"""
# Configuration module

This file contains all relevant configuration parameters for MSAexplorer, including
similarity matrices.
"""


import numpy as np

POSSIBLE_CHARS = [
        'A', 'R', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V', '*', 'X',  #AS
        'A', 'T','U', 'C', 'G', 'R', 'Y', 'S', 'W', 'K', 'M', 'B', 'D', 'H', 'V','N',  #RNA/DNA
        '-'  #GAP
]

AMBIG_CHARS = {
        'DNA': {
                'R': ['A', 'G'],
                'Y': ['C', 'T'],
                'S': ['G', 'C'],
                'W': ['A', 'T'],
                'K': ['G', 'T'],
                'M': ['A', 'C'],
                'B': ['C', 'G', 'T'],
                'D': ['A', 'G', 'T'],
                'H': ['A', 'C', 'T'],
                'V': ['A', 'C', 'G'],
                'N': ['A', 'C', 'G', 'T'],
                '-': ['A', 'C', 'G', 'T'],
        },
        'RNA': {
                'R': ['A', 'G'],
                'Y': ['C', 'U'],
                'S': ['G', 'C'],
                'W': ['A', 'U'],
                'K': ['G', 'U'],
                'M': ['A', 'C'],
                'B': ['C', 'G', 'U'],
                'D': ['A', 'G', 'U'],
                'H': ['A', 'C', 'U'],
                'V': ['A', 'C', 'G'],
                'N': ['A', 'C', 'G', 'U'],
                '-': ['A', 'C', 'G', 'U'],
        },
        'AA': {
                'X': ['A', 'R', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V', '*'],
                '-': ['A', 'R', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V', '*']
        }
    }

COMPLEMENT = {
        'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C',
        'R': 'Y', 'Y': 'R', 'S': 'S', 'W': 'W',  # R = A/G, Y = C/T, S = G/C, W = A/T
        'K': 'M', 'M': 'K', 'B': 'V', 'V': 'B',  # K = G/T, M = A/C, B = C/G/T, V = A/C/G
        'D': 'H', 'H': 'D', 'N': 'N', '-': '-',  # D = A/G/T, H = A/C/T, N = any base, '-' = gap
        'U': 'A'                                 # RNA
    }

START_CODONS = {
        'DNA': ['ATG'],
        'RNA': ['AUG'],
}

STOP_CODONS = {
        'DNA': ['TAG', 'TGA', 'TAA'],
        'RNA': ['UAG', 'UGA', 'UAA'],
}

# from: https://github.com/biopython/biopython/tree/master/Bio/Align/substitution_matrices/data
# extracted to nested dictionaries dependent on the alignment type
SUBS_MATRICES = {
        'AA': {
                'BLOSUM45': {
                        '*': {'*': 1.0, 'A': -5.0, 'B': -5.0, 'C': -5.0, 'D': -5.0, 'E': -5.0, 'F': -5.0, 'G': -5.0, 'H': -5.0, 'I': -5.0, 'K': -5.0, 'L': -5.0, 'M': -5.0, 'N': -5.0, 'P': -5.0, 'Q': -5.0, 'R': -5.0, 'S': -5.0, 'T': -5.0, 'V': -5.0, 'W': -5.0, 'X': -5.0, 'Y': -5.0, 'Z': -5.0},
                        'A': {'*': -5.0, 'A': 5.0, 'B': -1.0, 'C': -1.0, 'D': -2.0, 'E': -1.0, 'F': -2.0, 'G': 0.0, 'H': -2.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -2.0, 'S': 1.0, 'T': 0.0, 'V': 0.0, 'W': -2.0, 'X': 0.0, 'Y': -2.0, 'Z': -1.0},
                        'B': {'*': -5.0, 'A': -1.0, 'B': 4.0, 'C': -2.0, 'D': 5.0, 'E': 1.0, 'F': -3.0, 'G': -1.0, 'H': 0.0, 'I': -3.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 4.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': 2.0},
                        'C': {'*': -5.0, 'A': -1.0, 'B': -2.0, 'C': 12.0, 'D': -3.0, 'E': -3.0, 'F': -2.0, 'G': -3.0, 'H': -3.0, 'I': -3.0, 'K': -3.0, 'L': -2.0, 'M': -2.0, 'N': -2.0, 'P': -4.0, 'Q': -3.0, 'R': -3.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -5.0, 'X': -2.0, 'Y': -3.0, 'Z': -3.0},
                        'D': {'*': -5.0, 'A': -2.0, 'B': 5.0, 'C': -3.0, 'D': 7.0, 'E': 2.0, 'F': -4.0, 'G': -1.0, 'H': 0.0, 'I': -4.0, 'K': 0.0, 'L': -3.0, 'M': -3.0, 'N': 2.0, 'P': -1.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': 1.0},
                        'E': {'*': -5.0, 'A': -1.0, 'B': 1.0, 'C': -3.0, 'D': 2.0, 'E': 6.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 1.0, 'L': -2.0, 'M': -2.0, 'N': 0.0, 'P': 0.0, 'Q': 2.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 4.0},
                        'F': {'*': -5.0, 'A': -2.0, 'B': -3.0, 'C': -2.0, 'D': -4.0, 'E': -3.0, 'F': 8.0, 'G': -3.0, 'H': -2.0, 'I': 0.0, 'K': -3.0, 'L': 1.0, 'M': 0.0, 'N': -2.0, 'P': -3.0, 'Q': -4.0, 'R': -2.0, 'S': -2.0, 'T': -1.0, 'V': 0.0, 'W': 1.0, 'X': -1.0, 'Y': 3.0, 'Z': -3.0},
                        'G': {'*': -5.0, 'A': 0.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'E': -2.0, 'F': -3.0, 'G': 7.0, 'H': -2.0, 'I': -4.0, 'K': -2.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': -2.0, 'Q': -2.0, 'R': -2.0, 'S': 0.0, 'T': -2.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': -3.0, 'Z': -2.0},
                        'H': {'*': -5.0, 'A': -2.0, 'B': 0.0, 'C': -3.0, 'D': 0.0, 'E': 0.0, 'F': -2.0, 'G': -2.0, 'H': 10.0, 'I': -3.0, 'K': -1.0, 'L': -2.0, 'M': 0.0, 'N': 1.0, 'P': -2.0, 'Q': 1.0, 'R': 0.0, 'S': -1.0, 'T': -2.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': 2.0, 'Z': 0.0},
                        'I': {'*': -5.0, 'A': -1.0, 'B': -3.0, 'C': -3.0, 'D': -4.0, 'E': -3.0, 'F': 0.0, 'G': -4.0, 'H': -3.0, 'I': 5.0, 'K': -3.0, 'L': 2.0, 'M': 2.0, 'N': -2.0, 'P': -2.0, 'Q': -2.0, 'R': -3.0, 'S': -2.0, 'T': -1.0, 'V': 3.0, 'W': -2.0, 'X': -1.0, 'Y': 0.0, 'Z': -3.0},
                        'K': {'*': -5.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': 0.0, 'E': 1.0, 'F': -3.0, 'G': -2.0, 'H': -1.0, 'I': -3.0, 'K': 5.0, 'L': -3.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': 1.0, 'R': 3.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': 1.0},
                        'L': {'*': -5.0, 'A': -1.0, 'B': -3.0, 'C': -2.0, 'D': -3.0, 'E': -2.0, 'F': 1.0, 'G': -3.0, 'H': -2.0, 'I': 2.0, 'K': -3.0, 'L': 5.0, 'M': 2.0, 'N': -3.0, 'P': -3.0, 'Q': -2.0, 'R': -2.0, 'S': -3.0, 'T': -1.0, 'V': 1.0, 'W': -2.0, 'X': -1.0, 'Y': 0.0, 'Z': -2.0},
                        'M': {'*': -5.0, 'A': -1.0, 'B': -2.0, 'C': -2.0, 'D': -3.0, 'E': -2.0, 'F': 0.0, 'G': -2.0, 'H': 0.0, 'I': 2.0, 'K': -1.0, 'L': 2.0, 'M': 6.0, 'N': -2.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': -2.0, 'T': -1.0, 'V': 1.0, 'W': -2.0, 'X': -1.0, 'Y': 0.0, 'Z': -1.0},
                        'N': {'*': -5.0, 'A': -1.0, 'B': 4.0, 'C': -2.0, 'D': 2.0, 'E': 0.0, 'F': -2.0, 'G': 0.0, 'H': 1.0, 'I': -2.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 6.0, 'P': -2.0, 'Q': 0.0, 'R': 0.0, 'S': 1.0, 'T': 0.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'P': {'*': -5.0, 'A': -1.0, 'B': -2.0, 'C': -4.0, 'D': -1.0, 'E': 0.0, 'F': -3.0, 'G': -2.0, 'H': -2.0, 'I': -2.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': -2.0, 'P': 9.0, 'Q': -1.0, 'R': -2.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -3.0, 'Z': -1.0},
                        'Q': {'*': -5.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': 0.0, 'E': 2.0, 'F': -4.0, 'G': -2.0, 'H': 1.0, 'I': -2.0, 'K': 1.0, 'L': -2.0, 'M': 0.0, 'N': 0.0, 'P': -1.0, 'Q': 6.0, 'R': 1.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': 4.0},
                        'R': {'*': -5.0, 'A': -2.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'E': 0.0, 'F': -2.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 3.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': -2.0, 'Q': 1.0, 'R': 7.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': 0.0},
                        'S': {'*': -5.0, 'A': 1.0, 'B': 0.0, 'C': -1.0, 'D': 0.0, 'E': 0.0, 'F': -2.0, 'G': 0.0, 'H': -1.0, 'I': -2.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': 1.0, 'P': -1.0, 'Q': 0.0, 'R': -1.0, 'S': 4.0, 'T': 2.0, 'V': -1.0, 'W': -4.0, 'X': 0.0, 'Y': -2.0, 'Z': 0.0},
                        'T': {'*': -5.0, 'A': 0.0, 'B': 0.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': -1.0, 'G': -2.0, 'H': -2.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 2.0, 'T': 5.0, 'V': 0.0, 'W': -3.0, 'X': 0.0, 'Y': -1.0, 'Z': -1.0},
                        'V': {'*': -5.0, 'A': 0.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -3.0, 'F': 0.0, 'G': -3.0, 'H': -3.0, 'I': 3.0, 'K': -2.0, 'L': 1.0, 'M': 1.0, 'N': -3.0, 'P': -3.0, 'Q': -3.0, 'R': -2.0, 'S': -1.0, 'T': 0.0, 'V': 5.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'W': {'*': -5.0, 'A': -2.0, 'B': -4.0, 'C': -5.0, 'D': -4.0, 'E': -3.0, 'F': 1.0, 'G': -2.0, 'H': -3.0, 'I': -2.0, 'K': -2.0, 'L': -2.0, 'M': -2.0, 'N': -4.0, 'P': -3.0, 'Q': -2.0, 'R': -2.0, 'S': -4.0, 'T': -3.0, 'V': -3.0, 'W': 15.0, 'X': -2.0, 'Y': 3.0, 'Z': -2.0},
                        'X': {'*': -5.0, 'A': 0.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'E': -1.0, 'F': -1.0, 'G': -1.0, 'H': -1.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': -1.0},
                        'Y': {'*': -5.0, 'A': -2.0, 'B': -2.0, 'C': -3.0, 'D': -2.0, 'E': -2.0, 'F': 3.0, 'G': -3.0, 'H': 2.0, 'I': 0.0, 'K': -1.0, 'L': 0.0, 'M': 0.0, 'N': -2.0, 'P': -3.0, 'Q': -1.0, 'R': -1.0, 'S': -2.0, 'T': -1.0, 'V': -1.0, 'W': 3.0, 'X': -1.0, 'Y': 8.0, 'Z': -2.0},
                        'Z': {'*': -5.0, 'A': -1.0, 'B': 2.0, 'C': -3.0, 'D': 1.0, 'E': 4.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 1.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': 4.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': -2.0, 'Z': 4.0}
                },
                'BLOSUM50': {
                        '*': {'*': 1.0, 'A': -5.0, 'B': -5.0, 'C': -5.0, 'D': -5.0, 'E': -5.0, 'F': -5.0, 'G': -5.0, 'H': -5.0, 'I': -5.0, 'K': -5.0, 'L': -5.0, 'M': -5.0, 'N': -5.0, 'P': -5.0, 'Q': -5.0, 'R': -5.0, 'S': -5.0, 'T': -5.0, 'V': -5.0, 'W': -5.0, 'X': -5.0, 'Y': -5.0, 'Z': -5.0},
                        'A': {'*': -5.0, 'A': 5.0, 'B': -2.0, 'C': -1.0, 'D': -2.0, 'E': -1.0, 'F': -3.0, 'G': 0.0, 'H': -2.0, 'I': -1.0, 'K': -1.0, 'L': -2.0, 'M': -1.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -2.0, 'S': 1.0, 'T': 0.0, 'V': 0.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': -1.0},
                        'B': {'*': -5.0, 'A': -2.0, 'B': 5.0, 'C': -3.0, 'D': 5.0, 'E': 1.0, 'F': -4.0, 'G': -1.0, 'H': 0.0, 'I': -4.0, 'K': 0.0, 'L': -4.0, 'M': -3.0, 'N': 4.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -4.0, 'W': -5.0, 'X': -1.0, 'Y': -3.0, 'Z': 2.0},
                        'C': {'*': -5.0, 'A': -1.0, 'B': -3.0, 'C': 13.0, 'D': -4.0, 'E': -3.0, 'F': -2.0, 'G': -3.0, 'H': -3.0, 'I': -2.0, 'K': -3.0, 'L': -2.0, 'M': -2.0, 'N': -2.0, 'P': -4.0, 'Q': -3.0, 'R': -4.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -5.0, 'X': -2.0, 'Y': -3.0, 'Z': -3.0},
                        'D': {'*': -5.0, 'A': -2.0, 'B': 5.0, 'C': -4.0, 'D': 8.0, 'E': 2.0, 'F': -5.0, 'G': -1.0, 'H': -1.0, 'I': -4.0, 'K': -1.0, 'L': -4.0, 'M': -4.0, 'N': 2.0, 'P': -1.0, 'Q': 0.0, 'R': -2.0, 'S': 0.0, 'T': -1.0, 'V': -4.0, 'W': -5.0, 'X': -1.0, 'Y': -3.0, 'Z': 1.0},
                        'E': {'*': -5.0, 'A': -1.0, 'B': 1.0, 'C': -3.0, 'D': 2.0, 'E': 6.0, 'F': -3.0, 'G': -3.0, 'H': 0.0, 'I': -4.0, 'K': 1.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': -1.0, 'Q': 2.0, 'R': 0.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 5.0},
                        'F': {'*': -5.0, 'A': -3.0, 'B': -4.0, 'C': -2.0, 'D': -5.0, 'E': -3.0, 'F': 8.0, 'G': -4.0, 'H': -1.0, 'I': 0.0, 'K': -4.0, 'L': 1.0, 'M': 0.0, 'N': -4.0, 'P': -4.0, 'Q': -4.0, 'R': -3.0, 'S': -3.0, 'T': -2.0, 'V': -1.0, 'W': 1.0, 'X': -2.0, 'Y': 4.0, 'Z': -4.0},
                        'G': {'*': -5.0, 'A': 0.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'E': -3.0, 'F': -4.0, 'G': 8.0, 'H': -2.0, 'I': -4.0, 'K': -2.0, 'L': -4.0, 'M': -3.0, 'N': 0.0, 'P': -2.0, 'Q': -2.0, 'R': -3.0, 'S': 0.0, 'T': -2.0, 'V': -4.0, 'W': -3.0, 'X': -2.0, 'Y': -3.0, 'Z': -2.0},
                        'H': {'*': -5.0, 'A': -2.0, 'B': 0.0, 'C': -3.0, 'D': -1.0, 'E': 0.0, 'F': -1.0, 'G': -2.0, 'H': 10.0, 'I': -4.0, 'K': 0.0, 'L': -3.0, 'M': -1.0, 'N': 1.0, 'P': -2.0, 'Q': 1.0, 'R': 0.0, 'S': -1.0, 'T': -2.0, 'V': -4.0, 'W': -3.0, 'X': -1.0, 'Y': 2.0, 'Z': 0.0},
                        'I': {'*': -5.0, 'A': -1.0, 'B': -4.0, 'C': -2.0, 'D': -4.0, 'E': -4.0, 'F': 0.0, 'G': -4.0, 'H': -4.0, 'I': 5.0, 'K': -3.0, 'L': 2.0, 'M': 2.0, 'N': -3.0, 'P': -3.0, 'Q': -3.0, 'R': -4.0, 'S': -3.0, 'T': -1.0, 'V': 4.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'K': {'*': -5.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': -1.0, 'E': 1.0, 'F': -4.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 6.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': -1.0, 'Q': 2.0, 'R': 3.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 1.0},
                        'L': {'*': -5.0, 'A': -2.0, 'B': -4.0, 'C': -2.0, 'D': -4.0, 'E': -3.0, 'F': 1.0, 'G': -4.0, 'H': -3.0, 'I': 2.0, 'K': -3.0, 'L': 5.0, 'M': 3.0, 'N': -4.0, 'P': -4.0, 'Q': -2.0, 'R': -3.0, 'S': -3.0, 'T': -1.0, 'V': 1.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'M': {'*': -5.0, 'A': -1.0, 'B': -3.0, 'C': -2.0, 'D': -4.0, 'E': -2.0, 'F': 0.0, 'G': -3.0, 'H': -1.0, 'I': 2.0, 'K': -2.0, 'L': 3.0, 'M': 7.0, 'N': -2.0, 'P': -3.0, 'Q': 0.0, 'R': -2.0, 'S': -2.0, 'T': -1.0, 'V': 1.0, 'W': -1.0, 'X': -1.0, 'Y': 0.0, 'Z': -1.0},
                        'N': {'*': -5.0, 'A': -1.0, 'B': 4.0, 'C': -2.0, 'D': 2.0, 'E': 0.0, 'F': -4.0, 'G': 0.0, 'H': 1.0, 'I': -3.0, 'K': 0.0, 'L': -4.0, 'M': -2.0, 'N': 7.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': 1.0, 'T': 0.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'P': {'*': -5.0, 'A': -1.0, 'B': -2.0, 'C': -4.0, 'D': -1.0, 'E': -1.0, 'F': -4.0, 'G': -2.0, 'H': -2.0, 'I': -3.0, 'K': -1.0, 'L': -4.0, 'M': -3.0, 'N': -2.0, 'P': 10.0, 'Q': -1.0, 'R': -3.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -4.0, 'X': -2.0, 'Y': -3.0, 'Z': -1.0},
                        'Q': {'*': -5.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': 0.0, 'E': 2.0, 'F': -4.0, 'G': -2.0, 'H': 1.0, 'I': -3.0, 'K': 2.0, 'L': -2.0, 'M': 0.0, 'N': 0.0, 'P': -1.0, 'Q': 7.0, 'R': 1.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -1.0, 'X': -1.0, 'Y': -1.0, 'Z': 4.0},
                        'R': {'*': -5.0, 'A': -2.0, 'B': -1.0, 'C': -4.0, 'D': -2.0, 'E': 0.0, 'F': -3.0, 'G': -3.0, 'H': 0.0, 'I': -4.0, 'K': 3.0, 'L': -3.0, 'M': -2.0, 'N': -1.0, 'P': -3.0, 'Q': 1.0, 'R': 7.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': 0.0},
                        'S': {'*': -5.0, 'A': 1.0, 'B': 0.0, 'C': -1.0, 'D': 0.0, 'E': -1.0, 'F': -3.0, 'G': 0.0, 'H': -1.0, 'I': -3.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 1.0, 'P': -1.0, 'Q': 0.0, 'R': -1.0, 'S': 5.0, 'T': 2.0, 'V': -2.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'T': {'*': -5.0, 'A': 0.0, 'B': 0.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': -2.0, 'G': -2.0, 'H': -2.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 2.0, 'T': 5.0, 'V': 0.0, 'W': -3.0, 'X': 0.0, 'Y': -2.0, 'Z': -1.0},
                        'V': {'*': -5.0, 'A': 0.0, 'B': -4.0, 'C': -1.0, 'D': -4.0, 'E': -3.0, 'F': -1.0, 'G': -4.0, 'H': -4.0, 'I': 4.0, 'K': -3.0, 'L': 1.0, 'M': 1.0, 'N': -3.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -2.0, 'T': 0.0, 'V': 5.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'W': {'*': -5.0, 'A': -3.0, 'B': -5.0, 'C': -5.0, 'D': -5.0, 'E': -3.0, 'F': 1.0, 'G': -3.0, 'H': -3.0, 'I': -3.0, 'K': -3.0, 'L': -2.0, 'M': -1.0, 'N': -4.0, 'P': -4.0, 'Q': -1.0, 'R': -3.0, 'S': -4.0, 'T': -3.0, 'V': -3.0, 'W': 15.0, 'X': -3.0, 'Y': 2.0, 'Z': -2.0},
                        'X': {'*': -5.0, 'A': -1.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'E': -1.0, 'F': -2.0, 'G': -2.0, 'H': -1.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -1.0, 'P': -2.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': 0.0, 'V': -1.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -1.0},
                        'Y': {'*': -5.0, 'A': -2.0, 'B': -3.0, 'C': -3.0, 'D': -3.0, 'E': -2.0, 'F': 4.0, 'G': -3.0, 'H': 2.0, 'I': -1.0, 'K': -2.0, 'L': -1.0, 'M': 0.0, 'N': -2.0, 'P': -3.0, 'Q': -1.0, 'R': -1.0, 'S': -2.0, 'T': -2.0, 'V': -1.0, 'W': 2.0, 'X': -1.0, 'Y': 8.0, 'Z': -2.0},
                        'Z': {'*': -5.0, 'A': -1.0, 'B': 2.0, 'C': -3.0, 'D': 1.0, 'E': 5.0, 'F': -4.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 1.0, 'L': -3.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': 4.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': -2.0, 'Z': 5.0}},
                'BLOSUM65': {
                        '*': {'*': 1.0, 'A': -4.0, 'B': -4.0, 'C': -4.0, 'D': -4.0, 'E': -4.0, 'F': -4.0, 'G': -4.0, 'H': -4.0, 'I': -4.0, 'K': -4.0, 'L': -4.0, 'M': -4.0, 'N': -4.0, 'P': -4.0, 'Q': -4.0, 'R': -4.0, 'S': -4.0, 'T': -4.0, 'V': -4.0, 'W': -4.0, 'X': -4.0, 'Y': -4.0, 'Z': -4.0},
                        'A': {'*': -4.0, 'A': 4.0, 'B': -2.0, 'C': 0.0, 'D': -2.0, 'E': -1.0, 'F': -2.0, 'G': 0.0, 'H': -2.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -2.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 1.0, 'T': 0.0, 'V': 0.0, 'W': -3.0, 'X': 0.0, 'Y': -2.0, 'Z': -1.0},
                        'B': {'*': -4.0, 'A': -2.0, 'B': 4.0, 'C': -3.0, 'D': 4.0, 'E': 1.0, 'F': -3.0, 'G': -1.0, 'H': 0.0, 'I': -3.0, 'K': 0.0, 'L': -4.0, 'M': -3.0, 'N': 3.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': 1.0},
                        'C': {'*': -4.0, 'A': 0.0, 'B': -3.0, 'C': 9.0, 'D': -3.0, 'E': -4.0, 'F': -2.0, 'G': -3.0, 'H': -3.0, 'I': -1.0, 'K': -3.0, 'L': -1.0, 'M': -1.0, 'N': -3.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -2.0, 'X': -2.0, 'Y': -2.0, 'Z': -3.0},
                        'D': {'*': -4.0, 'A': -2.0, 'B': 4.0, 'C': -3.0, 'D': 6.0, 'E': 2.0, 'F': -3.0, 'G': -1.0, 'H': -1.0, 'I': -3.0, 'K': -1.0, 'L': -4.0, 'M': -3.0, 'N': 1.0, 'P': -1.0, 'Q': 0.0, 'R': -2.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': 1.0},
                        'E': {'*': -4.0, 'A': -1.0, 'B': 1.0, 'C': -4.0, 'D': 2.0, 'E': 5.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 1.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': -1.0, 'Q': 2.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 4.0},
                        'F': {'*': -4.0, 'A': -2.0, 'B': -3.0, 'C': -2.0, 'D': -3.0, 'E': -3.0, 'F': 6.0, 'G': -3.0, 'H': -1.0, 'I': 0.0, 'K': -3.0, 'L': 0.0, 'M': 0.0, 'N': -3.0, 'P': -4.0, 'Q': -3.0, 'R': -3.0, 'S': -2.0, 'T': -2.0, 'V': -1.0, 'W': 1.0, 'X': -1.0, 'Y': 3.0, 'Z': -3.0},
                        'G': {'*': -4.0, 'A': 0.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'E': -2.0, 'F': -3.0, 'G': 6.0, 'H': -2.0, 'I': -4.0, 'K': -2.0, 'L': -4.0, 'M': -3.0, 'N': 0.0, 'P': -2.0, 'Q': -2.0, 'R': -2.0, 'S': 0.0, 'T': -2.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': -3.0, 'Z': -2.0},
                        'H': {'*': -4.0, 'A': -2.0, 'B': 0.0, 'C': -3.0, 'D': -1.0, 'E': 0.0, 'F': -1.0, 'G': -2.0, 'H': 8.0, 'I': -3.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': 1.0, 'P': -2.0, 'Q': 0.0, 'R': 0.0, 'S': -1.0, 'T': -2.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': 2.0, 'Z': 0.0},
                        'I': {'*': -4.0, 'A': -1.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -3.0, 'F': 0.0, 'G': -4.0, 'H': -3.0, 'I': 4.0, 'K': -3.0, 'L': 2.0, 'M': 1.0, 'N': -3.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -2.0, 'T': -1.0, 'V': 3.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'K': {'*': -4.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': -1.0, 'E': 1.0, 'F': -3.0, 'G': -2.0, 'H': -1.0, 'I': -3.0, 'K': 5.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': 1.0, 'R': 2.0, 'S': 0.0, 'T': -1.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 1.0},
                        'L': {'*': -4.0, 'A': -1.0, 'B': -4.0, 'C': -1.0, 'D': -4.0, 'E': -3.0, 'F': 0.0, 'G': -4.0, 'H': -3.0, 'I': 2.0, 'K': -2.0, 'L': 4.0, 'M': 2.0, 'N': -3.0, 'P': -3.0, 'Q': -2.0, 'R': -2.0, 'S': -2.0, 'T': -1.0, 'V': 1.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'M': {'*': -4.0, 'A': -1.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -2.0, 'F': 0.0, 'G': -3.0, 'H': -2.0, 'I': 1.0, 'K': -1.0, 'L': 2.0, 'M': 5.0, 'N': -2.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': 1.0, 'W': -1.0, 'X': -1.0, 'Y': -1.0, 'Z': -1.0},
                        'N': {'*': -4.0, 'A': -2.0, 'B': 3.0, 'C': -3.0, 'D': 1.0, 'E': 0.0, 'F': -3.0, 'G': 0.0, 'H': 1.0, 'I': -3.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 6.0, 'P': -2.0, 'Q': 0.0, 'R': 0.0, 'S': 1.0, 'T': 0.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'P': {'*': -4.0, 'A': -1.0, 'B': -2.0, 'C': -3.0, 'D': -1.0, 'E': -1.0, 'F': -4.0, 'G': -2.0, 'H': -2.0, 'I': -3.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': -2.0, 'P': 7.0, 'Q': -1.0, 'R': -2.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -4.0, 'X': -2.0, 'Y': -3.0, 'Z': -1.0},
                        'Q': {'*': -4.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': 0.0, 'E': 2.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 1.0, 'L': -2.0, 'M': 0.0, 'N': 0.0, 'P': -1.0, 'Q': 5.0, 'R': 1.0, 'S': 0.0, 'T': -1.0, 'V': -2.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': 3.0},
                        'R': {'*': -4.0, 'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -2.0, 'E': 0.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 2.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': -2.0, 'Q': 1.0, 'R': 5.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'S': {'*': -4.0, 'A': 1.0, 'B': 0.0, 'C': -1.0, 'D': 0.0, 'E': 0.0, 'F': -2.0, 'G': 0.0, 'H': -1.0, 'I': -2.0, 'K': 0.0, 'L': -2.0, 'M': -1.0, 'N': 1.0, 'P': -1.0, 'Q': 0.0, 'R': -1.0, 'S': 4.0, 'T': 1.0, 'V': -2.0, 'W': -3.0, 'X': 0.0, 'Y': -2.0, 'Z': 0.0},
                        'T': {'*': -4.0, 'A': 0.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': -2.0, 'G': -2.0, 'H': -2.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 1.0, 'T': 5.0, 'V': 0.0, 'W': -2.0, 'X': 0.0, 'Y': -2.0, 'Z': -1.0},
                        'V': {'*': -4.0, 'A': 0.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -2.0, 'F': -1.0, 'G': -3.0, 'H': -3.0, 'I': 3.0, 'K': -2.0, 'L': 1.0, 'M': 1.0, 'N': -3.0, 'P': -2.0, 'Q': -2.0, 'R': -3.0, 'S': -2.0, 'T': 0.0, 'V': 4.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -2.0},
                        'W': {'*': -4.0, 'A': -3.0, 'B': -4.0, 'C': -2.0, 'D': -4.0, 'E': -3.0, 'F': 1.0, 'G': -2.0, 'H': -2.0, 'I': -3.0, 'K': -3.0, 'L': -2.0, 'M': -1.0, 'N': -4.0, 'P': -4.0, 'Q': -2.0, 'R': -3.0, 'S': -3.0, 'T': -2.0, 'V': -3.0, 'W': 11.0, 'X': -2.0, 'Y': 2.0, 'Z': -3.0},
                        'X': {'*': -4.0, 'A': 0.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'E': -1.0, 'F': -1.0, 'G': -1.0, 'H': -1.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -1.0, 'P': -2.0, 'Q': -1.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': -1.0},
                        'Y': {'*': -4.0, 'A': -2.0, 'B': -3.0, 'C': -2.0, 'D': -3.0, 'E': -2.0, 'F': 3.0, 'G': -3.0, 'H': 2.0, 'I': -1.0, 'K': -2.0, 'L': -1.0, 'M': -1.0, 'N': -2.0, 'P': -3.0, 'Q': -1.0, 'R': -2.0, 'S': -2.0, 'T': -2.0, 'V': -1.0, 'W': 2.0, 'X': -1.0, 'Y': 7.0, 'Z': -2.0},
                        'Z': {'*': -4.0, 'A': -1.0, 'B': 1.0, 'C': -3.0, 'D': 1.0, 'E': 4.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 1.0, 'L': -3.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': 3.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 4.0}
                },
                'BLOSUM80': {
                        '*': {'*': 1.0, 'A': -8.0, 'B': -8.0, 'C': -8.0, 'D': -8.0, 'E': -8.0, 'F': -8.0, 'G': -8.0, 'H': -8.0, 'I': -8.0, 'K': -8.0, 'L': -8.0, 'M': -8.0, 'N': -8.0, 'P': -8.0, 'Q': -8.0, 'R': -8.0, 'S': -8.0, 'T': -8.0, 'V': -8.0, 'W': -8.0, 'X': -8.0, 'Y': -8.0, 'Z': -8.0},
                        'A': {'*': -8.0, 'A': 7.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -2.0, 'F': -4.0, 'G': 0.0, 'H': -3.0, 'I': -3.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': -3.0, 'P': -1.0, 'Q': -2.0, 'R': -3.0, 'S': 2.0, 'T': 0.0, 'V': -1.0, 'W': -5.0, 'X': -1.0, 'Y': -4.0, 'Z': -2.0},
                        'B': {'*': -8.0, 'A': -3.0, 'B': 6.0, 'C': -6.0, 'D': 6.0, 'E': 1.0, 'F': -6.0, 'G': -2.0, 'H': -1.0, 'I': -6.0, 'K': -1.0, 'L': -7.0, 'M': -5.0, 'N': 5.0, 'P': -4.0, 'Q': -1.0, 'R': -2.0, 'S': 0.0, 'T': -1.0, 'V': -6.0, 'W': -8.0, 'X': -3.0, 'Y': -5.0, 'Z': 0.0},
                        'C': {'*': -8.0, 'A': -1.0, 'B': -6.0, 'C': 13.0, 'D': -7.0, 'E': -7.0, 'F': -4.0, 'G': -6.0, 'H': -7.0, 'I': -2.0, 'K': -6.0, 'L': -3.0, 'M': -3.0, 'N': -5.0, 'P': -6.0, 'Q': -5.0, 'R': -6.0, 'S': -2.0, 'T': -2.0, 'V': -2.0, 'W': -5.0, 'X': -4.0, 'Y': -5.0, 'Z': -7.0},
                        'D': {'*': -8.0, 'A': -3.0, 'B': 6.0, 'C': -7.0, 'D': 10.0, 'E': 2.0, 'F': -6.0, 'G': -3.0, 'H': -2.0, 'I': -7.0, 'K': -2.0, 'L': -7.0, 'M': -6.0, 'N': 2.0, 'P': -3.0, 'Q': -1.0, 'R': -3.0, 'S': -1.0, 'T': -2.0, 'V': -6.0, 'W': -8.0, 'X': -3.0, 'Y': -6.0, 'Z': 1.0},
                        'E': {'*': -8.0, 'A': -2.0, 'B': 1.0, 'C': -7.0, 'D': 2.0, 'E': 8.0, 'F': -6.0, 'G': -4.0, 'H': 0.0, 'I': -6.0, 'K': 1.0, 'L': -6.0, 'M': -4.0, 'N': -1.0, 'P': -2.0, 'Q': 3.0, 'R': -1.0, 'S': -1.0, 'T': -2.0, 'V': -4.0, 'W': -6.0, 'X': -2.0, 'Y': -5.0, 'Z': 6.0},
                        'F': {'*': -8.0, 'A': -4.0, 'B': -6.0, 'C': -4.0, 'D': -6.0, 'E': -6.0, 'F': 10.0, 'G': -6.0, 'H': -2.0, 'I': -1.0, 'K': -5.0, 'L': 0.0, 'M': 0.0, 'N': -6.0, 'P': -6.0, 'Q': -5.0, 'R': -5.0, 'S': -4.0, 'T': -4.0, 'V': -2.0, 'W': 0.0, 'X': -3.0, 'Y': 4.0, 'Z': -6.0},
                        'G': {'*': -8.0, 'A': 0.0, 'B': -2.0, 'C': -6.0, 'D': -3.0, 'E': -4.0, 'F': -6.0, 'G': 9.0, 'H': -4.0, 'I': -7.0, 'K': -3.0, 'L': -7.0, 'M': -5.0, 'N': -1.0, 'P': -5.0, 'Q': -4.0, 'R': -4.0, 'S': -1.0, 'T': -3.0, 'V': -6.0, 'W': -6.0, 'X': -3.0, 'Y': -6.0, 'Z': -4.0},
                        'H': {'*': -8.0, 'A': -3.0, 'B': -1.0, 'C': -7.0, 'D': -2.0, 'E': 0.0, 'F': -2.0, 'G': -4.0, 'H': 12.0, 'I': -6.0, 'K': -1.0, 'L': -5.0, 'M': -4.0, 'N': 1.0, 'P': -4.0, 'Q': 1.0, 'R': 0.0, 'S': -2.0, 'T': -3.0, 'V': -5.0, 'W': -4.0, 'X': -2.0, 'Y': 3.0, 'Z': 0.0},
                        'I': {'*': -8.0, 'A': -3.0, 'B': -6.0, 'C': -2.0, 'D': -7.0, 'E': -6.0, 'F': -1.0, 'G': -7.0, 'H': -6.0, 'I': 7.0, 'K': -5.0, 'L': 2.0, 'M': 2.0, 'N': -6.0, 'P': -5.0, 'Q': -5.0, 'R': -5.0, 'S': -4.0, 'T': -2.0, 'V': 4.0, 'W': -5.0, 'X': -2.0, 'Y': -3.0, 'Z': -6.0},
                        'K': {'*': -8.0, 'A': -1.0, 'B': -1.0, 'C': -6.0, 'D': -2.0, 'E': 1.0, 'F': -5.0, 'G': -3.0, 'H': -1.0, 'I': -5.0, 'K': 8.0, 'L': -4.0, 'M': -3.0, 'N': 0.0, 'P': -2.0, 'Q': 2.0, 'R': 3.0, 'S': -1.0, 'T': -1.0, 'V': -4.0, 'W': -6.0, 'X': -2.0, 'Y': -4.0, 'Z': 1.0},
                        'L': {'*': -8.0, 'A': -3.0, 'B': -7.0, 'C': -3.0, 'D': -7.0, 'E': -6.0, 'F': 0.0, 'G': -7.0, 'H': -5.0, 'I': 2.0, 'K': -4.0, 'L': 6.0, 'M': 3.0, 'N': -6.0, 'P': -5.0, 'Q': -4.0, 'R': -4.0, 'S': -4.0, 'T': -3.0, 'V': 1.0, 'W': -4.0, 'X': -2.0, 'Y': -2.0, 'Z': -5.0},
                        'M': {'*': -8.0, 'A': -2.0, 'B': -5.0, 'C': -3.0, 'D': -6.0, 'E': -4.0, 'F': 0.0, 'G': -5.0, 'H': -4.0, 'I': 2.0, 'K': -3.0, 'L': 3.0, 'M': 9.0, 'N': -4.0, 'P': -4.0, 'Q': -1.0, 'R': -3.0, 'S': -3.0, 'T': -1.0, 'V': 1.0, 'W': -3.0, 'X': -2.0, 'Y': -3.0, 'Z': -3.0},
                        'N': {'*': -8.0, 'A': -3.0, 'B': 5.0, 'C': -5.0, 'D': 2.0, 'E': -1.0, 'F': -6.0, 'G': -1.0, 'H': 1.0, 'I': -6.0, 'K': 0.0, 'L': -6.0, 'M': -4.0, 'N': 9.0, 'P': -4.0, 'Q': 0.0, 'R': -1.0, 'S': 1.0, 'T': 0.0, 'V': -5.0, 'W': -7.0, 'X': -2.0, 'Y': -4.0, 'Z': -1.0},
                        'P': {'*': -8.0, 'A': -1.0, 'B': -4.0, 'C': -6.0, 'D': -3.0, 'E': -2.0, 'F': -6.0, 'G': -5.0, 'H': -4.0, 'I': -5.0, 'K': -2.0, 'L': -5.0, 'M': -4.0, 'N': -4.0, 'P': 12.0, 'Q': -3.0, 'R': -3.0, 'S': -2.0, 'T': -3.0, 'V': -4.0, 'W': -7.0, 'X': -3.0, 'Y': -6.0, 'Z': -2.0},
                        'Q': {'*': -8.0, 'A': -2.0, 'B': -1.0, 'C': -5.0, 'D': -1.0, 'E': 3.0, 'F': -5.0, 'G': -4.0, 'H': 1.0, 'I': -5.0, 'K': 2.0, 'L': -4.0, 'M': -1.0, 'N': 0.0, 'P': -3.0, 'Q': 9.0, 'R': 1.0, 'S': -1.0, 'T': -1.0, 'V': -4.0, 'W': -4.0, 'X': -2.0, 'Y': -3.0, 'Z': 5.0},
                        'R': {'*': -8.0, 'A': -3.0, 'B': -2.0, 'C': -6.0, 'D': -3.0, 'E': -1.0, 'F': -5.0, 'G': -4.0, 'H': 0.0, 'I': -5.0, 'K': 3.0, 'L': -4.0, 'M': -3.0, 'N': -1.0, 'P': -3.0, 'Q': 1.0, 'R': 9.0, 'S': -2.0, 'T': -2.0, 'V': -4.0, 'W': -5.0, 'X': -2.0, 'Y': -4.0, 'Z': 0.0},
                        'S': {'*': -8.0, 'A': 2.0, 'B': 0.0, 'C': -2.0, 'D': -1.0, 'E': -1.0, 'F': -4.0, 'G': -1.0, 'H': -2.0, 'I': -4.0, 'K': -1.0, 'L': -4.0, 'M': -3.0, 'N': 1.0, 'P': -2.0, 'Q': -1.0, 'R': -2.0, 'S': 7.0, 'T': 2.0, 'V': -3.0, 'W': -6.0, 'X': -1.0, 'Y': -3.0, 'Z': -1.0},
                        'T': {'*': -8.0, 'A': 0.0, 'B': -1.0, 'C': -2.0, 'D': -2.0, 'E': -2.0, 'F': -4.0, 'G': -3.0, 'H': -3.0, 'I': -2.0, 'K': -1.0, 'L': -3.0, 'M': -1.0, 'N': 0.0, 'P': -3.0, 'Q': -1.0, 'R': -2.0, 'S': 2.0, 'T': 8.0, 'V': 0.0, 'W': -5.0, 'X': -1.0, 'Y': -3.0, 'Z': -2.0},
                        'V': {'*': -8.0, 'A': -1.0, 'B': -6.0, 'C': -2.0, 'D': -6.0, 'E': -4.0, 'F': -2.0, 'G': -6.0, 'H': -5.0, 'I': 4.0, 'K': -4.0, 'L': 1.0, 'M': 1.0, 'N': -5.0, 'P': -4.0, 'Q': -4.0, 'R': -4.0, 'S': -3.0, 'T': 0.0, 'V': 7.0, 'W': -5.0, 'X': -2.0, 'Y': -3.0, 'Z': -4.0},
                        'W': {'*': -8.0, 'A': -5.0, 'B': -8.0, 'C': -5.0, 'D': -8.0, 'E': -6.0, 'F': 0.0, 'G': -6.0, 'H': -4.0, 'I': -5.0, 'K': -6.0, 'L': -4.0, 'M': -3.0, 'N': -7.0, 'P': -7.0, 'Q': -4.0, 'R': -5.0, 'S': -6.0, 'T': -5.0, 'V': -5.0, 'W': 16.0, 'X': -5.0, 'Y': 3.0, 'Z': -5.0},
                        'X': {'*': -8.0, 'A': -1.0, 'B': -3.0, 'C': -4.0, 'D': -3.0, 'E': -2.0, 'F': -3.0, 'G': -3.0, 'H': -2.0, 'I': -2.0, 'K': -2.0, 'L': -2.0, 'M': -2.0, 'N': -2.0, 'P': -3.0, 'Q': -2.0, 'R': -2.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -5.0, 'X': -2.0, 'Y': -3.0, 'Z': -1.0},
                        'Y': {'*': -8.0, 'A': -4.0, 'B': -5.0, 'C': -5.0, 'D': -6.0, 'E': -5.0, 'F': 4.0, 'G': -6.0, 'H': 3.0, 'I': -3.0, 'K': -4.0, 'L': -2.0, 'M': -3.0, 'N': -4.0, 'P': -6.0, 'Q': -3.0, 'R': -4.0, 'S': -3.0, 'T': -3.0, 'V': -3.0, 'W': 3.0, 'X': -3.0, 'Y': 11.0, 'Z': -4.0},
                        'Z': {'*': -8.0, 'A': -2.0, 'B': 0.0, 'C': -7.0, 'D': 1.0, 'E': 6.0, 'F': -6.0, 'G': -4.0, 'H': 0.0, 'I': -6.0, 'K': 1.0, 'L': -5.0, 'M': -3.0, 'N': -1.0, 'P': -2.0, 'Q': 5.0, 'R': 0.0, 'S': -1.0, 'T': -2.0, 'V': -4.0, 'W': -5.0, 'X': -1.0, 'Y': -4.0, 'Z': 6.0}
                },
                'BLOSUM90': {
                         '*': {'*': 1.0, 'A': -6.0, 'B': -6.0, 'C': -6.0, 'D': -6.0, 'E': -6.0, 'F': -6.0, 'G': -6.0, 'H': -6.0, 'I': -6.0, 'K': -6.0, 'L': -6.0, 'M': -6.0, 'N': -6.0, 'P': -6.0, 'Q': -6.0, 'R': -6.0, 'S': -6.0, 'T': -6.0, 'V': -6.0, 'W': -6.0, 'X': -6.0, 'Y': -6.0, 'Z': -6.0},
                         'A': {'*': -6.0, 'A': 5.0, 'B': -2.0, 'C': -1.0, 'D': -3.0, 'E': -1.0, 'F': -3.0, 'G': 0.0, 'H': -2.0, 'I': -2.0, 'K': -1.0, 'L': -2.0, 'M': -2.0, 'N': -2.0, 'P': -1.0, 'Q': -1.0, 'R': -2.0, 'S': 1.0, 'T': 0.0, 'V': -1.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': -1.0},
                         'B': {'*': -6.0, 'A': -2.0, 'B': 4.0, 'C': -4.0, 'D': 4.0, 'E': 0.0, 'F': -4.0, 'G': -2.0, 'H': -1.0, 'I': -5.0, 'K': -1.0, 'L': -5.0, 'M': -4.0, 'N': 4.0, 'P': -3.0, 'Q': -1.0, 'R': -2.0, 'S': 0.0, 'T': -1.0, 'V': -4.0, 'W': -6.0, 'X': -2.0, 'Y': -4.0, 'Z': 0.0},
                         'C': {'*': -6.0, 'A': -1.0, 'B': -4.0, 'C': 9.0, 'D': -5.0, 'E': -6.0, 'F': -3.0, 'G': -4.0, 'H': -5.0, 'I': -2.0, 'K': -4.0, 'L': -2.0, 'M': -2.0, 'N': -4.0, 'P': -4.0, 'Q': -4.0, 'R': -5.0, 'S': -2.0, 'T': -2.0, 'V': -2.0, 'W': -4.0, 'X': -3.0, 'Y': -4.0, 'Z': -5.0},
                         'D': {'*': -6.0, 'A': -3.0, 'B': 4.0, 'C': -5.0, 'D': 7.0, 'E': 1.0, 'F': -5.0, 'G': -2.0, 'H': -2.0, 'I': -5.0, 'K': -1.0, 'L': -5.0, 'M': -4.0, 'N': 1.0, 'P': -3.0, 'Q': -1.0, 'R': -3.0, 'S': -1.0, 'T': -2.0, 'V': -5.0, 'W': -6.0, 'X': -2.0, 'Y': -4.0, 'Z': 0.0},
                         'E': {'*': -6.0, 'A': -1.0, 'B': 0.0, 'C': -6.0, 'D': 1.0, 'E': 6.0, 'F': -5.0, 'G': -3.0, 'H': -1.0, 'I': -4.0, 'K': 0.0, 'L': -4.0, 'M': -3.0, 'N': -1.0, 'P': -2.0, 'Q': 2.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -5.0, 'X': -2.0, 'Y': -4.0, 'Z': 4.0},
                         'F': {'*': -6.0, 'A': -3.0, 'B': -4.0, 'C': -3.0, 'D': -5.0, 'E': -5.0, 'F': 7.0, 'G': -5.0, 'H': -2.0, 'I': -1.0, 'K': -4.0, 'L': 0.0, 'M': -1.0, 'N': -4.0, 'P': -4.0, 'Q': -4.0, 'R': -4.0, 'S': -3.0, 'T': -3.0, 'V': -2.0, 'W': 0.0, 'X': -2.0, 'Y': 3.0, 'Z': -4.0},
                         'G': {'*': -6.0, 'A': 0.0, 'B': -2.0, 'C': -4.0, 'D': -2.0, 'E': -3.0, 'F': -5.0, 'G': 6.0, 'H': -3.0, 'I': -5.0, 'K': -2.0, 'L': -5.0, 'M': -4.0, 'N': -1.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -1.0, 'T': -3.0, 'V': -5.0, 'W': -4.0, 'X': -2.0, 'Y': -5.0, 'Z': -3.0},
                         'H': {'*': -6.0, 'A': -2.0, 'B': -1.0, 'C': -5.0, 'D': -2.0, 'E': -1.0, 'F': -2.0, 'G': -3.0, 'H': 8.0, 'I': -4.0, 'K': -1.0, 'L': -4.0, 'M': -3.0, 'N': 0.0, 'P': -3.0, 'Q': 1.0, 'R': 0.0, 'S': -2.0, 'T': -2.0, 'V': -4.0, 'W': -3.0, 'X': -2.0, 'Y': 1.0, 'Z': 0.0},
                         'I': {'*': -6.0, 'A': -2.0, 'B': -5.0, 'C': -2.0, 'D': -5.0, 'E': -4.0, 'F': -1.0, 'G': -5.0, 'H': -4.0, 'I': 5.0, 'K': -4.0, 'L': 1.0, 'M': 1.0, 'N': -4.0, 'P': -4.0, 'Q': -4.0, 'R': -4.0, 'S': -3.0, 'T': -1.0, 'V': 3.0, 'W': -4.0, 'X': -2.0, 'Y': -2.0, 'Z': -4.0},
                         'K': {'*': -6.0, 'A': -1.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'E': 0.0, 'F': -4.0, 'G': -2.0, 'H': -1.0, 'I': -4.0, 'K': 6.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': -2.0, 'Q': 1.0, 'R': 2.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -5.0, 'X': -1.0, 'Y': -3.0, 'Z': 1.0},
                         'L': {'*': -6.0, 'A': -2.0, 'B': -5.0, 'C': -2.0, 'D': -5.0, 'E': -4.0, 'F': 0.0, 'G': -5.0, 'H': -4.0, 'I': 1.0, 'K': -3.0, 'L': 5.0, 'M': 2.0, 'N': -4.0, 'P': -4.0, 'Q': -3.0, 'R': -3.0, 'S': -3.0, 'T': -2.0, 'V': 0.0, 'W': -3.0, 'X': -2.0, 'Y': -2.0, 'Z': -4.0},
                         'M': {'*': -6.0, 'A': -2.0, 'B': -4.0, 'C': -2.0, 'D': -4.0, 'E': -3.0, 'F': -1.0, 'G': -4.0, 'H': -3.0, 'I': 1.0, 'K': -2.0, 'L': 2.0, 'M': 7.0, 'N': -3.0, 'P': -3.0, 'Q': 0.0, 'R': -2.0, 'S': -2.0, 'T': -1.0, 'V': 0.0, 'W': -2.0, 'X': -1.0, 'Y': -2.0, 'Z': -2.0},
                         'N': {'*': -6.0, 'A': -2.0, 'B': 4.0, 'C': -4.0, 'D': 1.0, 'E': -1.0, 'F': -4.0, 'G': -1.0, 'H': 0.0, 'I': -4.0, 'K': 0.0, 'L': -4.0, 'M': -3.0, 'N': 7.0, 'P': -3.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -4.0, 'W': -5.0, 'X': -2.0, 'Y': -3.0, 'Z': -1.0},
                         'P': {'*': -6.0, 'A': -1.0, 'B': -3.0, 'C': -4.0, 'D': -3.0, 'E': -2.0, 'F': -4.0, 'G': -3.0, 'H': -3.0, 'I': -4.0, 'K': -2.0, 'L': -4.0, 'M': -3.0, 'N': -3.0, 'P': 8.0, 'Q': -2.0, 'R': -3.0, 'S': -2.0, 'T': -2.0, 'V': -3.0, 'W': -5.0, 'X': -2.0, 'Y': -4.0, 'Z': -2.0},
                         'Q': {'*': -6.0, 'A': -1.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'E': 2.0, 'F': -4.0, 'G': -3.0, 'H': 1.0, 'I': -4.0, 'K': 1.0, 'L': -3.0, 'M': 0.0, 'N': 0.0, 'P': -2.0, 'Q': 7.0, 'R': 1.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -3.0, 'Z': 4.0},
                         'R': {'*': -6.0, 'A': -2.0, 'B': -2.0, 'C': -5.0, 'D': -3.0, 'E': -1.0, 'F': -4.0, 'G': -3.0, 'H': 0.0, 'I': -4.0, 'K': 2.0, 'L': -3.0, 'M': -2.0, 'N': -1.0, 'P': -3.0, 'Q': 1.0, 'R': 6.0, 'S': -1.0, 'T': -2.0, 'V': -3.0, 'W': -4.0, 'X': -2.0, 'Y': -3.0, 'Z': 0.0},
                         'S': {'*': -6.0, 'A': 1.0, 'B': 0.0, 'C': -2.0, 'D': -1.0, 'E': -1.0, 'F': -3.0, 'G': -1.0, 'H': -2.0, 'I': -3.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': -2.0, 'Q': -1.0, 'R': -1.0, 'S': 5.0, 'T': 1.0, 'V': -2.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': -1.0},
                         'T': {'*': -6.0, 'A': 0.0, 'B': -1.0, 'C': -2.0, 'D': -2.0, 'E': -1.0, 'F': -3.0, 'G': -3.0, 'H': -2.0, 'I': -1.0, 'K': -1.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': -2.0, 'Q': -1.0, 'R': -2.0, 'S': 1.0, 'T': 6.0, 'V': -1.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': -1.0},
                         'V': {'*': -6.0, 'A': -1.0, 'B': -4.0, 'C': -2.0, 'D': -5.0, 'E': -3.0, 'F': -2.0, 'G': -5.0, 'H': -4.0, 'I': 3.0, 'K': -3.0, 'L': 0.0, 'M': 0.0, 'N': -4.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -2.0, 'T': -1.0, 'V': 5.0, 'W': -3.0, 'X': -2.0, 'Y': -3.0, 'Z': -3.0},
                         'W': {'*': -6.0, 'A': -4.0, 'B': -6.0, 'C': -4.0, 'D': -6.0, 'E': -5.0, 'F': 0.0, 'G': -4.0, 'H': -3.0, 'I': -4.0, 'K': -5.0, 'L': -3.0, 'M': -2.0, 'N': -5.0, 'P': -5.0, 'Q': -3.0, 'R': -4.0, 'S': -4.0, 'T': -4.0, 'V': -3.0, 'W': 11.0, 'X': -3.0, 'Y': 2.0, 'Z': -4.0},
                         'X': {'*': -6.0, 'A': -1.0, 'B': -2.0, 'C': -3.0, 'D': -2.0, 'E': -2.0, 'F': -2.0, 'G': -2.0, 'H': -2.0, 'I': -2.0, 'K': -1.0, 'L': -2.0, 'M': -1.0, 'N': -2.0, 'P': -2.0, 'Q': -1.0, 'R': -2.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -3.0, 'X': -2.0, 'Y': -2.0, 'Z': -1.0},
                         'Y': {'*': -6.0, 'A': -3.0, 'B': -4.0, 'C': -4.0, 'D': -4.0, 'E': -4.0, 'F': 3.0, 'G': -5.0, 'H': 1.0, 'I': -2.0, 'K': -3.0, 'L': -2.0, 'M': -2.0, 'N': -3.0, 'P': -4.0, 'Q': -3.0, 'R': -3.0, 'S': -3.0, 'T': -2.0, 'V': -3.0, 'W': 2.0, 'X': -2.0, 'Y': 8.0, 'Z': -3.0},
                         'Z': {'*': -6.0, 'A': -1.0, 'B': 0.0, 'C': -5.0, 'D': 0.0, 'E': 4.0, 'F': -4.0, 'G': -3.0, 'H': 0.0, 'I': -4.0, 'K': 1.0, 'L': -4.0, 'M': -2.0, 'N': -1.0, 'P': -2.0, 'Q': 4.0, 'R': 0.0, 'S': -1.0, 'T': -1.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': 4.0}
                         },
                'BLOSUM95': {
                    'A': {'A': 5, 'R': -2, 'N': -1, 'D': -2, 'C': -1, 'Q': -1, 'E': -1, 'G': 0, 'H': -2, 'I': -2, 'L': -2, 'K': -1, 'M': -1, 'F': -3, 'P': -1, 'S': 1, 'T': 0, 'W': -3, 'Y': -2, 'V': 0},
                    'R': {'A': -2, 'R': 6, 'N': -1, 'D': -2, 'C': -4, 'Q': 1, 'E': -1, 'G': -3, 'H': 0, 'I': -3, 'L': -3, 'K': 3, 'M': -2, 'F': -3, 'P': -2, 'S': -1, 'T': -1, 'W': -3, 'Y': -1, 'V': -3},
                    'N': {'A': -1, 'R': -1, 'N': 7, 'D': 2, 'C': -3, 'Q': 0, 'E': -1, 'G': 0, 'H': 1, 'I': -4, 'L': -4, 'K': 0, 'M': -3, 'F': -4, 'P': -2, 'S': 1, 'T': 0, 'W': -4, 'Y': -2, 'V': -4},
                    'D': {'A': -2, 'R': -2, 'N': 2, 'D': 8, 'C': -4, 'Q': 0, 'E': 2, 'G': -1, 'H': -1, 'I': -4, 'L': -5, 'K': -1, 'M': -4, 'F': -5, 'P': -2, 'S': 0, 'T': -1, 'W': -5, 'Y': -3, 'V': -4},
                    'C': {'A': -1, 'R': -4, 'N': -3, 'D': -4, 'C': 13, 'Q': -3, 'E': -4, 'G': -3, 'H': -3, 'I': -2, 'L': -2, 'K': -3, 'M': -2, 'F': -3, 'P': -4, 'S': -1, 'T': -1, 'W': -5, 'Y': -3, 'V': -1},
                    'Q': {'A': -1, 'R': 1, 'N': 0, 'D': 0, 'C': -3, 'Q': 7, 'E': 2, 'G': -2, 'H': 1, 'I': -3, 'L': -2, 'K': 2, 'M': 0, 'F': -4, 'P': -1, 'S': 0, 'T': -1, 'W': -2, 'Y': -1, 'V': -3},
                    'E': {'A': -1, 'R': -1, 'N': -1, 'D': 2, 'C': -4, 'Q': 2, 'E': 6, 'G': -2, 'H': 0, 'I': -4, 'L': -3, 'K': 1, 'M': -2, 'F': -3, 'P': -1, 'S': 0, 'T': -1, 'W': -3, 'Y': -2, 'V': -3},
                    'G': {'A': 0, 'R': -3, 'N': 0, 'D': -1, 'C': -3, 'Q': -2, 'E': -2, 'G': 8, 'H': -2, 'I': -4, 'L': -4, 'K': -2, 'M': -3, 'F': -4, 'P': -2, 'S': 0, 'T': -2, 'W': -4, 'Y': -3, 'V': -4},
                    'H': {'A': -2, 'R': 0, 'N': 1, 'D': -1, 'C': -3, 'Q': 1, 'E': 0, 'G': -2, 'H': 10, 'I': -4, 'L': -3, 'K': -1, 'M': -2, 'F': -2, 'P': -2, 'S': -1, 'T': -2, 'W': -3, 'Y': 2, 'V': -4},
                    'I': {'A': -2, 'R': -3, 'N': -4, 'D': -4, 'C': -2, 'Q': -3, 'E': -4, 'G': -4, 'H': -4, 'I': 5, 'L': 2, 'K': -3, 'M': 2, 'F': 0, 'P': -3, 'S': -2, 'T': -1, 'W': -3, 'Y': -1, 'V': 4},
                    'L': {'A': -2, 'R': -3, 'N': -4, 'D': -5, 'C': -2, 'Q': -2, 'E': -3, 'G': -4, 'H': -3, 'I': 2, 'L': 5, 'K': -3, 'M': 3, 'F': 1, 'P': -3, 'S': -3, 'T': -1, 'W': -2, 'Y': -1, 'V': 1},
                    'K': {'A': -1, 'R': 3, 'N': 0, 'D': -1, 'C': -3, 'Q': 2, 'E': 1, 'G': -2, 'H': -1, 'I': -3, 'L': -3, 'K': 6, 'M': -2, 'F': -3, 'P': -1, 'S': 0, 'T': -1, 'W': -3, 'Y': -2, 'V': -3},
                    'M': {'A': -1, 'R': -2, 'N': -3, 'D': -4, 'C': -2, 'Q': 0, 'E': -2, 'G': -3, 'H': -2, 'I': 2, 'L': 3, 'K': -2, 'M': 7, 'F': 0, 'P': -2, 'S': -2, 'T': -1, 'W': -1, 'Y': -1, 'V': 1},
                    'F': {'A': -3, 'R': -3, 'N': -4, 'D': -5, 'C': -3, 'Q': -4, 'E': -3, 'G': -4, 'H': -2, 'I': 0, 'L': 1, 'K': -3, 'M': 0, 'F': 8, 'P': -4, 'S': -3, 'T': -2, 'W': 1, 'Y': 4, 'V': -1},
                    'P': {'A': -1, 'R': -2, 'N': -2, 'D': -2, 'C': -4, 'Q': -1, 'E': -1, 'G': -2, 'H': -2, 'I': -3, 'L': -3, 'K': -1, 'M': -2, 'F': -4, 'P': 10, 'S': -1, 'T': -1, 'W': -4, 'Y': -3, 'V': -3},
                    'S': {'A': 1, 'R': -1, 'N': 1, 'D': 0, 'C': -1, 'Q': 0, 'E': 0, 'G': 0, 'H': -1, 'I': -2, 'L': -3, 'K': 0, 'M': -2, 'F': -3, 'P': -1, 'S': 5, 'T': 2, 'W': -4, 'Y': -2, 'V': -2},
                    'T': {'A': 0, 'R': -1, 'N': 0, 'D': -1, 'C': -1, 'Q': -1, 'E': -1, 'G': -2, 'H': -2, 'I': -1, 'L': -1, 'K': -1, 'M': -1, 'F': -2, 'P': -1, 'S': 2, 'T': 5, 'W': -3, 'Y': -2, 'V': 0},
                    'W': {'A': -3, 'R': -3, 'N': -4, 'D': -5, 'C': -5, 'Q': -2, 'E': -3, 'G': -4, 'H': -3, 'I': -3, 'L': -2, 'K': -3, 'M': -1, 'F': 1, 'P': -4, 'S': -4, 'T': -3, 'W': 15, 'Y': 2, 'V': -3},
                    'Y': {'A': -2, 'R': -1, 'N': -2, 'D': -3, 'C': -3, 'Q': -1, 'E': -2, 'G': -3, 'H': 2, 'I': -1, 'L': -1, 'K': -2, 'M': -1, 'F': 4, 'P': -3, 'S': -2, 'T': -2, 'W': 2, 'Y': 8, 'V': -1},
                    'V': {'A': 0, 'R': -3, 'N': -4, 'D': -4, 'C': -1, 'Q': -3, 'E': -3, 'G': -4, 'H': -4, 'I': 4, 'L': 1, 'K': -3, 'M': 1, 'F': -1, 'P': -3, 'S': -2, 'T': 0, 'W': -3, 'Y': -1, 'V': 5}
                },
                'BENNER22': {
                        'A': {'A': 2.5, 'C': -1.2, 'D': -0.2, 'E': -0.3, 'F': -3.1, 'G': 0.8, 'H': -1.6, 'I': -0.4, 'K': -1.0, 'L': -1.7, 'M': -0.8, 'N': 0.0, 'P': 0.8, 'Q': -0.9, 'R': -1.2, 'S': 1.3, 'T': 1.4, 'V': 0.4, 'W': -5.5, 'Y': -3.5},
                        'C': {'A': -1.2, 'C': 12.6, 'D': -3.7, 'E': -4.3, 'F': -0.1, 'G': -1.7, 'H': -1.5, 'I': -2.4, 'K': -3.3, 'L': -2.6, 'M': -2.5, 'N': -1.9, 'P': -3.1, 'Q': -3.3, 'R': -1.6, 'S': 0.3, 'T': -1.1, 'V': -1.7, 'W': 0.5, 'Y': 0.6},
                        'D': {'A': -0.2, 'C': -3.7, 'D': 4.8, 'E': 3.9, 'F': -5.4, 'G': 0.7, 'H': 0.3, 'I': -4.0, 'K': 0.2, 'L': -4.9, 'M': -3.9, 'N': 2.4, 'P': -1.8, 'Q': 0.6, 'R': -1.0, 'S': 0.1, 'T': -0.7, 'V': -3.0, 'W': -6.4, 'Y': -3.0},
                        'E': {'A': -0.3, 'C': -4.3, 'D': 3.9, 'E': 4.6, 'F': -5.7, 'G': 0.5, 'H': -0.2, 'I': -3.6, 'K': 1.0, 'L': -4.4, 'M': -3.4, 'N': 1.2, 'P': -1.7, 'Q': 1.7, 'R': -0.1, 'S': -0.5, 'T': -0.9, 'V': -2.7, 'W': -6.3, 'Y': -4.0},
                        'F': {'A': -3.1, 'C': -0.1, 'D': -5.4, 'E': -5.7, 'F': 7.7, 'G': -5.8, 'H': 0.3, 'I': 0.5, 'K': -5.1, 'L': 2.2, 'M': 0.7, 'N': -3.5, 'P': -3.4, 'Q': -3.6, 'R': -4.3, 'S': -2.2, 'T': -2.6, 'V': -0.1, 'W': 0.5, 'Y': 5.9},
                        'G': {'A': 0.8, 'C': -1.7, 'D': 0.7, 'E': 0.5, 'F': -5.8, 'G': 6.2, 'H': -2.0, 'I': -3.8, 'K': -1.0, 'L': -4.9, 'M': -3.8, 'N': 0.4, 'P': -1.8, 'Q': -1.4, 'R': -0.7, 'S': 0.6, 'T': -0.7, 'V': -2.5, 'W': -4.5, 'Y': -4.8},
                        'H': {'A': -1.6, 'C': -1.5, 'D': 0.3, 'E': -0.2, 'F': 0.3, 'G': -2.0, 'H': 6.1, 'I': -3.2, 'K': 0.8, 'L': -2.1, 'M': -2.4, 'N': 1.4, 'P': -0.4, 'Q': 2.4, 'R': 1.5, 'S': -0.5, 'T': -1.1, 'V': -3.0, 'W': -2.7, 'Y': 3.7},
                        'I': {'A': -0.4, 'C': -2.4, 'D': -4.0, 'E': -3.6, 'F': 0.5, 'G': -3.8, 'H': -3.2, 'I': 4.2, 'K': -3.0, 'L': 2.7, 'M': 3.1, 'N': -2.7, 'P': -2.3, 'Q': -2.7, 'R': -3.2, 'S': -1.4, 'T': 0.3, 'V': 3.6, 'W': -4.4, 'Y': -2.2},
                        'K': {'A': -1.0, 'C': -3.3, 'D': 0.2, 'E': 1.0, 'F': -5.1, 'G': -1.0, 'H': 0.8, 'I': -3.0, 'K': 4.4, 'L': -3.3, 'M': -2.0, 'N': 1.0, 'P': -1.6, 'Q': 2.2, 'R': 3.9, 'S': -0.4, 'T': -0.4, 'V': -2.7, 'W': -3.7, 'Y': -3.6},
                        'L': {'A': -1.7, 'C': -2.6, 'D': -4.9, 'E': -4.4, 'F': 2.2, 'G': -4.9, 'H': -2.1, 'I': 2.7, 'K': -3.3, 'L': 4.6, 'M': 3.2, 'N': -3.5, 'P': -1.3, 'Q': -2.0, 'R': -2.9, 'S': -2.1, 'T': -1.0, 'V': 2.0, 'W': -1.8, 'Y': -0.7},
                        'M': {'A': -0.8, 'C': -2.5, 'D': -3.9, 'E': -3.4, 'F': 0.7, 'G': -3.8, 'H': -2.4, 'I': 3.1, 'K': -2.0, 'L': 3.2, 'M': 4.9, 'N': -2.6, 'P': -2.0, 'Q': -1.7, 'R': -2.1, 'S': -1.5, 'T': 0.1, 'V': 2.5, 'W': -2.8, 'Y': -1.8},
                        'N': {'A': 0.0, 'C': -1.9, 'D': 2.4, 'E': 1.2, 'F': -3.5, 'G': 0.4, 'H': 1.4, 'I': -2.7, 'K': 1.0, 'L': -3.5, 'M': -2.6, 'N': 3.3, 'P': -1.1, 'Q': 0.5, 'R': 0.4, 'S': 1.1, 'T': 0.5, 'V': -2.3, 'W': -5.2, 'Y': -1.2},
                        'P': {'A': 0.8, 'C': -3.1, 'D': -1.8, 'E': -1.7, 'F': -3.4, 'G': -1.8, 'H': -0.4, 'I': -2.3, 'K': -1.6, 'L': -1.3, 'M': -2.0, 'N': -1.1, 'P': 7.0, 'Q': -0.1, 'R': -1.2, 'S': 1.1, 'T': 0.4, 'V': -1.7, 'W': -5.8, 'Y': -3.5},
                        'Q': {'A': -0.9, 'C': -3.3, 'D': 0.6, 'E': 1.7, 'F': -3.6, 'G': -1.4, 'H': 2.4, 'I': -2.7, 'K': 2.2, 'L': -2.0, 'M': -1.7, 'N': 0.5, 'P': -0.1, 'Q': 4.2, 'R': 2.2, 'S': -0.6, 'T': -0.7, 'V': -2.4, 'W': -3.3, 'Y': -1.9},
                        'R': {'A': -1.2, 'C': -1.6, 'D': -1.0, 'E': -0.1, 'F': -4.3, 'G': -0.7, 'H': 1.5, 'I': -3.2, 'K': 3.9, 'L': -2.9, 'M': -2.1, 'N': 0.4, 'P': -1.2, 'Q': 2.2, 'R': 5.0, 'S': -0.5, 'T': -0.7, 'V': -2.9, 'W': -1.1, 'Y': -2.7},
                        'S': {'A': 1.3, 'C': 0.3, 'D': 0.1, 'E': -0.5, 'F': -2.2, 'G': 0.6, 'H': -0.5, 'I': -1.4, 'K': -0.4, 'L': -2.1, 'M': -1.5, 'N': 1.1, 'P': 1.1, 'Q': -0.6, 'R': -0.5, 'S': 2.0, 'T': 1.5, 'V': -0.9, 'W': -3.9, 'Y': -1.9},
                        'T': {'A': 1.4, 'C': -1.1, 'D': -0.7, 'E': -0.9, 'F': -2.6, 'G': -0.7, 'H': -1.1, 'I': 0.3, 'K': -0.4, 'L': -1.0, 'M': 0.1, 'N': 0.5, 'P': 0.4, 'Q': -0.7, 'R': -0.7, 'S': 1.5, 'T': 2.5, 'V': 0.4, 'W': -4.5, 'Y': -3.0},
                        'V': {'A': 0.4, 'C': -1.7, 'D': -3.0, 'E': -2.7, 'F': -0.1, 'G': -2.5, 'H': -3.0, 'I': 3.6, 'K': -2.7, 'L': 2.0, 'M': 2.5, 'N': -2.3, 'P': -1.7, 'Q': -2.4, 'R': -2.9, 'S': -0.9, 'T': 0.4, 'V': 3.7, 'W': -4.5, 'Y': -2.6},
                        'W': {'A': -5.5, 'C': 0.5, 'D': -6.4, 'E': -6.3, 'F': 0.5, 'G': -4.5, 'H': -2.7, 'I': -4.4, 'K': -3.7, 'L': -1.8, 'M': -2.8, 'N': -5.2, 'P': -5.8, 'Q': -3.3, 'R': -1.1, 'S': -3.9, 'T': -4.5, 'V': -4.5, 'W': 15.7, 'Y': 1.5},
                        'Y': {'A': -3.5, 'C': 0.6, 'D': -3.0, 'E': -4.0, 'F': 5.9, 'G': -4.8, 'H': 3.7, 'I': -2.2, 'K': -3.6, 'L': -0.7, 'M': -1.8, 'N': -1.2, 'P': -3.5, 'Q': -1.9, 'R': -2.7, 'S': -1.9, 'T': -3.0, 'V': -2.6, 'W': 1.5, 'Y': 9.0}
                },
                'BENNER6': {
                        'A': {'A': 2.5, 'C': -1.7, 'D': -0.6, 'E': -0.7, 'F': -3.2, 'G': 0.8, 'H': -2.1, 'I': 0.1, 'K': -1.9, 'L': -1.3, 'M': -0.2, 'N': 0.0, 'P': 1.1, 'Q': -1.7, 'R': -1.7, 'S': 1.4, 'T': 1.7, 'V': 0.7, 'W': -4.3, 'Y': -4.0},
                        'C': {'A': -1.7, 'C': 12.1, 'D': -3.7, 'E': -4.7, 'F': -0.1, 'G': -1.3, 'H': -1.2, 'I': -3.6, 'K': -2.8, 'L': -3.8, 'M': -3.7, 'N': -1.6, 'P': -2.7, 'Q': -3.2, 'R': -0.4, 'S': 0.9, 'T': -1.5, 'V': -3.1, 'W': 1.6, 'Y': 2.6},
                        'D': {'A': -0.6, 'C': -3.7, 'D': 5.2, 'E': 4.4, 'F': -5.7, 'G': 0.8, 'H': 0.1, 'I': -4.2, 'K': -0.2, 'L': -5.3, 'M': -4.3, 'N': 2.5, 'P': -2.8, 'Q': 0.6, 'R': -1.5, 'S': -0.4, 'T': -1.2, 'V': -3.3, 'W': -6.3, 'Y': -2.3},
                        'E': {'A': -0.7, 'C': -4.7, 'D': 4.4, 'E': 5.2, 'F': -6.7, 'G': 0.5, 'H': -0.2, 'I': -4.1, 'K': 0.9, 'L': -5.0, 'M': -4.1, 'N': 1.1, 'P': -2.6, 'Q': 2.1, 'R': -0.4, 'S': -1.2, 'T': -1.6, 'V': -3.0, 'W': -5.6, 'Y': -4.1},
                        'F': {'A': -3.2, 'C': -0.1, 'D': -5.7, 'E': -6.7, 'F': 8.3, 'G': -5.7, 'H': 0.1, 'I': 0.0, 'K': -6.3, 'L': 2.4, 'M': -0.1, 'N': -3.5, 'P': -3.2, 'Q': -4.4, 'R': -4.9, 'S': -1.8, 'T': -2.4, 'V': -0.5, 'W': -1.6, 'Y': 5.6},
                        'G': {'A': 0.8, 'C': -1.3, 'D': 0.8, 'E': 0.5, 'F': -5.7, 'G': 5.8, 'H': -2.1, 'I': -3.4, 'K': -1.4, 'L': -4.6, 'M': -3.7, 'N': -0.1, 'P': -1.7, 'Q': -1.6, 'R': -0.1, 'S': 0.8, 'T': -0.5, 'V': -2.3, 'W': -1.7, 'Y': -4.9},
                        'H': {'A': -2.1, 'C': -1.2, 'D': 0.1, 'E': -0.2, 'F': 0.1, 'G': -2.1, 'H': 6.1, 'I': -3.7, 'K': 0.9, 'L': -2.2, 'M': -3.4, 'N': 1.4, 'P': -0.4, 'Q': 3.2, 'R': 1.8, 'S': -0.9, 'T': -1.7, 'V': -3.8, 'W': -2.8, 'Y': 4.4},
                        'I': {'A': 0.1, 'C': -3.6, 'D': -4.2, 'E': -4.1, 'F': 0.0, 'G': -3.4, 'H': -3.7, 'I': 4.4, 'K': -3.8, 'L': 2.4, 'M': 4.0, 'N': -2.5, 'P': -2.0, 'Q': -3.8, 'R': -3.8, 'S': -1.2, 'T': 0.7, 'V': 3.9, 'W': -5.0, 'Y': -3.3},
                        'K': {'A': -1.9, 'C': -2.8, 'D': -0.2, 'E': 0.9, 'F': -6.3, 'G': -1.4, 'H': 0.9, 'I': -3.8, 'K': 5.6, 'L': -4.1, 'M': -2.9, 'N': 1.0, 'P': -2.3, 'Q': 2.5, 'R': 4.3, 'S': -1.2, 'T': -1.1, 'V': -3.8, 'W': -1.4, 'Y': -4.0},
                        'L': {'A': -1.3, 'C': -3.8, 'D': -5.3, 'E': -5.0, 'F': 2.4, 'G': -4.6, 'H': -2.2, 'I': 2.4, 'K': -4.1, 'L': 4.8, 'M': -2.9, 'N': -3.4, 'P': -0.2, 'Q': -2.4, 'R': -3.2, 'S': -1.5, 'T': -0.4, 'V': 1.9, 'W': -3.0, 'Y': -1.6},
                        'M': {'A': -0.2, 'C': -3.7, 'D': -4.3, 'E': -4.1, 'F': -0.1, 'G': -3.7, 'H': -3.4, 'I': 4.0, 'K': -2.9, 'L': -2.9, 'M': 4.8, 'N': -2.5, 'P': -1.8, 'Q': -3.1, 'R': -3.0, 'S': -1.3, 'T': 0.6, 'V': 3.3, 'W': -4.4, 'Y': -3.6},
                        'N': {'A': 0.0, 'C': -1.6, 'D': 2.5, 'E': 1.1, 'F': -3.5, 'G': -0.1, 'H': 1.4, 'I': -2.5, 'K': 1.0, 'L': -3.4, 'M': -2.5, 'N': 3.6, 'P': -1.1, 'Q': 0.1, 'R': -0.1, 'S': 1.2, 'T': 0.5, 'V': -2.4, 'W': -4.4, 'Y': -0.9},
                        'P': {'A': 1.1, 'C': -2.7, 'D': -2.8, 'E': -2.6, 'F': -3.2, 'G': -1.7, 'H': -0.4, 'I': -2.0, 'K': -2.3, 'L': -0.2, 'M': -1.8, 'N': -1.1, 'P': 6.5, 'Q': 0.1, 'R': -1.3, 'S': 1.4, 'T': 0.6, 'V': -1.6, 'W': -4.8, 'Y': -3.8},
                        'Q': {'A': -1.7, 'C': -3.2, 'D': 0.6, 'E': 2.1, 'F': -4.4, 'G': -1.6, 'H': 3.2, 'I': -3.8, 'K': 2.5, 'L': -2.4, 'M': -3.1, 'N': 0.1, 'P': 0.1, 'Q': 5.3, 'R': 2.5, 'S': -1.4, 'T': -1.7, 'V': -3.5, 'W': -2.6, 'Y': -1.4},
                        'R': {'A': -1.7, 'C': -0.4, 'D': -1.5, 'E': -0.4, 'F': -4.9, 'G': -0.1, 'H': 1.8, 'I': -3.8, 'K': 4.3, 'L': -3.2, 'M': -3.0, 'N': -0.1, 'P': -1.3, 'Q': 2.5, 'R': 5.1, 'S': -0.9, 'T': -1.3, 'V': -3.7, 'W': 2.0, 'Y': -2.6},
                        'S': {'A': 1.4, 'C': 0.9, 'D': -0.4, 'E': -1.2, 'F': -1.8, 'G': 0.8, 'H': -0.9, 'I': -1.2, 'K': -1.2, 'L': -1.5, 'M': -1.3, 'N': 1.2, 'P': 1.4, 'Q': -1.4, 'R': -0.9, 'S': 2.1, 'T': 1.5, 'V': -0.9, 'W': -2.9, 'Y': -1.8},
                        'T': {'A': 1.7, 'C': -1.5, 'D': -1.2, 'E': -1.6, 'F': -2.4, 'G': -0.5, 'H': -1.7, 'I': 0.7, 'K': -1.1, 'L': -0.4, 'M': 0.6, 'N': 0.5, 'P': 0.6, 'Q': -1.7, 'R': -1.3, 'S': 1.5, 'T': 2.4, 'V': 0.6, 'W': -2.6, 'Y': -3.4},
                        'V': {'A': 0.7, 'C': -3.1, 'D': -3.3, 'E': -3.0, 'F': -0.5, 'G': -2.3, 'H': -3.8, 'I': 3.9, 'K': -3.8, 'L': 1.9, 'M': 3.3, 'N': -2.4, 'P': -1.6, 'Q': -3.5, 'R': -3.7, 'S': -0.9, 'T': 0.6, 'V': 4.0, 'W': -4.8, 'Y': -3.8},
                        'W': {'A': -4.3, 'C': 1.6, 'D': -6.3, 'E': -5.6, 'F': -1.6, 'G': -1.7, 'H': -2.8, 'I': -5.0, 'K': -1.4, 'L': -3.0, 'M': -4.4, 'N': -4.4, 'P': -4.8, 'Q': -2.6, 'R': 2.0, 'S': -2.9, 'T': -2.6, 'V': -4.8, 'W': 14.7, 'Y': -0.3},
                        'Y': {'A': -4.0, 'C': 2.6, 'D': -2.3, 'E': -4.1, 'F': 5.6, 'G': -4.9, 'H': 4.4, 'I': -3.3, 'K': -4.0, 'L': -1.6, 'M': -3.6, 'N': -0.9, 'P': -3.8, 'Q': -1.4, 'R': -2.6, 'S': -1.8, 'T': -3.4, 'V': -3.8, 'W': -0.3, 'Y': 9.5}
                },
                'BENNER74': {
                        'A': {'A': 2.4, 'C': 0.3, 'D': -0.3, 'E': -0.1, 'F': -2.6, 'G': 0.6, 'H': -1.0, 'I': -0.8, 'K': -0.4, 'L': -1.4, 'M': -0.8, 'N': -0.2, 'P': 0.4, 'Q': -0.3, 'R': -0.8, 'S': 1.1, 'T': 0.7, 'V': 0.1, 'W': -4.1, 'Y': -2.6},
                        'C': {'A': 0.3, 'C': 11.8, 'D': -3.2, 'E': -3.2, 'F': -0.7, 'G': -2.0, 'H': -1.3, 'I': -1.2, 'K': -2.9, 'L': -1.6, 'M': -1.2, 'N': -1.8, 'P': -3.1, 'Q': -2.6, 'R': -2.2, 'S': 0.1, 'T': -0.6, 'V': -0.2, 'W': -0.9, 'Y': -0.4},
                        'D': {'A': -0.3, 'C': -3.2, 'D': 4.8, 'E': 2.9, 'F': -4.7, 'G': 0.2, 'H': 0.4, 'I': -3.9, 'K': 0.4, 'L': -4.2, 'M': -3.2, 'N': 2.2, 'P': -1.0, 'Q': 0.8, 'R': -0.5, 'S': 0.4, 'T': -0.2, 'V': -2.9, 'W': -5.5, 'Y': -2.8},
                        'E': {'A': -0.1, 'C': -3.2, 'D': 2.9, 'E': 3.7, 'F': -4.3, 'G': -0.5, 'H': 0.2, 'I': -2.9, 'K': 1.2, 'L': -3.1, 'M': -2.2, 'N': 1.0, 'P': -0.7, 'Q': 1.7, 'R': 0.3, 'S': 0.1, 'T': -0.2, 'V': -2.1, 'W': -4.7, 'Y': -3.0},
                        'F': {'A': -2.6, 'C': -0.7, 'D': -4.7, 'E': -4.3, 'F': 7.2, 'G': -5.4, 'H': 0.0, 'I': 0.9, 'K': -3.6, 'L': 2.1, 'M': 1.3, 'N': -3.2, 'P': -3.8, 'Q': -2.8, 'R': -3.5, 'S': -2.6, 'T': -2.2, 'V': 0.1, 'W': 3.0, 'Y': 5.3},
                        'G': {'A': 0.6, 'C': -2.0, 'D': 0.2, 'E': -0.5, 'F': -5.4, 'G': 6.6, 'H': -1.6, 'I': -4.3, 'K': -1.1, 'L': -4.6, 'M': -3.5, 'N': 0.4, 'P': -1.7, 'Q': -1.1, 'R': -1.0, 'S': 0.4, 'T': -1.0, 'V': -3.1, 'W': -4.1, 'Y': -4.3},
                        'H': {'A': -1.0, 'C': -1.3, 'D': 0.4, 'E': 0.2, 'F': 0.0, 'G': -1.6, 'H': 6.1, 'I': -2.3, 'K': 0.6, 'L': -1.9, 'M': -1.5, 'N': 1.2, 'P': -1.0, 'Q': 1.4, 'R': 1.0, 'S': -0.3, 'T': -0.5, 'V': -2.1, 'W': -1.0, 'Y': 2.5},
                        'I': {'A': -0.8, 'C': -1.2, 'D': -3.9, 'E': -2.9, 'F': 0.9, 'G': -4.3, 'H': -2.3, 'I': 4.0, 'K': -2.3, 'L': 2.8, 'M': 2.6, 'N': -2.8, 'P': -2.6, 'Q': -2.0, 'R': -2.6, 'S': -1.8, 'T': -0.3, 'V': 3.2, 'W': -2.3, 'Y': -1.0},
                        'K': {'A': -0.4, 'C': -2.9, 'D': 0.4, 'E': 1.2, 'F': -3.6, 'G': -1.1, 'H': 0.6, 'I': -2.3, 'K': 3.4, 'L': -2.4, 'M': -1.5, 'N': 0.9, 'P': -0.8, 'Q': 1.7, 'R': 2.9, 'S': 0.0, 'T': 0.1, 'V': -1.9, 'W': -3.6, 'Y': -2.4},
                        'L': {'A': -1.4, 'C': -1.6, 'D': -4.2, 'E': -3.1, 'F': 2.1, 'G': -4.6, 'H': -1.9, 'I': 2.8, 'K': -2.4, 'L': 4.2, 'M': 2.9, 'N': -3.1, 'P': -2.2, 'Q': -1.7, 'R': -2.4, 'S': -2.2, 'T': -1.1, 'V': 1.9, 'W': -0.9, 'Y': -0.1},
                        'M': {'A': -0.8, 'C': -1.2, 'D': -3.2, 'E': -2.2, 'F': 1.3, 'G': -3.5, 'H': -1.5, 'I': 2.6, 'K': -1.5, 'L': 2.9, 'M': 4.5, 'N': -2.2, 'P': -2.4, 'Q': -1.0, 'R': -1.8, 'S': -1.4, 'T': -0.4, 'V': 1.8, 'W': -1.3, 'Y': -0.5},
                        'N': {'A': -0.2, 'C': -1.8, 'D': 2.2, 'E': 1.0, 'F': -3.2, 'G': 0.4, 'H': 1.2, 'I': -2.8, 'K': 0.9, 'L': -3.1, 'M': -2.2, 'N': 3.6, 'P': -1.0, 'Q': 0.7, 'R': 0.3, 'S': 0.9, 'T': 0.4, 'V': -2.2, 'W': -4.0, 'Y': -1.4},
                        'P': {'A': 0.4, 'C': -3.1, 'D': -1.0, 'E': -0.7, 'F': -3.8, 'G': -1.7, 'H': -1.0, 'I': -2.6, 'K': -0.8, 'L': -2.2, 'M': -2.4, 'N': -1.0, 'P': 7.5, 'Q': -0.2, 'R': -0.1, 'S': 0.5, 'T': 0.1, 'V': -1.9, 'W': -5.2, 'Y': -3.4},
                        'Q': {'A': -0.3, 'C': -2.6, 'D': 0.8, 'E': 1.7, 'F': -2.8, 'G': -1.1, 'H': 1.4, 'I': -2.0, 'K': 1.7, 'L': -1.7, 'M': -1.0, 'N': 0.7, 'P': -0.2, 'Q': 3.0, 'R': 1.6, 'S': 0.1, 'T': -0.1, 'V': -1.7, 'W': -2.8, 'Y': -1.8},
                        'R': {'A': -0.8, 'C': -2.2, 'D': -0.5, 'E': 0.3, 'F': -3.5, 'G': -1.0, 'H': 1.0, 'I': -2.6, 'K': 2.9, 'L': -2.4, 'M': -1.8, 'N': 0.3, 'P': -0.1, 'Q': 1.6, 'R': 4.8, 'S': -0.2, 'T': -0.3, 'V': -2.2, 'W': -1.6, 'Y': -2.0},
                        'S': {'A': 1.1, 'C': 0.1, 'D': 0.4, 'E': 0.1, 'F': -2.6, 'G': 0.4, 'H': -0.3, 'I': -1.8, 'K': 0.0, 'L': -2.2, 'M': -1.4, 'N': 0.9, 'P': 0.5, 'Q': 0.1, 'R': -0.2, 'S': 2.1, 'T': 1.4, 'V': -1.0, 'W': -3.4, 'Y': -1.9},
                        'T': {'A': 0.7, 'C': -0.6, 'D': -0.2, 'E': -0.2, 'F': -2.2, 'G': -1.0, 'H': -0.5, 'I': -0.3, 'K': 0.1, 'L': -1.1, 'M': -0.4, 'N': 0.4, 'P': 0.1, 'Q': -0.1, 'R': -0.3, 'S': 1.4, 'T': 2.5, 'V': 0.2, 'W': -3.7, 'Y': -2.1},
                        'V': {'A': 0.1, 'C': -0.2, 'D': -2.9, 'E': -2.1, 'F': 0.1, 'G': -3.1, 'H': -2.1, 'I': 3.2, 'K': -1.9, 'L': 1.9, 'M': 1.8, 'N': -2.2, 'P': -1.9, 'Q': -1.7, 'R': -2.2, 'S': -1.0, 'T': 0.2, 'V': 3.4, 'W': -2.9, 'Y': -1.4},
                        'W': {'A': -4.1, 'C': -0.9, 'D': -5.5, 'E': -4.7, 'F': 3.0, 'G': -4.1, 'H': -1.0, 'I': -2.3, 'K': -3.6, 'L': -0.9, 'M': -1.3, 'N': -4.0, 'P': -5.2, 'Q': -2.8, 'R': -1.6, 'S': -3.4, 'T': -3.7, 'V': -2.9, 'W': 14.7, 'Y': 3.6},
                        'Y': {'A': -2.6, 'C': -0.4, 'D': -2.8, 'E': -3.0, 'F': 5.3, 'G': -4.3, 'H': 2.5, 'I': -1.0, 'K': -2.4, 'L': -0.1, 'M': -0.5, 'N': -1.4, 'P': -3.4, 'Q': -1.8, 'R': -2.0, 'S': -1.9, 'T': -2.1, 'V': -1.4, 'W': 3.6, 'Y': 8.1}

                },
                'BLASTP': {
                        '*': {'*': 1.0, 'A': -4.0, 'B': -4.0, 'C': -4.0, 'D': -4.0, 'E': -4.0, 'F': -4.0, 'G': -4.0, 'H': -4.0, 'I': -4.0, 'J': -4.0, 'K': -4.0, 'L': -4.0, 'M': -4.0, 'N': -4.0, 'O': -4.0, 'P': -4.0, 'Q': -4.0, 'R': -4.0, 'S': -4.0, 'T': -4.0, 'U': -4.0, 'V': -4.0, 'W': -4.0, 'X': -4.0, 'Y': -4.0, 'Z': -4.0},
                        'A': {'*': -4.0, 'A': 4.0, 'B': -2.0, 'C': 0.0, 'D': -2.0, 'E': -1.0, 'F': -2.0, 'G': 0.0, 'H': -2.0, 'I': -1.0, 'J': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -2.0, 'O': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 1.0, 'T': 0.0, 'U': 0.0, 'V': 0.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': -1.0},
                        'B': {'*': -4.0, 'A': -2.0, 'B': 4.0, 'C': -3.0, 'D': 4.0, 'E': 1.0, 'F': -3.0, 'G': -1.0, 'H': 0.0, 'I': -3.0, 'J': -3.0, 'K': 0.0, 'L': -4.0, 'M': -3.0, 'N': 4.0, 'O': -1.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': -1.0, 'U': -3.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': 0.0},
                        'C': {'*': -4.0, 'A': 0.0, 'B': -3.0, 'C': 9.0, 'D': -3.0, 'E': -4.0, 'F': -2.0, 'G': -3.0, 'H': -3.0, 'I': -1.0, 'J': -1.0, 'K': -3.0, 'L': -1.0, 'M': -1.0, 'N': -3.0, 'O': -1.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -1.0, 'T': -1.0, 'U': 9.0, 'V': -1.0, 'W': -2.0, 'X': -1.0, 'Y': -2.0, 'Z': -3.0},
                        'D': {'*': -4.0, 'A': -2.0, 'B': 4.0, 'C': -3.0, 'D': 6.0, 'E': 2.0, 'F': -3.0, 'G': -1.0, 'H': -1.0, 'I': -3.0, 'J': -3.0, 'K': -1.0, 'L': -4.0, 'M': -3.0, 'N': 1.0, 'O': -1.0, 'P': -1.0, 'Q': 0.0, 'R': -2.0, 'S': 0.0, 'T': -1.0, 'U': -3.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': 1.0},
                        'E': {'*': -4.0, 'A': -1.0, 'B': 1.0, 'C': -4.0, 'D': 2.0, 'E': 5.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'J': -3.0, 'K': 1.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'O': -1.0, 'P': -1.0, 'Q': 2.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'U': -4.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 4.0},
                        'F': {'*': -4.0, 'A': -2.0, 'B': -3.0, 'C': -2.0, 'D': -3.0, 'E': -3.0, 'F': 6.0, 'G': -3.0, 'H': -1.0, 'I': 0.0, 'J': 0.0, 'K': -3.0, 'L': 0.0, 'M': 0.0, 'N': -3.0, 'O': -1.0, 'P': -4.0, 'Q': -3.0, 'R': -3.0, 'S': -2.0, 'T': -2.0, 'U': -2.0, 'V': -1.0, 'W': 1.0, 'X': -1.0, 'Y': 3.0, 'Z': -3.0},
                        'G': {'*': -4.0, 'A': 0.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'E': -2.0, 'F': -3.0, 'G': 6.0, 'H': -2.0, 'I': -4.0, 'J': -4.0, 'K': -2.0, 'L': -4.0, 'M': -3.0, 'N': 0.0, 'O': -1.0, 'P': -2.0, 'Q': -2.0, 'R': -2.0, 'S': 0.0, 'T': -2.0, 'U': -3.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': -3.0, 'Z': -2.0},
                        'H': {'*': -4.0, 'A': -2.0, 'B': 0.0, 'C': -3.0, 'D': -1.0, 'E': 0.0, 'F': -1.0, 'G': -2.0, 'H': 8.0, 'I': -3.0, 'J': -3.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': 1.0, 'O': -1.0, 'P': -2.0, 'Q': 0.0, 'R': 0.0, 'S': -1.0, 'T': -2.0, 'U': -3.0, 'V': -3.0, 'W': -2.0, 'X': -1.0, 'Y': 2.0, 'Z': 0.0},
                        'I': {'*': -4.0, 'A': -1.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -3.0, 'F': 0.0, 'G': -4.0, 'H': -3.0, 'I': 4.0, 'J': 3.0, 'K': -3.0, 'L': 2.0, 'M': 1.0, 'N': -3.0, 'O': -1.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -2.0, 'T': -1.0, 'U': -1.0, 'V': 3.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'J': {'*': -4.0, 'A': -1.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -3.0, 'F': 0.0, 'G': -4.0, 'H': -3.0, 'I': 3.0, 'J': 3.0, 'K': -3.0, 'L': 3.0, 'M': 2.0, 'N': -3.0, 'O': -1.0, 'P': -3.0, 'Q': -2.0, 'R': -2.0, 'S': -2.0, 'T': -1.0, 'U': -1.0, 'V': 2.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'K': {'*': -4.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': -1.0, 'E': 1.0, 'F': -3.0, 'G': -2.0, 'H': -1.0, 'I': -3.0, 'J': -3.0, 'K': 5.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'O': -1.0, 'P': -1.0, 'Q': 1.0, 'R': 2.0, 'S': 0.0, 'T': -1.0, 'U': -3.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 1.0},
                        'L': {'*': -4.0, 'A': -1.0, 'B': -4.0, 'C': -1.0, 'D': -4.0, 'E': -3.0, 'F': 0.0, 'G': -4.0, 'H': -3.0, 'I': 2.0, 'J': 3.0, 'K': -2.0, 'L': 4.0, 'M': 2.0, 'N': -3.0, 'O': -1.0, 'P': -3.0, 'Q': -2.0, 'R': -2.0, 'S': -2.0, 'T': -1.0, 'U': -1.0, 'V': 1.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'M': {'*': -4.0, 'A': -1.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -2.0, 'F': 0.0, 'G': -3.0, 'H': -2.0, 'I': 1.0, 'J': 2.0, 'K': -1.0, 'L': 2.0, 'M': 5.0, 'N': -2.0, 'O': -1.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'U': -1.0, 'V': 1.0, 'W': -1.0, 'X': -1.0, 'Y': -1.0, 'Z': -1.0},
                        'N': {'*': -4.0, 'A': -2.0, 'B': 4.0, 'C': -3.0, 'D': 1.0, 'E': 0.0, 'F': -3.0, 'G': 0.0, 'H': 1.0, 'I': -3.0, 'J': -3.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 6.0, 'O': -1.0, 'P': -2.0, 'Q': 0.0, 'R': 0.0, 'S': 1.0, 'T': 0.0, 'U': -3.0, 'V': -3.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'O': {'*': -4.0, 'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': -1.0, 'G': -1.0, 'H': -1.0, 'I': -1.0, 'J': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -1.0, 'O': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'X': -1.0, 'Y': -1.0, 'Z': -1.0},
                        'P': {'*': -4.0, 'A': -1.0, 'B': -2.0, 'C': -3.0, 'D': -1.0, 'E': -1.0, 'F': -4.0, 'G': -2.0, 'H': -2.0, 'I': -3.0, 'J': -3.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': -2.0, 'O': -1.0, 'P': 7.0, 'Q': -1.0, 'R': -2.0, 'S': -1.0, 'T': -1.0, 'U': -3.0, 'V': -2.0, 'W': -4.0, 'X': -1.0, 'Y': -3.0, 'Z': -1.0},
                        'Q': {'*': -4.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': 0.0, 'E': 2.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'J': -2.0, 'K': 1.0, 'L': -2.0, 'M': 0.0, 'N': 0.0, 'O': -1.0, 'P': -1.0, 'Q': 5.0, 'R': 1.0, 'S': 0.0, 'T': -1.0, 'U': -3.0, 'V': -2.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': 4.0},
                        'R': {'*': -4.0, 'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -2.0, 'E': 0.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'J': -2.0, 'K': 2.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'O': -1.0, 'P': -2.0, 'Q': 1.0, 'R': 5.0, 'S': -1.0, 'T': -1.0, 'U': -3.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'S': {'*': -4.0, 'A': 1.0, 'B': 0.0, 'C': -1.0, 'D': 0.0, 'E': 0.0, 'F': -2.0, 'G': 0.0, 'H': -1.0, 'I': -2.0, 'J': -2.0, 'K': 0.0, 'L': -2.0, 'M': -1.0, 'N': 1.0, 'O': -1.0, 'P': -1.0, 'Q': 0.0, 'R': -1.0, 'S': 4.0, 'T': 1.0, 'U': -1.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': -2.0, 'Z': 0.0},
                        'T': {'*': -4.0, 'A': 0.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': -2.0, 'G': -2.0, 'H': -2.0, 'I': -1.0, 'J': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'O': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 1.0, 'T': 5.0, 'U': -1.0, 'V': 0.0, 'W': -2.0, 'X': -1.0, 'Y': -2.0, 'Z': -1.0},
                        'U': {'*': -4.0, 'A': 0.0, 'B': -3.0, 'C': 9.0, 'D': -3.0, 'E': -4.0, 'F': -2.0, 'G': -3.0, 'H': -3.0, 'I': -1.0, 'J': -1.0, 'K': -3.0, 'L': -1.0, 'M': -1.0, 'N': -3.0, 'O': -1.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -1.0, 'T': -1.0, 'U': 9.0, 'V': -1.0, 'W': -2.0, 'X': -1.0, 'Y': -2.0, 'Z': -3.0},
                        'V': {'*': -4.0, 'A': 0.0, 'B': -3.0, 'C': -1.0, 'D': -3.0, 'E': -2.0, 'F': -1.0, 'G': -3.0, 'H': -3.0, 'I': 3.0, 'J': 2.0, 'K': -2.0, 'L': 1.0, 'M': 1.0, 'N': -3.0, 'O': -1.0, 'P': -2.0, 'Q': -2.0, 'R': -3.0, 'S': -2.0, 'T': 0.0, 'U': -1.0, 'V': 4.0, 'W': -3.0, 'X': -1.0, 'Y': -1.0, 'Z': -2.0},
                        'W': {'*': -4.0, 'A': -3.0, 'B': -4.0, 'C': -2.0, 'D': -4.0, 'E': -3.0, 'F': 1.0, 'G': -2.0, 'H': -2.0, 'I': -3.0, 'J': -2.0, 'K': -3.0, 'L': -2.0, 'M': -1.0, 'N': -4.0, 'O': -1.0, 'P': -4.0, 'Q': -2.0, 'R': -3.0, 'S': -3.0, 'T': -2.0, 'U': -2.0, 'V': -3.0, 'W': 11.0, 'X': -1.0, 'Y': 2.0, 'Z': -2.0},
                        'X': {'*': -4.0, 'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': -1.0, 'G': -1.0, 'H': -1.0, 'I': -1.0, 'J': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': -1.0, 'O': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'X': -1.0, 'Y': -1.0, 'Z': -1.0},
                        'Y': {'*': -4.0, 'A': -2.0, 'B': -3.0, 'C': -2.0, 'D': -3.0, 'E': -2.0, 'F': 3.0, 'G': -3.0, 'H': 2.0, 'I': -1.0, 'J': -1.0, 'K': -2.0, 'L': -1.0, 'M': -1.0, 'N': -2.0, 'O': -1.0, 'P': -3.0, 'Q': -1.0, 'R': -2.0, 'S': -2.0, 'T': -2.0, 'U': -2.0, 'V': -1.0, 'W': 2.0, 'X': -1.0, 'Y': 7.0, 'Z': -2.0},
                        'Z': {'*': -4.0, 'A': -1.0, 'B': 0.0, 'C': -3.0, 'D': 1.0, 'E': 4.0, 'F': -3.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'J': -3.0, 'K': 1.0, 'L': -3.0, 'M': -1.0, 'N': 0.0, 'O': -1.0, 'P': -1.0, 'Q': 4.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'U': -3.0, 'V': -2.0, 'W': -2.0, 'X': -1.0, 'Y': -2.0, 'Z': 4.0}
                },
                'DAYHOFF': {
                        'A': {'A': 0.2, 'C': -0.2, 'D': 0.0, 'E': 0.0, 'F': -0.4, 'G': 0.1, 'H': -0.1, 'I': -0.1, 'K': -0.1, 'L': -0.2, 'M': -0.1, 'N': 0.0, 'P': 0.1, 'Q': 0.0, 'R': -0.2, 'S': 0.1, 'T': 0.1, 'V': 0.0, 'W': -0.6, 'Y': -0.3},
                        'C': {'A': -0.2, 'C': 1.2, 'D': -0.5, 'E': -0.5, 'F': -0.4, 'G': -0.3, 'H': -0.3, 'I': -0.2, 'K': -0.5, 'L': -0.6, 'M': -0.5, 'N': -0.4, 'P': -0.3, 'Q': -0.5, 'R': -0.4, 'S': 0.0, 'T': -0.2, 'V': -0.2, 'W': -0.8, 'Y': 0.0},
                        'D': {'A': 0.0, 'C': -0.5, 'D': 0.4, 'E': 0.3, 'F': -0.6, 'G': 0.1, 'H': 0.1, 'I': -0.2, 'K': 0.0, 'L': -0.4, 'M': -0.3, 'N': 0.2, 'P': -0.1, 'Q': 0.2, 'R': -0.1, 'S': 0.0, 'T': 0.0, 'V': -0.2, 'W': -0.7, 'Y': -0.4},
                        'E': {'A': 0.0, 'C': -0.5, 'D': 0.3, 'E': 0.4, 'F': -0.5, 'G': 0.0, 'H': 0.1, 'I': -0.2, 'K': 0.0, 'L': -0.3, 'M': -0.2, 'N': 0.1, 'P': -0.1, 'Q': 0.2, 'R': -0.1, 'S': 0.0, 'T': 0.0, 'V': -0.2, 'W': -0.7, 'Y': -0.4},
                        'F': {'A': -0.4, 'C': -0.4, 'D': -0.6, 'E': -0.5, 'F': 0.9, 'G': -0.5, 'H': -0.2, 'I': 0.1, 'K': -0.5, 'L': 0.2, 'M': 0.0, 'N': -0.4, 'P': -0.5, 'Q': -0.5, 'R': -0.4, 'S': -0.3, 'T': -0.3, 'V': -0.1, 'W': 0.0, 'Y': 0.7},
                        'G': {'A': 0.1, 'C': -0.3, 'D': 0.1, 'E': 0.0, 'F': -0.5, 'G': 0.5, 'H': -0.2, 'I': -0.3, 'K': -0.2, 'L': -0.4, 'M': -0.3, 'N': 0.0, 'P': -0.1, 'Q': -0.1, 'R': -0.3, 'S': 0.1, 'T': 0.0, 'V': -0.1, 'W': -0.7, 'Y': -0.5},
                        'H': {'A': -0.1, 'C': -0.3, 'D': 0.1, 'E': 0.1, 'F': -0.2, 'G': -0.2, 'H': 0.6, 'I': -0.2, 'K': 0.0, 'L': -0.2, 'M': -0.2, 'N': 0.2, 'P': 0.0, 'Q': 0.3, 'R': 0.2, 'S': -0.1, 'T': -0.1, 'V': -0.2, 'W': -0.3, 'Y': 0.0},
                        'I': {'A': -0.1, 'C': -0.2, 'D': -0.2, 'E': -0.2, 'F': 0.1, 'G': -0.3, 'H': -0.2, 'I': 0.5, 'K': -0.2, 'L': 0.2, 'M': 0.2, 'N': -0.2, 'P': -0.2, 'Q': -0.2, 'R': -0.2, 'S': -0.1, 'T': 0.0, 'V': 0.4, 'W': -0.5, 'Y': -0.1},
                        'K': {'A': -0.1, 'C': -0.5, 'D': 0.0, 'E': 0.0, 'F': -0.5, 'G': -0.2, 'H': 0.0, 'I': -0.2, 'K': 0.5, 'L': -0.3, 'M': 0.0, 'N': 0.1, 'P': -0.1, 'Q': 0.1, 'R': 0.3, 'S': 0.0, 'T': 0.0, 'V': -0.2, 'W': -0.3, 'Y': -0.4},
                        'L': {'A': -0.2, 'C': -0.6, 'D': -0.4, 'E': -0.3, 'F': 0.2, 'G': -0.4, 'H': -0.2, 'I': 0.2, 'K': -0.3, 'L': 0.6, 'M': 0.4, 'N': -0.3, 'P': -0.3, 'Q': -0.2, 'R': -0.3, 'S': -0.3, 'T': -0.2, 'V': 0.2, 'W': -0.2, 'Y': -0.1},
                        'M': {'A': -0.1, 'C': -0.5, 'D': -0.3, 'E': -0.2, 'F': 0.0, 'G': -0.3, 'H': -0.2, 'I': 0.2, 'K': 0.0, 'L': 0.4, 'M': 0.6, 'N': -0.2, 'P': -0.2, 'Q': -0.1, 'R': 0.0, 'S': -0.2, 'T': -0.1, 'V': 0.2, 'W': -0.4, 'Y': -0.2},
                        'N': {'A': 0.0, 'C': -0.4, 'D': 0.2, 'E': 0.1, 'F': -0.4, 'G': 0.0, 'H': 0.2, 'I': -0.2, 'K': 0.1, 'L': -0.3, 'M': -0.2, 'N': 0.2, 'P': -0.1, 'Q': 0.1, 'R': 0.0, 'S': 0.1, 'T': 0.0, 'V': -0.2, 'W': -0.4, 'Y': -0.2},
                        'P': {'A': 0.1, 'C': -0.3, 'D': -0.1, 'E': -0.1, 'F': -0.5, 'G': -0.1, 'H': 0.0, 'I': -0.2, 'K': -0.1, 'L': -0.3, 'M': -0.2, 'N': -0.1, 'P': 0.6, 'Q': 0.0, 'R': 0.0, 'S': 0.1, 'T': 0.0, 'V': -0.1, 'W': -0.6, 'Y': -0.5},
                        'Q': {'A': 0.0, 'C': -0.5, 'D': 0.2, 'E': 0.2, 'F': -0.5, 'G': -0.1, 'H': 0.3, 'I': -0.2, 'K': 0.1, 'L': -0.2, 'M': -0.1, 'N': 0.1, 'P': 0.0, 'Q': 0.4, 'R': 0.1, 'S': -0.1, 'T': -0.1, 'V': -0.2, 'W': -0.5, 'Y': -0.4},
                        'R': {'A': -0.2, 'C': -0.4, 'D': -0.1, 'E': -0.1, 'F': -0.4, 'G': -0.3, 'H': 0.2, 'I': -0.2, 'K': 0.3, 'L': -0.3, 'M': 0.0, 'N': 0.0, 'P': 0.0, 'Q': 0.1, 'R': 0.6, 'S': 0.0, 'T': -0.1, 'V': -0.2, 'W': 0.2, 'Y': -0.4},
                        'S': {'A': 0.1, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -0.3, 'G': 0.1, 'H': -0.1, 'I': -0.1, 'K': 0.0, 'L': -0.3, 'M': -0.2, 'N': 0.1, 'P': 0.1, 'Q': -0.1, 'R': 0.0, 'S': 0.2, 'T': 0.1, 'V': -0.1, 'W': -0.2, 'Y': -0.3},
                        'T': {'A': 0.1, 'C': -0.2, 'D': 0.0, 'E': 0.0, 'F': -0.3, 'G': 0.0, 'H': -0.1, 'I': 0.0, 'K': 0.0, 'L': -0.2, 'M': -0.1, 'N': 0.0, 'P': 0.0, 'Q': -0.1, 'R': -0.1, 'S': 0.1, 'T': 0.3, 'V': 0.0, 'W': -0.5, 'Y': -0.3},
                        'V': {'A': 0.0, 'C': -0.2, 'D': -0.2, 'E': -0.2, 'F': -0.1, 'G': -0.1, 'H': -0.2, 'I': 0.4, 'K': -0.2, 'L': 0.2, 'M': 0.2, 'N': -0.2, 'P': -0.1, 'Q': -0.2, 'R': -0.2, 'S': -0.1, 'T': 0.0, 'V': 0.4, 'W': -0.6, 'Y': -0.2},
                        'W': {'A': -0.6, 'C': -0.8, 'D': -0.7, 'E': -0.7, 'F': 0.0, 'G': -0.7, 'H': -0.3, 'I': -0.5, 'K': -0.3, 'L': -0.2, 'M': -0.4, 'N': -0.4, 'P': -0.6, 'Q': -0.5, 'R': 0.2, 'S': -0.2, 'T': -0.5, 'V': -0.6, 'W': 1.7, 'Y': 0.0},
                        'Y': {'A': -0.3, 'C': 0.0, 'D': -0.4, 'E': -0.4, 'F': 0.7, 'G': -0.5, 'H': 0.0, 'I': -0.1, 'K': -0.4, 'L': -0.1, 'M': -0.2, 'N': -0.2, 'P': -0.5, 'Q': -0.4, 'R': -0.4, 'S': -0.3, 'T': -0.3, 'V': -0.2, 'W': 0.0, 'Y': 1.0}
                },
                'FENG': {
                        'A': {'A': 6.0, 'C': 2.0, 'D': 4.0, 'E': 4.0, 'F': 2.0, 'G': 5.0, 'H': 2.0, 'I': 2.0, 'K': 3.0, 'L': 2.0, 'M': 2.0, 'N': 3.0, 'P': 5.0, 'Q': 3.0, 'R': 2.0, 'S': 5.0, 'T': 5.0, 'V': 5.0, 'W': 2.0, 'Y': 2.0},
                        'C': {'A': 2.0, 'C': 6.0, 'D': 1.0, 'E': 0.0, 'F': 3.0, 'G': 3.0, 'H': 2.0, 'I': 2.0, 'K': 0.0, 'L': 2.0, 'M': 2.0, 'N': 2.0, 'P': 2.0, 'Q': 1.0, 'R': 2.0, 'S': 4.0, 'T': 2.0, 'V': 2.0, 'W': 3.0, 'Y': 3.0},
                        'D': {'A': 4.0, 'C': 1.0, 'D': 6.0, 'E': 5.0, 'F': 1.0, 'G': 4.0, 'H': 3.0, 'I': 1.0, 'K': 3.0, 'L': 1.0, 'M': 0.0, 'N': 5.0, 'P': 2.0, 'Q': 4.0, 'R': 2.0, 'S': 3.0, 'T': 2.0, 'V': 3.0, 'W': 0.0, 'Y': 2.0},
                        'E': {'A': 4.0, 'C': 0.0, 'D': 5.0, 'E': 6.0, 'F': 0.0, 'G': 4.0, 'H': 2.0, 'I': 1.0, 'K': 4.0, 'L': 1.0, 'M': 1.0, 'N': 3.0, 'P': 3.0, 'Q': 4.0, 'R': 2.0, 'S': 3.0, 'T': 3.0, 'V': 4.0, 'W': 1.0, 'Y': 1.0},
                        'F': {'A': 2.0, 'C': 3.0, 'D': 1.0, 'E': 0.0, 'F': 6.0, 'G': 1.0, 'H': 2.0, 'I': 4.0, 'K': 0.0, 'L': 4.0, 'M': 2.0, 'N': 1.0, 'P': 2.0, 'Q': 1.0, 'R': 1.0, 'S': 3.0, 'T': 1.0, 'V': 4.0, 'W': 3.0, 'Y': 5.0},
                        'G': {'A': 5.0, 'C': 3.0, 'D': 4.0, 'E': 4.0, 'F': 1.0, 'G': 6.0, 'H': 1.0, 'I': 2.0, 'K': 2.0, 'L': 2.0, 'M': 1.0, 'N': 3.0, 'P': 3.0, 'Q': 2.0, 'R': 3.0, 'S': 5.0, 'T': 2.0, 'V': 4.0, 'W': 3.0, 'Y': 2.0},
                        'H': {'A': 2.0, 'C': 2.0, 'D': 3.0, 'E': 2.0, 'F': 2.0, 'G': 1.0, 'H': 6.0, 'I': 1.0, 'K': 3.0, 'L': 3.0, 'M': 1.0, 'N': 4.0, 'P': 3.0, 'Q': 4.0, 'R': 4.0, 'S': 3.0, 'T': 2.0, 'V': 1.0, 'W': 1.0, 'Y': 3.0},
                        'I': {'A': 2.0, 'C': 2.0, 'D': 1.0, 'E': 1.0, 'F': 4.0, 'G': 2.0, 'H': 1.0, 'I': 6.0, 'K': 2.0, 'L': 5.0, 'M': 4.0, 'N': 2.0, 'P': 2.0, 'Q': 1.0, 'R': 2.0, 'S': 2.0, 'T': 3.0, 'V': 5.0, 'W': 2.0, 'Y': 3.0},
                        'K': {'A': 3.0, 'C': 0.0, 'D': 3.0, 'E': 4.0, 'F': 0.0, 'G': 2.0, 'H': 3.0, 'I': 2.0, 'K': 6.0, 'L': 2.0, 'M': 2.0, 'N': 4.0, 'P': 2.0, 'Q': 4.0, 'R': 5.0, 'S': 3.0, 'T': 4.0, 'V': 3.0, 'W': 1.0, 'Y': 1.0},
                        'L': {'A': 2.0, 'C': 2.0, 'D': 1.0, 'E': 1.0, 'F': 4.0, 'G': 2.0, 'H': 3.0, 'I': 5.0, 'K': 2.0, 'L': 6.0, 'M': 5.0, 'N': 1.0, 'P': 3.0, 'Q': 2.0, 'R': 2.0, 'S': 2.0, 'T': 2.0, 'V': 5.0, 'W': 4.0, 'Y': 3.0},
                        'M': {'A': 2.0, 'C': 2.0, 'D': 0.0, 'E': 1.0, 'F': 2.0, 'G': 1.0, 'H': 1.0, 'I': 4.0, 'K': 2.0, 'L': 5.0, 'M': 6.0, 'N': 1.0, 'P': 2.0, 'Q': 2.0, 'R': 2.0, 'S': 1.0, 'T': 3.0, 'V': 4.0, 'W': 3.0, 'Y': 2.0},
                        'N': {'A': 3.0, 'C': 2.0, 'D': 5.0, 'E': 3.0, 'F': 1.0, 'G': 3.0, 'H': 4.0, 'I': 2.0, 'K': 4.0, 'L': 1.0, 'M': 1.0, 'N': 6.0, 'P': 2.0, 'Q': 3.0, 'R': 2.0, 'S': 5.0, 'T': 4.0, 'V': 2.0, 'W': 0.0, 'Y': 3.0},
                        'P': {'A': 5.0, 'C': 2.0, 'D': 2.0, 'E': 3.0, 'F': 2.0, 'G': 3.0, 'H': 3.0, 'I': 2.0, 'K': 2.0, 'L': 3.0, 'M': 2.0, 'N': 2.0, 'P': 6.0, 'Q': 3.0, 'R': 3.0, 'S': 4.0, 'T': 4.0, 'V': 3.0, 'W': 2.0, 'Y': 2.0},
                        'Q': {'A': 3.0, 'C': 1.0, 'D': 4.0, 'E': 4.0, 'F': 1.0, 'G': 2.0, 'H': 4.0, 'I': 1.0, 'K': 4.0, 'L': 2.0, 'M': 2.0, 'N': 3.0, 'P': 3.0, 'Q': 6.0, 'R': 3.0, 'S': 3.0, 'T': 3.0, 'V': 2.0, 'W': 1.0, 'Y': 2.0},
                        'R': {'A': 2.0, 'C': 2.0, 'D': 2.0, 'E': 2.0, 'F': 1.0, 'G': 3.0, 'H': 4.0, 'I': 2.0, 'K': 5.0, 'L': 2.0, 'M': 2.0, 'N': 2.0, 'P': 3.0, 'Q': 3.0, 'R': 6.0, 'S': 3.0, 'T': 3.0, 'V': 2.0, 'W': 2.0, 'Y': 1.0},
                        'S': {'A': 5.0, 'C': 4.0, 'D': 3.0, 'E': 3.0, 'F': 3.0, 'G': 5.0, 'H': 3.0, 'I': 2.0, 'K': 3.0, 'L': 2.0, 'M': 1.0, 'N': 5.0, 'P': 4.0, 'Q': 3.0, 'R': 3.0, 'S': 6.0, 'T': 5.0, 'V': 2.0, 'W': 2.0, 'Y': 3.0},
                        'T': {'A': 5.0, 'C': 2.0, 'D': 2.0, 'E': 3.0, 'F': 1.0, 'G': 2.0, 'H': 2.0, 'I': 3.0, 'K': 4.0, 'L': 2.0, 'M': 3.0, 'N': 4.0, 'P': 4.0, 'Q': 3.0, 'R': 3.0, 'S': 5.0, 'T': 6.0, 'V': 3.0, 'W': 1.0, 'Y': 2.0},
                        'V': {'A': 5.0, 'C': 2.0, 'D': 3.0, 'E': 4.0, 'F': 4.0, 'G': 4.0, 'H': 1.0, 'I': 5.0, 'K': 3.0, 'L': 5.0, 'M': 4.0, 'N': 2.0, 'P': 3.0, 'Q': 2.0, 'R': 2.0, 'S': 2.0, 'T': 3.0, 'V': 6.0, 'W': 3.0, 'Y': 3.0},
                        'W': {'A': 2.0, 'C': 3.0, 'D': 0.0, 'E': 1.0, 'F': 3.0, 'G': 3.0, 'H': 1.0, 'I': 2.0, 'K': 1.0, 'L': 4.0, 'M': 3.0, 'N': 0.0, 'P': 2.0, 'Q': 1.0, 'R': 2.0, 'S': 2.0, 'T': 1.0, 'V': 3.0, 'W': 6.0, 'Y': 3.0},
                        'Y': {'A': 2.0, 'C': 3.0, 'D': 2.0, 'E': 1.0, 'F': 5.0, 'G': 2.0, 'H': 3.0, 'I': 3.0, 'K': 1.0, 'L': 3.0, 'M': 2.0, 'N': 3.0, 'P': 2.0, 'Q': 2.0, 'R': 1.0, 'S': 3.0, 'T': 2.0, 'V': 3.0, 'W': 3.0, 'Y': 6.0}
                },
                'GENETIC': {
                        'A': {'A': 4.0, 'C': -1.9, 'D': 1.0, 'E': 1.3, 'F': -2.4, 'G': 1.2, 'H': -2.1, 'I': -1.8, 'K': -1.9, 'L': -2.3, 'M': -2.0, 'N': -1.7, 'P': 0.8, 'Q': -2.1, 'R': -1.6, 'S': 0.1, 'T': 0.9, 'V': 1.0, 'W': -2.2, 'Y': -2.4},
                        'C': {'A': -1.9, 'C': 5.5, 'D': -1.6, 'E': -3.0, 'F': 1.8, 'G': 1.0, 'H': -1.6, 'I': -1.9, 'K': -3.2, 'L': -1.3, 'M': -2.7, 'N': -1.5, 'P': -1.9, 'Q': -3.1, 'R': 0.7, 'S': 1.5, 'T': -1.9, 'V': -2.2, 'W': 4.1, 'Y': 2.6},
                        'D': {'A': 1.0, 'C': -1.6, 'D': 4.8, 'E': 3.8, 'F': -1.7, 'G': 1.1, 'H': 1.7, 'I': -2.1, 'K': 0.3, 'L': -2.4, 'M': -2.5, 'N': 1.7, 'P': -2.2, 'Q': 0.3, 'R': -2.3, 'S': -2.1, 'T': -2.1, 'V': 1.0, 'W': -2.9, 'Y': 2.3},
                        'E': {'A': 1.3, 'C': -3.0, 'D': 3.8, 'E': 5.7, 'F': -2.9, 'G': 1.4, 'H': 0.3, 'I': -2.3, 'K': 2.0, 'L': -2.5, 'M': -1.8, 'N': 0.3, 'P': -2.1, 'Q': 2.0, 'R': -2.0, 'S': -2.8, 'T': -2.1, 'V': 1.3, 'W': -3.2, 'Y': -0.9},
                        'F': {'A': -2.4, 'C': 1.8, 'D': -1.7, 'E': -2.9, 'F': 4.5, 'G': -1.9, 'H': -1.1, 'I': 1.3, 'K': -2.8, 'L': 2.2, 'M': 0.5, 'N': -1.3, 'P': -1.8, 'Q': -2.1, 'R': -1.5, 'S': 0.0, 'T': -2.1, 'V': 1.0, 'W': 0.0, 'Y': 2.0},
                        'G': {'A': 1.2, 'C': 1.0, 'D': 1.1, 'E': 1.4, 'F': -1.9, 'G': 4.2, 'H': -2.2, 'I': -2.5, 'K': -2.2, 'L': -2.2, 'M': -2.3, 'N': -2.6, 'P': -1.8, 'Q': -2.1, 'R': 0.8, 'S': -0.6, 'T': -2.1, 'V': 1.1, 'W': 1.4, 'Y': -1.8},
                        'H': {'A': -2.1, 'C': -1.6, 'D': 1.7, 'E': 0.3, 'F': -1.1, 'G': -2.2, 'H': 4.7, 'I': -1.8, 'K': 0.6, 'L': -0.1, 'M': -1.8, 'N': 1.8, 'P': 0.7, 'Q': 3.6, 'R': 3.6, 'S': -1.6, 'T': -1.8, 'V': -2.1, 'W': -2.1, 'Y': 2.3},
                        'I': {'A': -1.8, 'C': -1.9, 'D': -2.1, 'E': -2.3, 'F': 1.3, 'G': -2.5, 'H': -1.8, 'I': 4.1, 'K': 0.7, 'L': 1.2, 'M': 3.3, 'N': 0.9, 'P': -1.6, 'Q': -1.9, 'R': -1.2, 'S': -0.5, 'T': 0.8, 'V': 1.0, 'W': -2.2, 'Y': -1.6},
                        'K': {'A': -1.9, 'C': -3.2, 'D': 0.3, 'E': 2.0, 'F': -2.8, 'G': -2.2, 'H': 0.6, 'I': 0.7, 'K': 5.6, 'L': -2.0, 'M': 1.6, 'N': 3.5, 'P': -1.5, 'Q': 2.2, 'R': -0.2, 'S': -1.5, 'T': 1.0, 'V': -2.1, 'W': -3.0, 'Y': -0.8},
                        'L': {'A': -2.3, 'C': -1.3, 'D': -2.4, 'E': -2.5, 'F': 2.2, 'G': -2.2, 'H': -0.1, 'I': 1.2, 'K': -2.0, 'L': 3.4, 'M': 1.5, 'N': -2.2, 'P': 0.0, 'Q': 0.1, 'R': -0.4, 'S': -1.2, 'T': -1.9, 'V': 1.1, 'W': -0.3, 'Y': -1.6},
                        'M': {'A': -2.0, 'C': -2.7, 'D': -2.5, 'E': -1.8, 'F': 0.5, 'G': -2.3, 'H': -1.8, 'I': 3.3, 'K': 1.6, 'L': 1.5, 'M': 5.4, 'N': 0.1, 'P': -1.4, 'Q': -1.2, 'R': -0.4, 'S': -1.3, 'T': 0.7, 'V': 1.0, 'W': -2.0, 'Y': -2.9},
                        'N': {'A': -1.7, 'C': -1.5, 'D': 1.7, 'E': 0.3, 'F': -1.3, 'G': -2.6, 'H': 1.8, 'I': 0.9, 'K': 3.5, 'L': -2.2, 'M': 0.1, 'N': 4.7, 'P': -1.6, 'Q': 0.4, 'R': -1.5, 'S': -0.3, 'T': 0.9, 'V': -2.2, 'W': -3.0, 'Y': 2.5},
                        'P': {'A': 0.8, 'C': -1.9, 'D': -2.2, 'E': -2.1, 'F': -1.8, 'G': -1.8, 'H': 0.7, 'I': -1.6, 'K': -1.5, 'L': 0.0, 'M': -1.4, 'N': -1.6, 'P': 3.8, 'Q': 1.0, 'R': 0.3, 'S': 0.4, 'T': 1.1, 'V': -2.1, 'W': -1.6, 'Y': -2.3},
                        'Q': {'A': -2.1, 'C': -3.1, 'D': 0.3, 'E': 2.0, 'F': -2.1, 'G': -2.1, 'H': 3.6, 'I': -1.9, 'K': 2.2, 'L': 0.1, 'M': -1.2, 'N': 0.4, 'P': 1.0, 'Q': 5.5, 'R': 0.3, 'S': -2.3, 'T': -1.7, 'V': -2.0, 'W': -2.3, 'Y': -0.8},
                        'R': {'A': -1.6, 'C': 0.7, 'D': -2.3, 'E': -2.0, 'F': -1.5, 'G': 0.8, 'H': 3.6, 'I': -1.2, 'K': -0.2, 'L': -0.4, 'M': -0.4, 'N': -1.5, 'P': 0.3, 'Q': 0.3, 'R': 2.9, 'S': 0.3, 'T': -0.6, 'V': -2.1, 'W': 1.8, 'Y': -1.9},
                        'S': {'A': 0.1, 'C': 1.5, 'D': -2.1, 'E': -2.8, 'F': 0.0, 'G': -0.6, 'H': -1.6, 'I': -0.5, 'K': -1.5, 'L': -1.2, 'M': -1.3, 'N': -0.3, 'P': 0.4, 'Q': -2.3, 'R': 0.3, 'S': 2.6, 'T': 1.0, 'V': -2.2, 'W': 0.8, 'Y': 0.3},
                        'T': {'A': 0.9, 'C': -1.9, 'D': -2.1, 'E': -2.1, 'F': -2.1, 'G': -2.1, 'H': -1.8, 'I': 0.8, 'K': 1.0, 'L': -1.9, 'M': 0.7, 'N': 0.9, 'P': 1.1, 'Q': -1.7, 'R': -0.6, 'S': 1.0, 'T': 4.0, 'V': -2.2, 'W': -2.2, 'Y': -2.1},
                        'V': {'A': 1.0, 'C': -2.2, 'D': 1.0, 'E': 1.3, 'F': 1.0, 'G': 1.1, 'H': -2.1, 'I': 1.0, 'K': -2.1, 'L': 1.1, 'M': 1.0, 'N': -2.2, 'P': -2.1, 'Q': -2.0, 'R': -2.1, 'S': -2.2, 'T': -2.2, 'V': 4.1, 'W': -2.1, 'Y': -2.2},
                        'W': {'A': -2.2, 'C': 4.1, 'D': -2.9, 'E': -3.2, 'F': 0.0, 'G': 1.4, 'H': -2.1, 'I': -2.2, 'K': -3.0, 'L': -0.3, 'M': -2.0, 'N': -3.0, 'P': -1.6, 'Q': -2.3, 'R': 1.8, 'S': 0.8, 'T': -2.2, 'V': -2.1, 'W': 7.5, 'Y': -0.5},
                        'Y': {'A': -2.4, 'C': 2.6, 'D': 2.3, 'E': -0.9, 'F': 2.0, 'G': -1.8, 'H': 2.3, 'I': -1.6, 'K': -0.8, 'L': -1.6, 'M': -2.9, 'N': 2.5, 'P': -2.3, 'Q': -0.8, 'R': -1.9, 'S': 0.3, 'T': -2.1, 'V': -2.2, 'W': -0.5, 'Y': 6.5}
                },
                'GONNET1992': {
                        'A': {'A': 2.4, 'C': 0.5, 'D': -0.3, 'E': 0.0, 'F': -2.3, 'G': 0.5, 'H': -0.8, 'I': -0.8, 'K': -0.4, 'L': -1.2, 'M': -0.7, 'N': -0.3, 'P': 0.3, 'Q': -0.2, 'R': -0.6, 'S': 1.1, 'T': 0.6, 'V': 0.1, 'W': -3.6, 'Y': -2.2},
                        'C': {'A': 0.5, 'C': 11.5, 'D': -3.2, 'E': -3.0, 'F': -0.8, 'G': -2.0, 'H': -1.3, 'I': -1.1, 'K': -2.8, 'L': -1.5, 'M': -0.9, 'N': -1.8, 'P': -3.1, 'Q': -2.4, 'R': -2.2, 'S': 0.1, 'T': -0.5, 'V': 0.0, 'W': -1.0, 'Y': -0.5},
                        'D': {'A': -0.3, 'C': -3.2, 'D': 4.7, 'E': 2.7, 'F': -4.5, 'G': 0.1, 'H': 0.4, 'I': -3.8, 'K': 0.5, 'L': -4.0, 'M': -3.0, 'N': 2.2, 'P': -0.7, 'Q': 0.9, 'R': -0.3, 'S': 0.5, 'T': 0.0, 'V': -2.9, 'W': -5.2, 'Y': -2.8},
                        'E': {'A': 0.0, 'C': -3.0, 'D': 2.7, 'E': 3.6, 'F': -3.9, 'G': -0.8, 'H': 0.4, 'I': -2.7, 'K': 1.2, 'L': -2.8, 'M': -2.0, 'N': 0.9, 'P': -0.5, 'Q': 1.7, 'R': 0.4, 'S': 0.2, 'T': -0.1, 'V': -1.9, 'W': -4.3, 'Y': -2.7},
                        'F': {'A': -2.3, 'C': -0.8, 'D': -4.5, 'E': -3.9, 'F': 7.0, 'G': -5.2, 'H': -0.1, 'I': 1.0, 'K': -3.3, 'L': 2.0, 'M': 1.6, 'N': -3.1, 'P': -3.8, 'Q': -2.6, 'R': -3.2, 'S': -2.8, 'T': -2.2, 'V': 0.1, 'W': 3.6, 'Y': 5.1},
                        'G': {'A': 0.5, 'C': -2.0, 'D': 0.1, 'E': -0.8, 'F': -5.2, 'G': 6.6, 'H': -1.4, 'I': -4.5, 'K': -1.1, 'L': -4.4, 'M': -3.5, 'N': 0.4, 'P': -1.6, 'Q': -1.0, 'R': -1.0, 'S': 0.4, 'T': -1.1, 'V': -3.3, 'W': -4.0, 'Y': -4.0},
                        'H': {'A': -0.8, 'C': -1.3, 'D': 0.4, 'E': 0.4, 'F': -0.1, 'G': -1.4, 'H': 6.0, 'I': -2.2, 'K': 0.6, 'L': -1.9, 'M': -1.3, 'N': 1.2, 'P': -1.1, 'Q': 1.2, 'R': 0.6, 'S': -0.2, 'T': -0.3, 'V': -2.0, 'W': -0.8, 'Y': 2.2},
                        'I': {'A': -0.8, 'C': -1.1, 'D': -3.8, 'E': -2.7, 'F': 1.0, 'G': -4.5, 'H': -2.2, 'I': 4.0, 'K': -2.1, 'L': 2.8, 'M': 2.5, 'N': -2.8, 'P': -2.6, 'Q': -1.9, 'R': -2.4, 'S': -1.8, 'T': -0.6, 'V': 3.1, 'W': -1.8, 'Y': -0.7},
                        'K': {'A': -0.4, 'C': -2.8, 'D': 0.5, 'E': 1.2, 'F': -3.3, 'G': -1.1, 'H': 0.6, 'I': -2.1, 'K': 3.2, 'L': -2.1, 'M': -1.4, 'N': 0.8, 'P': -0.6, 'Q': 1.5, 'R': 2.7, 'S': 0.1, 'T': 0.1, 'V': -1.7, 'W': -3.5, 'Y': -2.1},
                        'L': {'A': -1.2, 'C': -1.5, 'D': -4.0, 'E': -2.8, 'F': 2.0, 'G': -4.4, 'H': -1.9, 'I': 2.8, 'K': -2.1, 'L': 4.0, 'M': 2.8, 'N': -3.0, 'P': -2.3, 'Q': -1.6, 'R': -2.2, 'S': -2.1, 'T': -1.3, 'V': 1.8, 'W': -0.7, 'Y': 0.0},
                        'M': {'A': -0.7, 'C': -0.9, 'D': -3.0, 'E': -2.0, 'F': 1.6, 'G': -3.5, 'H': -1.3, 'I': 2.5, 'K': -1.4, 'L': 2.8, 'M': 4.3, 'N': -2.2, 'P': -2.4, 'Q': -1.0, 'R': -1.7, 'S': -1.4, 'T': -0.6, 'V': 1.6, 'W': -1.0, 'Y': -0.2},
                        'N': {'A': -0.3, 'C': -1.8, 'D': 2.2, 'E': 0.9, 'F': -3.1, 'G': 0.4, 'H': 1.2, 'I': -2.8, 'K': 0.8, 'L': -3.0, 'M': -2.2, 'N': 3.8, 'P': -0.9, 'Q': 0.7, 'R': 0.3, 'S': 0.9, 'T': 0.5, 'V': -2.2, 'W': -3.6, 'Y': -1.4},
                        'P': {'A': 0.3, 'C': -3.1, 'D': -0.7, 'E': -0.5, 'F': -3.8, 'G': -1.6, 'H': -1.1, 'I': -2.6, 'K': -0.6, 'L': -2.3, 'M': -2.4, 'N': -0.9, 'P': 7.6, 'Q': -0.2, 'R': -0.9, 'S': 0.4, 'T': 0.1, 'V': -1.8, 'W': -5.0, 'Y': -3.1},
                        'Q': {'A': -0.2, 'C': -2.4, 'D': 0.9, 'E': 1.7, 'F': -2.6, 'G': -1.0, 'H': 1.2, 'I': -1.9, 'K': 1.5, 'L': -1.6, 'M': -1.0, 'N': 0.7, 'P': -0.2, 'Q': 2.7, 'R': 1.5, 'S': 0.2, 'T': 0.0, 'V': -1.5, 'W': -2.7, 'Y': -1.7},
                        'R': {'A': -0.6, 'C': -2.2, 'D': -0.3, 'E': 0.4, 'F': -3.2, 'G': -1.0, 'H': 0.6, 'I': -2.4, 'K': 2.7, 'L': -2.2, 'M': -1.7, 'N': 0.3, 'P': -0.9, 'Q': 1.5, 'R': 4.7, 'S': -0.2, 'T': -0.2, 'V': -2.0, 'W': -1.6, 'Y': -1.8},
                        'S': {'A': 1.1, 'C': 0.1, 'D': 0.5, 'E': 0.2, 'F': -2.8, 'G': 0.4, 'H': -0.2, 'I': -1.8, 'K': 0.1, 'L': -2.1, 'M': -1.4, 'N': 0.9, 'P': 0.4, 'Q': 0.2, 'R': -0.2, 'S': 2.2, 'T': 1.5, 'V': -1.0, 'W': -3.3, 'Y': -1.9},
                        'T': {'A': 0.6, 'C': -0.5, 'D': 0.0, 'E': -0.1, 'F': -2.2, 'G': -1.1, 'H': -0.3, 'I': -0.6, 'K': 0.1, 'L': -1.3, 'M': -0.6, 'N': 0.5, 'P': 0.1, 'Q': 0.0, 'R': -0.2, 'S': 1.5, 'T': 2.5, 'V': 0.0, 'W': -3.5, 'Y': -1.9},
                        'V': {'A': 0.1, 'C': 0.0, 'D': -2.9, 'E': -1.9, 'F': 0.1, 'G': -3.3, 'H': -2.0, 'I': 3.1, 'K': -1.7, 'L': 1.8, 'M': 1.6, 'N': -2.2, 'P': -1.8, 'Q': -1.5, 'R': -2.0, 'S': -1.0, 'T': 0.0, 'V': 3.4, 'W': -2.6, 'Y': -1.1},
                        'W': {'A': -3.6, 'C': -1.0, 'D': -5.2, 'E': -4.3, 'F': 3.6, 'G': -4.0, 'H': -0.8, 'I': -1.8, 'K': -3.5, 'L': -0.7, 'M': -1.0, 'N': -3.6, 'P': -5.0, 'Q': -2.7, 'R': -1.6, 'S': -3.3, 'T': -3.5, 'V': -2.6, 'W': 14.2, 'Y': 4.1},
                        'Y': {'A': -2.2, 'C': -0.5, 'D': -2.8, 'E': -2.7, 'F': 5.1, 'G': -4.0, 'H': 2.2, 'I': -0.7, 'K': -2.1, 'L': 0.0, 'M': -0.2, 'N': -1.4, 'P': -3.1, 'Q': -1.7, 'R': -1.8, 'S': -1.9, 'T': -1.9, 'V': -1.1, 'W': 4.1, 'Y': 7.8}
                },
                'JOHNSON': {
                        'A': {'A': 0.6, 'C': -0.34, 'D': -0.16, 'E': -0.07, 'F': -0.32, 'G': -0.05, 'H': -0.31, 'I': -0.22, 'K': -0.09, 'L': -0.33, 'M': -0.15, 'N': -0.14, 'P': -0.1, 'Q': -0.06, 'R': -0.16, 'S': 0.0, 'T': -0.08, 'V': -0.05, 'W': -0.58, 'Y': -0.4},
                        'C': {'A': -0.34, 'C': 1.61, 'D': -0.97, 'E': -0.69, 'F': -0.44, 'G': -0.82, 'H': -0.82, 'I': -0.77, 'K': -0.87, 'L': -0.87, 'M': -0.44, 'N': -0.76, 'P': -0.89, 'Q': -0.69, 'R': -0.56, 'S': -0.77, 'T': -0.6, 'V': -0.48, 'W': -0.91, 'Y': -0.77},
                        'D': {'A': -0.16, 'C': -0.97, 'D': 0.85, 'E': 0.24, 'F': -0.7, 'G': -0.21, 'H': -0.07, 'I': -0.48, 'K': -0.15, 'L': -0.8, 'M': -0.59, 'N': 0.26, 'P': -0.1, 'Q': -0.11, 'R': -0.34, 'S': -0.02, 'T': -0.18, 'V': -0.52, 'W': -0.6, 'Y': -0.38},
                        'E': {'A': -0.07, 'C': -0.69, 'D': 0.24, 'E': 0.86, 'F': -0.64, 'G': -0.25, 'H': -0.23, 'I': -0.48, 'K': 0.11, 'L': -0.56, 'M': -0.28, 'N': -0.07, 'P': -0.15, 'Q': 0.24, 'R': -0.02, 'S': -0.22, 'T': -0.05, 'V': -0.42, 'W': -0.76, 'Y': -0.37},
                        'F': {'A': -0.32, 'C': -0.44, 'D': -0.7, 'E': -0.64, 'F': 1.04, 'G': -0.86, 'H': -0.17, 'I': 0.05, 'K': -0.56, 'L': 0.18, 'M': -0.06, 'N': -0.38, 'P': -0.5, 'Q': -0.64, 'R': -0.6, 'S': -0.48, 'T': -0.5, 'V': -0.13, 'W': 0.34, 'Y': 0.34},
                        'G': {'A': -0.05, 'C': -0.82, 'D': -0.21, 'E': -0.25, 'F': -0.86, 'G': 0.8, 'H': -0.32, 'I': -0.55, 'K': -0.35, 'L': -0.72, 'M': -0.52, 'N': -0.14, 'P': -0.25, 'Q': -0.28, 'R': -0.28, 'S': -0.13, 'T': -0.38, 'V': -0.56, 'W': -0.63, 'Y': -0.54},
                        'H': {'A': -0.31, 'C': -0.82, 'D': -0.07, 'E': -0.23, 'F': -0.17, 'G': -0.32, 'H': 1.27, 'I': -0.51, 'K': 0.01, 'L': -0.42, 'M': -0.23, 'N': 0.17, 'P': -0.43, 'Q': 0.14, 'R': 0.01, 'S': -0.26, 'T': -0.3, 'V': -0.39, 'W': -0.4, 'Y': -0.04},
                        'I': {'A': -0.22, 'C': -0.77, 'D': -0.48, 'E': -0.48, 'F': 0.05, 'G': -0.55, 'H': -0.51, 'I': 0.81, 'K': -0.47, 'L': 0.26, 'M': 0.26, 'N': -0.47, 'P': -0.57, 'Q': -0.7, 'R': -0.54, 'S': -0.47, 'T': -0.32, 'V': 0.39, 'W': -0.33, 'Y': -0.25},
                        'K': {'A': -0.09, 'C': -0.87, 'D': -0.15, 'E': 0.11, 'F': -0.56, 'G': -0.35, 'H': 0.01, 'I': -0.47, 'K': 0.76, 'L': -0.34, 'M': -0.19, 'N': 0.01, 'P': -0.06, 'Q': 0.11, 'R': 0.32, 'S': -0.15, 'T': -0.02, 'V': -0.37, 'W': -0.54, 'Y': -0.37},
                        'L': {'A': -0.33, 'C': -0.87, 'D': -0.8, 'E': -0.56, 'F': 0.18, 'G': -0.72, 'H': -0.42, 'I': 0.26, 'K': -0.34, 'L': 0.73, 'M': 0.44, 'N': -0.48, 'P': -0.28, 'Q': -0.44, 'R': -0.37, 'S': -0.52, 'T': -0.46, 'V': 0.18, 'W': -0.1, 'Y': -0.24},
                        'M': {'A': -0.15, 'C': -0.44, 'D': -0.59, 'E': -0.28, 'F': -0.06, 'G': -0.52, 'H': -0.23, 'I': 0.26, 'K': -0.19, 'L': 0.44, 'M': 1.12, 'N': -0.37, 'P': -0.98, 'Q': -0.06, 'R': -0.42, 'S': -0.48, 'T': -0.32, 'V': 0.07, 'W': -0.09, 'Y': -0.13},
                        'N': {'A': -0.14, 'C': -0.76, 'D': 0.26, 'E': -0.07, 'F': -0.38, 'G': -0.14, 'H': 0.17, 'I': -0.47, 'K': 0.01, 'L': -0.48, 'M': -0.37, 'N': 0.8, 'P': -0.24, 'Q': -0.08, 'R': -0.15, 'S': 0.1, 'T': 0.01, 'V': -0.57, 'W': -0.61, 'Y': -0.13},
                        'P': {'A': -0.1, 'C': -0.89, 'D': -0.1, 'E': -0.15, 'F': -0.5, 'G': -0.25, 'H': -0.43, 'I': -0.57, 'K': -0.06, 'L': -0.28, 'M': -0.98, 'N': -0.24, 'P': 1.03, 'Q': -0.36, 'R': -0.36, 'S': -0.1, 'T': -0.2, 'V': -0.52, 'W': -0.74, 'Y': -0.7},
                        'Q': {'A': -0.06, 'C': -0.69, 'D': -0.11, 'E': 0.24, 'F': -0.64, 'G': -0.28, 'H': 0.14, 'I': -0.7, 'K': 0.11, 'L': -0.44, 'M': -0.06, 'N': -0.08, 'P': -0.36, 'Q': 0.9, 'R': 0.21, 'S': -0.12, 'T': -0.04, 'V': -0.36, 'W': -0.82, 'Y': -0.51},
                        'R': {'A': -0.16, 'C': -0.56, 'D': -0.34, 'E': -0.02, 'F': -0.6, 'G': -0.28, 'H': 0.01, 'I': -0.54, 'K': 0.32, 'L': -0.37, 'M': -0.42, 'N': -0.15, 'P': -0.36, 'Q': 0.21, 'R': 1.0, 'S': -0.06, 'T': -0.14, 'V': -0.49, 'W': -0.38, 'Y': -0.21},
                        'S': {'A': 0.0, 'C': -0.77, 'D': -0.02, 'E': -0.22, 'F': -0.48, 'G': -0.13, 'H': -0.26, 'I': -0.47, 'K': -0.15, 'L': -0.52, 'M': -0.48, 'N': 0.1, 'P': -0.1, 'Q': -0.12, 'R': -0.06, 'S': 0.58, 'T': 0.2, 'V': -0.43, 'W': -0.62, 'Y': -0.34},
                        'T': {'A': -0.08, 'C': -0.6, 'D': -0.18, 'E': -0.05, 'F': -0.5, 'G': -0.38, 'H': -0.3, 'I': -0.32, 'K': -0.02, 'L': -0.46, 'M': -0.32, 'N': 0.01, 'P': -0.2, 'Q': -0.04, 'R': -0.14, 'S': 0.2, 'T': 0.68, 'V': -0.19, 'W': -0.93, 'Y': -0.27},
                        'V': {'A': -0.05, 'C': -0.48, 'D': -0.52, 'E': -0.42, 'F': -0.13, 'G': -0.56, 'H': -0.39, 'I': 0.39, 'K': -0.37, 'L': 0.18, 'M': 0.07, 'N': -0.57, 'P': -0.52, 'Q': -0.36, 'R': -0.49, 'S': -0.43, 'T': -0.19, 'V': 0.7, 'W': -0.49, 'Y': -0.18},
                        'W': {'A': -0.58, 'C': -0.91, 'D': -0.6, 'E': -0.76, 'F': 0.34, 'G': -0.63, 'H': -0.4, 'I': -0.33, 'K': -0.54, 'L': -0.1, 'M': -0.09, 'N': -0.61, 'P': -0.74, 'Q': -0.82, 'R': -0.38, 'S': -0.62, 'T': -0.93, 'V': -0.49, 'W': 1.52, 'Y': 0.23},
                        'Y': {'A': -0.4, 'C': -0.77, 'D': -0.38, 'E': -0.37, 'F': 0.34, 'G': -0.54, 'H': -0.04, 'I': -0.25, 'K': -0.37, 'L': -0.24, 'M': -0.13, 'N': -0.13, 'P': -0.7, 'Q': -0.51, 'R': -0.21, 'S': -0.34, 'T': -0.27, 'V': -0.18, 'W': 0.23, 'Y': 1.05}
                },
                'JONES': {
                        'A': {'A': 0.2, 'C': -0.1, 'D': 0.0, 'E': -0.1, 'F': -0.3, 'G': 0.1, 'H': -0.2, 'I': 0.0, 'K': -0.1, 'L': -0.1, 'M': -0.1, 'N': 0.0, 'P': 0.1, 'Q': -0.1, 'R': -0.1, 'S': 0.1, 'T': 0.2, 'V': 0.1, 'W': -0.4, 'Y': -0.3},
                        'C': {'A': -0.1, 'C': 1.1, 'D': -0.3, 'E': -0.4, 'F': 0.0, 'G': -0.1, 'H': 0.0, 'I': -0.2, 'K': -0.3, 'L': -0.3, 'M': -0.2, 'N': -0.1, 'P': -0.2, 'Q': -0.3, 'R': -0.1, 'S': 0.1, 'T': -0.1, 'V': -0.2, 'W': 0.1, 'Y': 0.2},
                        'D': {'A': 0.0, 'C': -0.3, 'D': 0.5, 'E': 0.4, 'F': -0.5, 'G': 0.1, 'H': 0.0, 'I': -0.3, 'K': 0.0, 'L': -0.4, 'M': -0.3, 'N': 0.2, 'P': -0.2, 'Q': 0.1, 'R': -0.1, 'S': 0.0, 'T': -0.1, 'V': -0.2, 'W': -0.5, 'Y': -0.2},
                        'E': {'A': -0.1, 'C': -0.4, 'D': 0.4, 'E': 0.5, 'F': -0.5, 'G': 0.0, 'H': 0.0, 'I': -0.3, 'K': 0.1, 'L': -0.4, 'M': -0.3, 'N': 0.1, 'P': -0.2, 'Q': 0.2, 'R': 0.0, 'S': -0.1, 'T': -0.1, 'V': -0.2, 'W': -0.5, 'Y': -0.4},
                        'F': {'A': -0.3, 'C': 0.0, 'D': -0.5, 'E': -0.5, 'F': 0.8, 'G': -0.5, 'H': 0.0, 'I': 0.0, 'K': -0.5, 'L': 0.2, 'M': 0.0, 'N': -0.3, 'P': -0.3, 'Q': -0.4, 'R': -0.4, 'S': -0.2, 'T': -0.2, 'V': 0.0, 'W': -0.1, 'Y': 0.5},
                        'G': {'A': 0.1, 'C': -0.1, 'D': 0.1, 'E': 0.0, 'F': -0.5, 'G': 0.5, 'H': -0.2, 'I': -0.3, 'K': -0.1, 'L': -0.4, 'M': -0.3, 'N': 0.0, 'P': -0.1, 'Q': -0.1, 'R': 0.0, 'S': 0.1, 'T': -0.1, 'V': -0.2, 'W': -0.2, 'Y': -0.4},
                        'H': {'A': -0.2, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': 0.0, 'G': -0.2, 'H': 0.6, 'I': -0.3, 'K': 0.1, 'L': -0.2, 'M': -0.2, 'N': 0.1, 'P': 0.0, 'Q': 0.2, 'R': 0.2, 'S': -0.1, 'T': -0.1, 'V': -0.3, 'W': -0.3, 'Y': 0.4},
                        'I': {'A': 0.0, 'C': -0.2, 'D': -0.3, 'E': -0.3, 'F': 0.0, 'G': -0.3, 'H': -0.3, 'I': 0.4, 'K': -0.3, 'L': 0.2, 'M': 0.3, 'N': -0.2, 'P': -0.2, 'Q': -0.3, 'R': -0.3, 'S': -0.1, 'T': 0.1, 'V': 0.4, 'W': -0.4, 'Y': -0.2},
                        'K': {'A': -0.1, 'C': -0.3, 'D': 0.0, 'E': 0.1, 'F': -0.5, 'G': -0.1, 'H': 0.1, 'I': -0.3, 'K': 0.5, 'L': -0.3, 'M': -0.2, 'N': 0.1, 'P': -0.2, 'Q': 0.2, 'R': 0.4, 'S': -0.1, 'T': -0.1, 'V': -0.3, 'W': -0.3, 'Y': -0.3},
                        'L': {'A': -0.1, 'C': -0.3, 'D': -0.4, 'E': -0.4, 'F': 0.2, 'G': -0.4, 'H': -0.2, 'I': 0.2, 'K': -0.3, 'L': 0.5, 'M': 0.3, 'N': -0.3, 'P': 0.0, 'Q': -0.2, 'R': -0.3, 'S': -0.2, 'T': -0.1, 'V': 0.2, 'W': -0.2, 'Y': -0.1},
                        'M': {'A': -0.1, 'C': -0.2, 'D': -0.3, 'E': -0.3, 'F': 0.0, 'G': -0.3, 'H': -0.2, 'I': 0.3, 'K': -0.2, 'L': 0.3, 'M': 0.6, 'N': -0.2, 'P': -0.2, 'Q': -0.2, 'R': -0.2, 'S': -0.1, 'T': 0.0, 'V': 0.2, 'W': -0.3, 'Y': -0.2},
                        'N': {'A': 0.0, 'C': -0.1, 'D': 0.2, 'E': 0.1, 'F': -0.3, 'G': 0.0, 'H': 0.1, 'I': -0.2, 'K': 0.1, 'L': -0.3, 'M': -0.2, 'N': 0.3, 'P': -0.1, 'Q': 0.0, 'R': 0.0, 'S': 0.1, 'T': 0.1, 'V': -0.2, 'W': -0.5, 'Y': -0.1},
                        'P': {'A': 0.1, 'C': -0.2, 'D': -0.2, 'E': -0.2, 'F': -0.3, 'G': -0.1, 'H': 0.0, 'I': -0.2, 'K': -0.2, 'L': 0.0, 'M': -0.2, 'N': -0.1, 'P': 0.6, 'Q': 0.0, 'R': -0.1, 'S': 0.1, 'T': 0.1, 'V': -0.1, 'W': -0.4, 'Y': -0.3},
                        'Q': {'A': -0.1, 'C': -0.3, 'D': 0.1, 'E': 0.2, 'F': -0.4, 'G': -0.1, 'H': 0.2, 'I': -0.3, 'K': 0.2, 'L': -0.2, 'M': -0.2, 'N': 0.0, 'P': 0.0, 'Q': 0.5, 'R': 0.2, 'S': -0.1, 'T': -0.1, 'V': -0.3, 'W': -0.3, 'Y': -0.2},
                        'R': {'A': -0.1, 'C': -0.1, 'D': -0.1, 'E': 0.0, 'F': -0.4, 'G': 0.0, 'H': 0.2, 'I': -0.3, 'K': 0.4, 'L': -0.3, 'M': -0.2, 'N': 0.0, 'P': -0.1, 'Q': 0.2, 'R': 0.5, 'S': -0.1, 'T': -0.1, 'V': -0.3, 'W': 0.0, 'Y': -0.2},
                        'S': {'A': 0.1, 'C': 0.1, 'D': 0.0, 'E': -0.1, 'F': -0.2, 'G': 0.1, 'H': -0.1, 'I': -0.1, 'K': -0.1, 'L': -0.2, 'M': -0.1, 'N': 0.1, 'P': 0.1, 'Q': -0.1, 'R': -0.1, 'S': 0.2, 'T': 0.1, 'V': -0.1, 'W': -0.3, 'Y': -0.1},
                        'T': {'A': 0.2, 'C': -0.1, 'D': -0.1, 'E': -0.1, 'F': -0.2, 'G': -0.1, 'H': -0.1, 'I': 0.1, 'K': -0.1, 'L': -0.1, 'M': 0.0, 'N': 0.1, 'P': 0.1, 'Q': -0.1, 'R': -0.1, 'S': 0.1, 'T': 0.2, 'V': 0.0, 'W': -0.4, 'Y': -0.3},
                        'V': {'A': 0.1, 'C': -0.2, 'D': -0.2, 'E': -0.2, 'F': 0.0, 'G': -0.2, 'H': -0.3, 'I': 0.4, 'K': -0.3, 'L': 0.2, 'M': 0.2, 'N': -0.2, 'P': -0.1, 'Q': -0.3, 'R': -0.3, 'S': -0.1, 'T': 0.0, 'V': 0.4, 'W': -0.3, 'Y': -0.3},
                        'W': {'A': -0.4, 'C': 0.1, 'D': -0.5, 'E': -0.5, 'F': -0.1, 'G': -0.2, 'H': -0.3, 'I': -0.4, 'K': -0.3, 'L': -0.2, 'M': -0.3, 'N': -0.5, 'P': -0.4, 'Q': -0.3, 'R': 0.0, 'S': -0.3, 'T': -0.4, 'V': -0.3, 'W': 1.5, 'Y': 0.0},
                        'Y': {'A': -0.3, 'C': 0.2, 'D': -0.2, 'E': -0.4, 'F': 0.5, 'G': -0.4, 'H': 0.4, 'I': -0.2, 'K': -0.3, 'L': -0.1, 'M': -0.2, 'N': -0.1, 'P': -0.3, 'Q': -0.2, 'R': -0.2, 'S': -0.1, 'T': -0.3, 'V': -0.3, 'W': 0.0, 'Y': 0.9}
                },
                'LEVIN': {
                        'A': {'A': 2.0, 'C': 0.0, 'D': 0.0, 'E': 1.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': 0.0, 'K': 0.0, 'L': 0.0, 'M': 0.0, 'N': 0.0, 'P': -1.0, 'Q': 0.0, 'R': 0.0, 'S': 1.0, 'T': 0.0, 'V': 0.0, 'W': -1.0, 'Y': -1.0},
                        'C': {'A': 0.0, 'C': 2.0, 'D': 0.0, 'E': 0.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': 0.0, 'K': 0.0, 'L': 0.0, 'M': 0.0, 'N': 0.0, 'P': 0.0, 'Q': 0.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': 0.0, 'W': -1.0, 'Y': -1.0},
                        'D': {'A': 0.0, 'C': 0.0, 'D': 2.0, 'E': 1.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 0.0, 'L': -1.0, 'M': -1.0, 'N': 1.0, 'P': 0.0, 'Q': 0.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'E': {'A': 1.0, 'C': 0.0, 'D': 1.0, 'E': 2.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 0.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': 1.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'F': {'A': -1.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': 2.0, 'G': -1.0, 'H': -1.0, 'I': 1.0, 'K': -1.0, 'L': 0.0, 'M': 0.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': 0.0, 'W': 0.0, 'Y': 1.0},
                        'G': {'A': 0.0, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -1.0, 'G': 2.0, 'H': 0.0, 'I': -1.0, 'K': 0.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': 0.0, 'Q': 0.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'H': {'A': 0.0, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -1.0, 'G': 0.0, 'H': 2.0, 'I': -1.0, 'K': 0.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': 0.0, 'Q': 0.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': 0.0},
                        'I': {'A': 0.0, 'C': 0.0, 'D': -1.0, 'E': -1.0, 'F': 1.0, 'G': -1.0, 'H': -1.0, 'I': 2.0, 'K': -1.0, 'L': 0.0, 'M': 0.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': 0.0, 'V': 1.0, 'W': 0.0, 'Y': 0.0},
                        'K': {'A': 0.0, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 2.0, 'L': -1.0, 'M': -1.0, 'N': 1.0, 'P': 0.0, 'Q': 0.0, 'R': 1.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'L': {'A': 0.0, 'C': 0.0, 'D': -1.0, 'E': -1.0, 'F': 0.0, 'G': -1.0, 'H': -1.0, 'I': 0.0, 'K': -1.0, 'L': 2.0, 'M': 2.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': 0.0, 'V': 1.0, 'W': 0.0, 'Y': 0.0},
                        'M': {'A': 0.0, 'C': 0.0, 'D': -1.0, 'E': -1.0, 'F': 0.0, 'G': -1.0, 'H': -1.0, 'I': 0.0, 'K': -1.0, 'L': 2.0, 'M': 2.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': 0.0, 'V': 0.0, 'W': 0.0, 'Y': 0.0},
                        'N': {'A': 0.0, 'C': 0.0, 'D': 1.0, 'E': 0.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 1.0, 'L': -1.0, 'M': -1.0, 'N': 3.0, 'P': 0.0, 'Q': 1.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'P': {'A': -1.0, 'C': 0.0, 'D': 0.0, 'E': -1.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 0.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': 3.0, 'Q': 0.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'Q': {'A': 0.0, 'C': 0.0, 'D': 0.0, 'E': 1.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 0.0, 'L': -1.0, 'M': -1.0, 'N': 1.0, 'P': 0.0, 'Q': 2.0, 'R': 0.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'R': {'A': 0.0, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 1.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': 0.0, 'Q': 0.0, 'R': 2.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': 0.0, 'Y': -1.0},
                        'S': {'A': 1.0, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': -1.0, 'K': 0.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': 0.0, 'Q': 0.0, 'R': 0.0, 'S': 2.0, 'T': 0.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'T': {'A': 0.0, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -1.0, 'G': 0.0, 'H': 0.0, 'I': 0.0, 'K': 0.0, 'L': 0.0, 'M': 0.0, 'N': 0.0, 'P': 0.0, 'Q': 0.0, 'R': 0.0, 'S': 0.0, 'T': 2.0, 'V': 0.0, 'W': -1.0, 'Y': -1.0},
                        'V': {'A': 0.0, 'C': 0.0, 'D': -1.0, 'E': -1.0, 'F': 0.0, 'G': -1.0, 'H': -1.0, 'I': 1.0, 'K': -1.0, 'L': 1.0, 'M': 0.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': 0.0, 'V': 2.0, 'W': 0.0, 'Y': 0.0},
                        'W': {'A': -1.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': 0.0, 'G': -1.0, 'H': -1.0, 'I': 0.0, 'K': -1.0, 'L': 0.0, 'M': 0.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': 0.0, 'S': -1.0, 'T': -1.0, 'V': 0.0, 'W': 2.0, 'Y': 0.0},
                        'Y': {'A': -1.0, 'C': -1.0, 'D': -1.0, 'E': -1.0, 'F': 1.0, 'G': -1.0, 'H': 0.0, 'I': 0.0, 'K': -1.0, 'L': 0.0, 'M': 0.0, 'N': -1.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': 0.0, 'W': 0.0, 'Y': 2.0}
                },
                'MCLACHLAN': {
                        'A': {'A': 8.0, 'C': 1.0, 'D': 3.0, 'E': 4.0, 'F': 1.0, 'G': 3.0, 'H': 3.0, 'I': 2.0, 'K': 3.0, 'L': 2.0, 'M': 3.0, 'N': 3.0, 'P': 4.0, 'Q': 3.0, 'R': 2.0, 'S': 4.0, 'T': 3.0, 'V': 3.0, 'W': 1.0, 'Y': 1.0},
                        'C': {'A': 1.0, 'C': 9.0, 'D': 1.0, 'E': 0.0, 'F': 0.0, 'G': 1.0, 'H': 3.0, 'I': 1.0, 'K': 0.0, 'L': 0.0, 'M': 3.0, 'N': 1.0, 'P': 0.0, 'Q': 0.0, 'R': 1.0, 'S': 2.0, 'T': 2.0, 'V': 1.0, 'W': 2.0, 'Y': 1.0},
                        'D': {'A': 3.0, 'C': 1.0, 'D': 8.0, 'E': 5.0, 'F': 1.0, 'G': 3.0, 'H': 4.0, 'I': 1.0, 'K': 3.0, 'L': 1.0, 'M': 2.0, 'N': 5.0, 'P': 3.0, 'Q': 4.0, 'R': 1.0, 'S': 3.0, 'T': 3.0, 'V': 1.0, 'W': 0.0, 'Y': 1.0},
                        'E': {'A': 4.0, 'C': 0.0, 'D': 5.0, 'E': 8.0, 'F': 0.0, 'G': 3.0, 'H': 2.0, 'I': 1.0, 'K': 4.0, 'L': 1.0, 'M': 1.0, 'N': 4.0, 'P': 4.0, 'Q': 5.0, 'R': 3.0, 'S': 4.0, 'T': 4.0, 'V': 2.0, 'W': 1.0, 'Y': 2.0},
                        'F': {'A': 1.0, 'C': 0.0, 'D': 1.0, 'E': 0.0, 'F': 9.0, 'G': 0.0, 'H': 4.0, 'I': 3.0, 'K': 0.0, 'L': 5.0, 'M': 5.0, 'N': 0.0, 'P': 1.0, 'Q': 0.0, 'R': 1.0, 'S': 2.0, 'T': 1.0, 'V': 3.0, 'W': 6.0, 'Y': 6.0},
                        'G': {'A': 3.0, 'C': 1.0, 'D': 3.0, 'E': 3.0, 'F': 0.0, 'G': 8.0, 'H': 2.0, 'I': 1.0, 'K': 3.0, 'L': 1.0, 'M': 1.0, 'N': 3.0, 'P': 3.0, 'Q': 2.0, 'R': 3.0, 'S': 3.0, 'T': 2.0, 'V': 2.0, 'W': 1.0, 'Y': 0.0},
                        'H': {'A': 3.0, 'C': 3.0, 'D': 4.0, 'E': 2.0, 'F': 4.0, 'G': 2.0, 'H': 8.0, 'I': 2.0, 'K': 4.0, 'L': 2.0, 'M': 3.0, 'N': 4.0, 'P': 3.0, 'Q': 4.0, 'R': 5.0, 'S': 3.0, 'T': 4.0, 'V': 2.0, 'W': 3.0, 'Y': 4.0},
                        'I': {'A': 2.0, 'C': 1.0, 'D': 1.0, 'E': 1.0, 'F': 3.0, 'G': 1.0, 'H': 2.0, 'I': 8.0, 'K': 1.0, 'L': 5.0, 'M': 5.0, 'N': 1.0, 'P': 1.0, 'Q': 0.0, 'R': 1.0, 'S': 2.0, 'T': 3.0, 'V': 5.0, 'W': 3.0, 'Y': 3.0},
                        'K': {'A': 3.0, 'C': 0.0, 'D': 3.0, 'E': 4.0, 'F': 0.0, 'G': 3.0, 'H': 4.0, 'I': 1.0, 'K': 8.0, 'L': 2.0, 'M': 1.0, 'N': 4.0, 'P': 3.0, 'Q': 4.0, 'R': 5.0, 'S': 3.0, 'T': 3.0, 'V': 2.0, 'W': 1.0, 'Y': 1.0},
                        'L': {'A': 2.0, 'C': 0.0, 'D': 1.0, 'E': 1.0, 'F': 5.0, 'G': 1.0, 'H': 2.0, 'I': 5.0, 'K': 2.0, 'L': 8.0, 'M': 6.0, 'N': 1.0, 'P': 1.0, 'Q': 3.0, 'R': 2.0, 'S': 2.0, 'T': 3.0, 'V': 5.0, 'W': 3.0, 'Y': 3.0},
                        'M': {'A': 3.0, 'C': 3.0, 'D': 2.0, 'E': 1.0, 'F': 5.0, 'G': 1.0, 'H': 3.0, 'I': 5.0, 'K': 1.0, 'L': 6.0, 'M': 8.0, 'N': 2.0, 'P': 1.0, 'Q': 3.0, 'R': 1.0, 'S': 2.0, 'T': 3.0, 'V': 4.0, 'W': 1.0, 'Y': 2.0},
                        'N': {'A': 3.0, 'C': 1.0, 'D': 5.0, 'E': 4.0, 'F': 0.0, 'G': 3.0, 'H': 4.0, 'I': 1.0, 'K': 4.0, 'L': 1.0, 'M': 2.0, 'N': 8.0, 'P': 1.0, 'Q': 4.0, 'R': 3.0, 'S': 5.0, 'T': 3.0, 'V': 1.0, 'W': 0.0, 'Y': 2.0},
                        'P': {'A': 4.0, 'C': 0.0, 'D': 3.0, 'E': 4.0, 'F': 1.0, 'G': 3.0, 'H': 3.0, 'I': 1.0, 'K': 3.0, 'L': 1.0, 'M': 1.0, 'N': 1.0, 'P': 8.0, 'Q': 3.0, 'R': 3.0, 'S': 3.0, 'T': 3.0, 'V': 2.0, 'W': 0.0, 'Y': 0.0},
                        'Q': {'A': 3.0, 'C': 0.0, 'D': 4.0, 'E': 5.0, 'F': 0.0, 'G': 2.0, 'H': 4.0, 'I': 0.0, 'K': 4.0, 'L': 3.0, 'M': 3.0, 'N': 4.0, 'P': 3.0, 'Q': 8.0, 'R': 5.0, 'S': 4.0, 'T': 3.0, 'V': 2.0, 'W': 2.0, 'Y': 1.0},
                        'R': {'A': 2.0, 'C': 1.0, 'D': 1.0, 'E': 3.0, 'F': 1.0, 'G': 3.0, 'H': 5.0, 'I': 1.0, 'K': 5.0, 'L': 2.0, 'M': 1.0, 'N': 3.0, 'P': 3.0, 'Q': 5.0, 'R': 8.0, 'S': 4.0, 'T': 3.0, 'V': 2.0, 'W': 3.0, 'Y': 2.0},
                        'S': {'A': 4.0, 'C': 2.0, 'D': 3.0, 'E': 4.0, 'F': 2.0, 'G': 3.0, 'H': 3.0, 'I': 2.0, 'K': 3.0, 'L': 2.0, 'M': 2.0, 'N': 5.0, 'P': 3.0, 'Q': 4.0, 'R': 4.0, 'S': 8.0, 'T': 5.0, 'V': 2.0, 'W': 3.0, 'Y': 3.0},
                        'T': {'A': 3.0, 'C': 2.0, 'D': 3.0, 'E': 4.0, 'F': 1.0, 'G': 2.0, 'H': 4.0, 'I': 3.0, 'K': 3.0, 'L': 3.0, 'M': 3.0, 'N': 3.0, 'P': 3.0, 'Q': 3.0, 'R': 3.0, 'S': 5.0, 'T': 8.0, 'V': 3.0, 'W': 2.0, 'Y': 1.0},
                        'V': {'A': 3.0, 'C': 1.0, 'D': 1.0, 'E': 2.0, 'F': 3.0, 'G': 2.0, 'H': 2.0, 'I': 5.0, 'K': 2.0, 'L': 5.0, 'M': 4.0, 'N': 1.0, 'P': 2.0, 'Q': 2.0, 'R': 2.0, 'S': 2.0, 'T': 3.0, 'V': 8.0, 'W': 2.0, 'Y': 3.0},
                        'W': {'A': 1.0, 'C': 2.0, 'D': 0.0, 'E': 1.0, 'F': 6.0, 'G': 1.0, 'H': 3.0, 'I': 3.0, 'K': 1.0, 'L': 3.0, 'M': 1.0, 'N': 0.0, 'P': 0.0, 'Q': 2.0, 'R': 3.0, 'S': 3.0, 'T': 2.0, 'V': 2.0, 'W': 9.0, 'Y': 6.0},
                        'Y': {'A': 1.0, 'C': 1.0, 'D': 1.0, 'E': 2.0, 'F': 6.0, 'G': 0.0, 'H': 4.0, 'I': 3.0, 'K': 1.0, 'L': 3.0, 'M': 2.0, 'N': 2.0, 'P': 0.0, 'Q': 1.0, 'R': 2.0, 'S': 3.0, 'T': 1.0, 'V': 3.0, 'W': 6.0, 'Y': 9.0}
                },
                'MDM78': {
                        'A': {'A': 0.18, 'C': -0.2, 'D': 0.03, 'E': 0.03, 'F': -0.35, 'G': 0.13, 'H': -0.14, 'I': -0.05, 'K': -0.12, 'L': -0.19, 'M': -0.11, 'N': 0.02, 'P': 0.11, 'Q': -0.04, 'R': -0.15, 'S': 0.11, 'T': 0.12, 'V': 0.02, 'W': -0.58, 'Y': -0.35},
                        'C': {'A': -0.2, 'C': 1.19, 'D': -0.51, 'E': -0.53, 'F': -0.43, 'G': -0.34, 'H': -0.34, 'I': -0.23, 'K': -0.54, 'L': -0.6, 'M': -0.52, 'N': -0.36, 'P': -0.28, 'Q': -0.54, 'R': -0.36, 'S': 0.0, 'T': -0.22, 'V': -0.19, 'W': -0.78, 'Y': 0.03},
                        'D': {'A': 0.03, 'C': -0.51, 'D': 0.39, 'E': 0.34, 'F': -0.56, 'G': 0.06, 'H': 0.07, 'I': -0.24, 'K': 0.01, 'L': -0.4, 'M': -0.26, 'N': 0.21, 'P': -0.1, 'Q': 0.16, 'R': -0.13, 'S': 0.03, 'T': -0.01, 'V': -0.21, 'W': -0.68, 'Y': -0.43},
                        'E': {'A': 0.03, 'C': -0.53, 'D': 0.34, 'E': 0.38, 'F': -0.54, 'G': 0.02, 'H': 0.07, 'I': -0.2, 'K': -0.01, 'L': -0.34, 'M': -0.21, 'N': 0.14, 'P': -0.06, 'Q': 0.25, 'R': -0.11, 'S': 0.0, 'T': -0.04, 'V': -0.18, 'W': -0.7, 'Y': -0.43},
                        'F': {'A': -0.35, 'C': -0.43, 'D': -0.56, 'E': -0.54, 'F': 0.91, 'G': -0.48, 'H': -0.18, 'I': 0.1, 'K': -0.53, 'L': 0.18, 'M': 0.02, 'N': -0.35, 'P': -0.46, 'Q': -0.47, 'R': -0.45, 'S': -0.32, 'T': -0.31, 'V': -0.12, 'W': 0.04, 'Y': 0.7},
                        'G': {'A': 0.13, 'C': -0.34, 'D': 0.06, 'E': 0.02, 'F': -0.48, 'G': 0.48, 'H': -0.21, 'I': -0.26, 'K': -0.17, 'L': -0.41, 'M': -0.28, 'N': 0.03, 'P': -0.05, 'Q': -0.12, 'R': -0.26, 'S': 0.11, 'T': 0.0, 'V': -0.14, 'W': -0.7, 'Y': -0.52},
                        'H': {'A': -0.14, 'C': -0.34, 'D': 0.07, 'E': 0.07, 'F': -0.18, 'G': -0.21, 'H': 0.65, 'I': -0.24, 'K': 0.0, 'L': -0.21, 'M': -0.21, 'N': 0.16, 'P': -0.02, 'Q': 0.29, 'R': 0.16, 'S': -0.08, 'T': -0.13, 'V': -0.22, 'W': -0.28, 'Y': -0.01},
                        'I': {'A': -0.05, 'C': -0.23, 'D': -0.24, 'E': -0.2, 'F': 0.1, 'G': -0.26, 'H': -0.24, 'I': 0.45, 'K': -0.19, 'L': 0.24, 'M': 0.22, 'N': -0.18, 'P': -0.2, 'Q': -0.2, 'R': -0.2, 'S': -0.14, 'T': 0.01, 'V': 0.37, 'W': -0.51, 'Y': -0.09},
                        'K': {'A': -0.12, 'C': -0.54, 'D': 0.01, 'E': -0.01, 'F': -0.53, 'G': -0.17, 'H': 0.0, 'I': -0.19, 'K': 0.47, 'L': -0.29, 'M': 0.04, 'N': 0.1, 'P': -0.11, 'Q': 0.07, 'R': 0.34, 'S': -0.02, 'T': 0.0, 'V': -0.24, 'W': -0.35, 'Y': -0.44},
                        'L': {'A': -0.19, 'C': -0.6, 'D': -0.4, 'E': -0.34, 'F': 0.18, 'G': -0.41, 'H': -0.21, 'I': 0.24, 'K': -0.29, 'L': 0.59, 'M': 0.37, 'N': -0.29, 'P': -0.25, 'Q': -0.18, 'R': -0.3, 'S': -0.28, 'T': -0.17, 'V': 0.19, 'W': -0.18, 'Y': -0.09},
                        'M': {'A': -0.11, 'C': -0.52, 'D': -0.26, 'E': -0.21, 'F': 0.02, 'G': -0.28, 'H': -0.21, 'I': 0.22, 'K': 0.04, 'L': 0.37, 'M': 0.64, 'N': -0.17, 'P': -0.21, 'Q': -0.1, 'R': -0.04, 'S': -0.16, 'T': -0.06, 'V': 0.18, 'W': -0.42, 'Y': -0.24},
                        'N': {'A': 0.02, 'C': -0.36, 'D': 0.21, 'E': 0.14, 'F': -0.35, 'G': 0.03, 'H': 0.16, 'I': -0.18, 'K': 0.1, 'L': -0.29, 'M': -0.17, 'N': 0.2, 'P': -0.05, 'Q': 0.08, 'R': 0.0, 'S': 0.07, 'T': 0.04, 'V': -0.17, 'W': -0.42, 'Y': -0.21},
                        'P': {'A': 0.11, 'C': -0.28, 'D': -0.1, 'E': -0.06, 'F': -0.46, 'G': -0.05, 'H': -0.02, 'I': -0.2, 'K': -0.11, 'L': -0.25, 'M': -0.21, 'N': -0.05, 'P': 0.59, 'Q': 0.02, 'R': -0.02, 'S': 0.09, 'T': 0.03, 'V': -0.12, 'W': -0.56, 'Y': -0.49},
                        'Q': {'A': -0.04, 'C': -0.54, 'D': 0.16, 'E': 0.25, 'F': -0.47, 'G': -0.12, 'H': 0.29, 'I': -0.2, 'K': 0.07, 'L': -0.18, 'M': -0.1, 'N': 0.08, 'P': 0.02, 'Q': 0.4, 'R': 0.13, 'S': -0.05, 'T': -0.08, 'V': -0.19, 'W': -0.48, 'Y': -0.4},
                        'R': {'A': -0.15, 'C': -0.36, 'D': -0.13, 'E': -0.11, 'F': -0.45, 'G': -0.26, 'H': 0.16, 'I': -0.2, 'K': 0.34, 'L': -0.3, 'M': -0.04, 'N': 0.0, 'P': -0.02, 'Q': 0.13, 'R': 0.61, 'S': -0.03, 'T': -0.09, 'V': -0.25, 'W': 0.22, 'Y': -0.42},
                        'S': {'A': 0.11, 'C': 0.0, 'D': 0.03, 'E': 0.0, 'F': -0.32, 'G': 0.11, 'H': -0.08, 'I': -0.14, 'K': -0.02, 'L': -0.28, 'M': -0.16, 'N': 0.07, 'P': 0.09, 'Q': -0.05, 'R': -0.03, 'S': 0.16, 'T': 0.13, 'V': -0.1, 'W': -0.25, 'Y': -0.28},
                        'T': {'A': 0.12, 'C': -0.22, 'D': -0.01, 'E': -0.04, 'F': -0.31, 'G': 0.0, 'H': -0.13, 'I': 0.01, 'K': 0.0, 'L': -0.17, 'M': -0.06, 'N': 0.04, 'P': 0.03, 'Q': -0.08, 'R': -0.09, 'S': 0.13, 'T': 0.26, 'V': 0.03, 'W': -0.52, 'Y': -0.27},
                        'V': {'A': 0.02, 'C': -0.19, 'D': -0.21, 'E': -0.18, 'F': -0.12, 'G': -0.14, 'H': -0.22, 'I': 0.37, 'K': -0.24, 'L': 0.19, 'M': 0.18, 'N': -0.17, 'P': -0.12, 'Q': -0.19, 'R': -0.25, 'S': -0.1, 'T': 0.03, 'V': 0.43, 'W': -0.62, 'Y': -0.25},
                        'W': {'A': -0.58, 'C': -0.78, 'D': -0.68, 'E': -0.7, 'F': 0.04, 'G': -0.7, 'H': -0.28, 'I': -0.51, 'K': -0.35, 'L': -0.18, 'M': -0.42, 'N': -0.42, 'P': -0.56, 'Q': -0.48, 'R': 0.22, 'S': -0.25, 'T': -0.52, 'V': -0.62, 'W': 1.73, 'Y': -0.02},
                        'Y': {'A': -0.35, 'C': 0.03, 'D': -0.43, 'E': -0.43, 'F': 0.7, 'G': -0.52, 'H': -0.01, 'I': -0.09, 'K': -0.44, 'L': -0.09, 'M': -0.24, 'N': -0.21, 'P': -0.49, 'Q': -0.4, 'R': -0.42, 'S': -0.28, 'T': -0.27, 'V': -0.25, 'W': -0.02, 'Y': 1.01}
                },
                'MEGABLAST': {
                        'A': {'A': 1.0, 'B': -2.0, 'C': -2.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -2.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -2.0, 'T': -2.0, 'V': -1.0, 'W': -1.0, 'Y': -2.0},
                        'B': {'A': -2.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'C': {'A': -2.0, 'B': -1.0, 'C': 1.0, 'D': -2.0, 'G': -2.0, 'H': -1.0, 'K': -2.0, 'M': -1.0, 'N': -1.0, 'R': -2.0, 'S': -1.0, 'T': -2.0, 'V': -1.0, 'W': -2.0, 'Y': -1.0},
                        'D': {'A': -1.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'G': {'A': -2.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': 1.0, 'H': -2.0, 'K': -1.0, 'M': -2.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -2.0, 'V': -1.0, 'W': -2.0, 'Y': -2.0},
                        'H': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'K': {'A': -2.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -2.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'M': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -2.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -2.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'N': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'R': {'A': -1.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -2.0, 'V': -1.0, 'W': -1.0, 'Y': -2.0},
                        'S': {'A': -2.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -2.0, 'V': -1.0, 'W': -2.0, 'Y': -1.0},
                        'T': {'A': -2.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -1.0, 'M': -2.0, 'N': -1.0, 'R': -2.0, 'S': -2.0, 'T': 1.0, 'V': -2.0, 'W': -1.0, 'Y': -1.0},
                        'V': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -2.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'W': {'A': -1.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -2.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'Y': {'A': -2.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -2.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0}
                },
                'PAM250': {
                        '*': {'*': 1.0, 'A': -8.0, 'B': -8.0, 'C': -8.0, 'D': -8.0, 'E': -8.0, 'F': -8.0, 'G': -8.0, 'H': -8.0, 'I': -8.0, 'K': -8.0, 'L': -8.0, 'M': -8.0, 'N': -8.0, 'P': -8.0, 'Q': -8.0, 'R': -8.0, 'S': -8.0, 'T': -8.0, 'V': -8.0, 'W': -8.0, 'X': -8.0, 'Y': -8.0, 'Z': -8.0},
                        'A': {'*': -8.0, 'A': 2.0, 'B': 0.0, 'C': -2.0, 'D': 0.0, 'E': 0.0, 'F': -3.0, 'G': 1.0, 'H': -1.0, 'I': -1.0, 'K': -1.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': 1.0, 'Q': 0.0, 'R': -2.0, 'S': 1.0, 'T': 1.0, 'V': 0.0, 'W': -6.0, 'X': 0.0, 'Y': -3.0, 'Z': 0.0},
                        'B': {'*': -8.0, 'A': 0.0, 'B': 3.0, 'C': -4.0, 'D': 3.0, 'E': 3.0, 'F': -4.0, 'G': 0.0, 'H': 1.0, 'I': -2.0, 'K': 1.0, 'L': -3.0, 'M': -2.0, 'N': 2.0, 'P': -1.0, 'Q': 1.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -2.0, 'W': -5.0, 'X': -1.0, 'Y': -3.0, 'Z': 2.0},
                        'C': {'*': -8.0, 'A': -2.0, 'B': -4.0, 'C': 12.0, 'D': -5.0, 'E': -5.0, 'F': -4.0, 'G': -3.0, 'H': -3.0, 'I': -2.0, 'K': -5.0, 'L': -6.0, 'M': -5.0, 'N': -4.0, 'P': -3.0, 'Q': -5.0, 'R': -4.0, 'S': 0.0, 'T': -2.0, 'V': -2.0, 'W': -8.0, 'X': -3.0, 'Y': 0.0, 'Z': -5.0},
                        'D': {'*': -8.0, 'A': 0.0, 'B': 3.0, 'C': -5.0, 'D': 4.0, 'E': 3.0, 'F': -6.0, 'G': 1.0, 'H': 1.0, 'I': -2.0, 'K': 0.0, 'L': -4.0, 'M': -3.0, 'N': 2.0, 'P': -1.0, 'Q': 2.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -2.0, 'W': -7.0, 'X': -1.0, 'Y': -4.0, 'Z': 3.0},
                        'E': {'*': -8.0, 'A': 0.0, 'B': 3.0, 'C': -5.0, 'D': 3.0, 'E': 4.0, 'F': -5.0, 'G': 0.0, 'H': 1.0, 'I': -2.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 1.0, 'P': -1.0, 'Q': 2.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -2.0, 'W': -7.0, 'X': -1.0, 'Y': -4.0, 'Z': 3.0},
                        'F': {'*': -8.0, 'A': -3.0, 'B': -4.0, 'C': -4.0, 'D': -6.0, 'E': -5.0, 'F': 9.0, 'G': -5.0, 'H': -2.0, 'I': 1.0, 'K': -5.0, 'L': 2.0, 'M': 0.0, 'N': -3.0, 'P': -5.0, 'Q': -5.0, 'R': -4.0, 'S': -3.0, 'T': -3.0, 'V': -1.0, 'W': 0.0, 'X': -2.0, 'Y': 7.0, 'Z': -5.0},
                        'G': {'*': -8.0, 'A': 1.0, 'B': 0.0, 'C': -3.0, 'D': 1.0, 'E': 0.0, 'F': -5.0, 'G': 5.0, 'H': -2.0, 'I': -3.0, 'K': -2.0, 'L': -4.0, 'M': -3.0, 'N': 0.0, 'P': 0.0, 'Q': -1.0, 'R': -3.0, 'S': 1.0, 'T': 0.0, 'V': -1.0, 'W': -7.0, 'X': -1.0, 'Y': -5.0, 'Z': 0.0},
                        'H': {'*': -8.0, 'A': -1.0, 'B': 1.0, 'C': -3.0, 'D': 1.0, 'E': 1.0, 'F': -2.0, 'G': -2.0, 'H': 6.0, 'I': -2.0, 'K': 0.0, 'L': -2.0, 'M': -2.0, 'N': 2.0, 'P': 0.0, 'Q': 3.0, 'R': 2.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': 0.0, 'Z': 2.0},
                        'I': {'*': -8.0, 'A': -1.0, 'B': -2.0, 'C': -2.0, 'D': -2.0, 'E': -2.0, 'F': 1.0, 'G': -3.0, 'H': -2.0, 'I': 5.0, 'K': -2.0, 'L': 2.0, 'M': 2.0, 'N': -2.0, 'P': -2.0, 'Q': -2.0, 'R': -2.0, 'S': -1.0, 'T': 0.0, 'V': 4.0, 'W': -5.0, 'X': -1.0, 'Y': -1.0, 'Z': -2.0},
                        'K': {'*': -8.0, 'A': -1.0, 'B': 1.0, 'C': -5.0, 'D': 0.0, 'E': 0.0, 'F': -5.0, 'G': -2.0, 'H': 0.0, 'I': -2.0, 'K': 5.0, 'L': -3.0, 'M': 0.0, 'N': 1.0, 'P': -1.0, 'Q': 1.0, 'R': 3.0, 'S': 0.0, 'T': 0.0, 'V': -2.0, 'W': -3.0, 'X': -1.0, 'Y': -4.0, 'Z': 0.0},
                        'L': {'*': -8.0, 'A': -2.0, 'B': -3.0, 'C': -6.0, 'D': -4.0, 'E': -3.0, 'F': 2.0, 'G': -4.0, 'H': -2.0, 'I': 2.0, 'K': -3.0, 'L': 6.0, 'M': 4.0, 'N': -3.0, 'P': -3.0, 'Q': -2.0, 'R': -3.0, 'S': -3.0, 'T': -2.0, 'V': 2.0, 'W': -2.0, 'X': -1.0, 'Y': -1.0, 'Z': -3.0},
                        'M': {'*': -8.0, 'A': -1.0, 'B': -2.0, 'C': -5.0, 'D': -3.0, 'E': -2.0, 'F': 0.0, 'G': -3.0, 'H': -2.0, 'I': 2.0, 'K': 0.0, 'L': 4.0, 'M': 6.0, 'N': -2.0, 'P': -2.0, 'Q': -1.0, 'R': 0.0, 'S': -2.0, 'T': -1.0, 'V': 2.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': -2.0},
                        'N': {'*': -8.0, 'A': 0.0, 'B': 2.0, 'C': -4.0, 'D': 2.0, 'E': 1.0, 'F': -3.0, 'G': 0.0, 'H': 2.0, 'I': -2.0, 'K': 1.0, 'L': -3.0, 'M': -2.0, 'N': 2.0, 'P': 0.0, 'Q': 1.0, 'R': 0.0, 'S': 1.0, 'T': 0.0, 'V': -2.0, 'W': -4.0, 'X': 0.0, 'Y': -2.0, 'Z': 1.0},
                        'P': {'*': -8.0, 'A': 1.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'E': -1.0, 'F': -5.0, 'G': 0.0, 'H': 0.0, 'I': -2.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': 6.0, 'Q': 0.0, 'R': 0.0, 'S': 1.0, 'T': 0.0, 'V': -1.0, 'W': -6.0, 'X': -1.0, 'Y': -5.0, 'Z': 0.0},
                        'Q': {'*': -8.0, 'A': 0.0, 'B': 1.0, 'C': -5.0, 'D': 2.0, 'E': 2.0, 'F': -5.0, 'G': -1.0, 'H': 3.0, 'I': -2.0, 'K': 1.0, 'L': -2.0, 'M': -1.0, 'N': 1.0, 'P': 0.0, 'Q': 4.0, 'R': 1.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -5.0, 'X': -1.0, 'Y': -4.0, 'Z': 3.0},
                        'R': {'*': -8.0, 'A': -2.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'E': -1.0, 'F': -4.0, 'G': -3.0, 'H': 2.0, 'I': -2.0, 'K': 3.0, 'L': -3.0, 'M': 0.0, 'N': 0.0, 'P': 0.0, 'Q': 1.0, 'R': 6.0, 'S': 0.0, 'T': -1.0, 'V': -2.0, 'W': 2.0, 'X': -1.0, 'Y': -4.0, 'Z': 0.0},
                        'S': {'*': -8.0, 'A': 1.0, 'B': 0.0, 'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': -3.0, 'G': 1.0, 'H': -1.0, 'I': -1.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 1.0, 'P': 1.0, 'Q': -1.0, 'R': 0.0, 'S': 2.0, 'T': 1.0, 'V': -1.0, 'W': -2.0, 'X': 0.0, 'Y': -3.0, 'Z': 0.0},
                        'T': {'*': -8.0, 'A': 1.0, 'B': 0.0, 'C': -2.0, 'D': 0.0, 'E': 0.0, 'F': -3.0, 'G': 0.0, 'H': -1.0, 'I': 0.0, 'K': 0.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': 0.0, 'Q': -1.0, 'R': -1.0, 'S': 1.0, 'T': 3.0, 'V': 0.0, 'W': -5.0, 'X': 0.0, 'Y': -3.0, 'Z': -1.0},
                        'V': {'*': -8.0, 'A': 0.0, 'B': -2.0, 'C': -2.0, 'D': -2.0, 'E': -2.0, 'F': -1.0, 'G': -1.0, 'H': -2.0, 'I': 4.0, 'K': -2.0, 'L': 2.0, 'M': 2.0, 'N': -2.0, 'P': -1.0, 'Q': -2.0, 'R': -2.0, 'S': -1.0, 'T': 0.0, 'V': 4.0, 'W': -6.0, 'X': -1.0, 'Y': -2.0, 'Z': -2.0},
                        'W': {'*': -8.0, 'A': -6.0, 'B': -5.0, 'C': -8.0, 'D': -7.0, 'E': -7.0, 'F': 0.0, 'G': -7.0, 'H': -3.0, 'I': -5.0, 'K': -3.0, 'L': -2.0, 'M': -4.0, 'N': -4.0, 'P': -6.0, 'Q': -5.0, 'R': 2.0, 'S': -2.0, 'T': -5.0, 'V': -6.0, 'W': 17.0, 'X': -4.0, 'Y': 0.0, 'Z': -6.0},
                        'X': {'*': -8.0, 'A': 0.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'E': -1.0, 'F': -2.0, 'G': -1.0, 'H': -1.0, 'I': -1.0, 'K': -1.0, 'L': -1.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': -1.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -1.0, 'W': -4.0, 'X': -1.0, 'Y': -2.0, 'Z': -1.0},
                        'Y': {'*': -8.0, 'A': -3.0, 'B': -3.0, 'C': 0.0, 'D': -4.0, 'E': -4.0, 'F': 7.0, 'G': -5.0, 'H': 0.0, 'I': -1.0, 'K': -4.0, 'L': -1.0, 'M': -2.0, 'N': -2.0, 'P': -5.0, 'Q': -4.0, 'R': -4.0, 'S': -3.0, 'T': -3.0, 'V': -2.0, 'W': 0.0, 'X': -2.0, 'Y': 10.0, 'Z': -4.0},
                        'Z': {'*': -8.0, 'A': 0.0, 'B': 2.0, 'C': -5.0, 'D': 3.0, 'E': 3.0, 'F': -5.0, 'G': 0.0, 'H': 2.0, 'I': -2.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 1.0, 'P': 0.0, 'Q': 3.0, 'R': 0.0, 'S': 0.0, 'T': -1.0, 'V': -2.0, 'W': -6.0, 'X': -1.0, 'Y': -4.0, 'Z': 3.0}
                },
                'PAM30': {
                        '*': {'*': 1.0, 'A': -17.0, 'B': -17.0, 'C': -17.0, 'D': -17.0, 'E': -17.0, 'F': -17.0, 'G': -17.0, 'H': -17.0, 'I': -17.0, 'K': -17.0, 'L': -17.0, 'M': -17.0, 'N': -17.0, 'P': -17.0, 'Q': -17.0, 'R': -17.0, 'S': -17.0, 'T': -17.0, 'V': -17.0, 'W': -17.0, 'X': -17.0, 'Y': -17.0, 'Z': -17.0},
                        'A': {'*': -17.0, 'A': 6.0, 'B': -3.0, 'C': -6.0, 'D': -3.0, 'E': -2.0, 'F': -8.0, 'G': -2.0, 'H': -7.0, 'I': -5.0, 'K': -7.0, 'L': -6.0, 'M': -5.0, 'N': -4.0, 'P': -2.0, 'Q': -4.0, 'R': -7.0, 'S': 0.0, 'T': -1.0, 'V': -2.0, 'W': -13.0, 'X': -3.0, 'Y': -8.0, 'Z': -3.0},
                        'B': {'*': -17.0, 'A': -3.0, 'B': 6.0, 'C': -12.0, 'D': 6.0, 'E': 1.0, 'F': -10.0, 'G': -3.0, 'H': -1.0, 'I': -6.0, 'K': -2.0, 'L': -9.0, 'M': -10.0, 'N': 6.0, 'P': -7.0, 'Q': -3.0, 'R': -7.0, 'S': -1.0, 'T': -3.0, 'V': -8.0, 'W': -10.0, 'X': -5.0, 'Y': -6.0, 'Z': 0.0},
                        'C': {'*': -17.0, 'A': -6.0, 'B': -12.0, 'C': 10.0, 'D': -14.0, 'E': -14.0, 'F': -13.0, 'G': -9.0, 'H': -7.0, 'I': -6.0, 'K': -14.0, 'L': -15.0, 'M': -13.0, 'N': -11.0, 'P': -8.0, 'Q': -14.0, 'R': -8.0, 'S': -3.0, 'T': -8.0, 'V': -6.0, 'W': -15.0, 'X': -9.0, 'Y': -4.0, 'Z': -14.0},
                        'D': {'*': -17.0, 'A': -3.0, 'B': 6.0, 'C': -14.0, 'D': 8.0, 'E': 2.0, 'F': -15.0, 'G': -3.0, 'H': -4.0, 'I': -7.0, 'K': -4.0, 'L': -12.0, 'M': -11.0, 'N': 2.0, 'P': -8.0, 'Q': -2.0, 'R': -10.0, 'S': -4.0, 'T': -5.0, 'V': -8.0, 'W': -15.0, 'X': -5.0, 'Y': -11.0, 'Z': 1.0},
                        'E': {'*': -17.0, 'A': -2.0, 'B': 1.0, 'C': -14.0, 'D': 2.0, 'E': 8.0, 'F': -14.0, 'G': -4.0, 'H': -5.0, 'I': -5.0, 'K': -4.0, 'L': -9.0, 'M': -7.0, 'N': -2.0, 'P': -5.0, 'Q': 1.0, 'R': -9.0, 'S': -4.0, 'T': -6.0, 'V': -6.0, 'W': -17.0, 'X': -5.0, 'Y': -8.0, 'Z': 6.0},
                        'F': {'*': -17.0, 'A': -8.0, 'B': -10.0, 'C': -13.0, 'D': -15.0, 'E': -14.0, 'F': 9.0, 'G': -9.0, 'H': -6.0, 'I': -2.0, 'K': -14.0, 'L': -3.0, 'M': -4.0, 'N': -9.0, 'P': -10.0, 'Q': -13.0, 'R': -9.0, 'S': -6.0, 'T': -9.0, 'V': -8.0, 'W': -4.0, 'X': -8.0, 'Y': 2.0, 'Z': -13.0},
                        'G': {'*': -17.0, 'A': -2.0, 'B': -3.0, 'C': -9.0, 'D': -3.0, 'E': -4.0, 'F': -9.0, 'G': 6.0, 'H': -9.0, 'I': -11.0, 'K': -7.0, 'L': -10.0, 'M': -8.0, 'N': -3.0, 'P': -6.0, 'Q': -7.0, 'R': -9.0, 'S': -2.0, 'T': -6.0, 'V': -5.0, 'W': -15.0, 'X': -5.0, 'Y': -14.0, 'Z': -5.0},
                        'H': {'*': -17.0, 'A': -7.0, 'B': -1.0, 'C': -7.0, 'D': -4.0, 'E': -5.0, 'F': -6.0, 'G': -9.0, 'H': 9.0, 'I': -9.0, 'K': -6.0, 'L': -6.0, 'M': -10.0, 'N': 0.0, 'P': -4.0, 'Q': 1.0, 'R': -2.0, 'S': -6.0, 'T': -7.0, 'V': -6.0, 'W': -7.0, 'X': -5.0, 'Y': -3.0, 'Z': -1.0},
                        'I': {'*': -17.0, 'A': -5.0, 'B': -6.0, 'C': -6.0, 'D': -7.0, 'E': -5.0, 'F': -2.0, 'G': -11.0, 'H': -9.0, 'I': 8.0, 'K': -6.0, 'L': -1.0, 'M': -1.0, 'N': -5.0, 'P': -8.0, 'Q': -8.0, 'R': -5.0, 'S': -7.0, 'T': -2.0, 'V': 2.0, 'W': -14.0, 'X': -5.0, 'Y': -6.0, 'Z': -6.0},
                        'K': {'*': -17.0, 'A': -7.0, 'B': -2.0, 'C': -14.0, 'D': -4.0, 'E': -4.0, 'F': -14.0, 'G': -7.0, 'H': -6.0, 'I': -6.0, 'K': 7.0, 'L': -8.0, 'M': -2.0, 'N': -1.0, 'P': -6.0, 'Q': -3.0, 'R': 0.0, 'S': -4.0, 'T': -3.0, 'V': -9.0, 'W': -12.0, 'X': -5.0, 'Y': -9.0, 'Z': -4.0},
                        'L': {'*': -17.0, 'A': -6.0, 'B': -9.0, 'C': -15.0, 'D': -12.0, 'E': -9.0, 'F': -3.0, 'G': -10.0, 'H': -6.0, 'I': -1.0, 'K': -8.0, 'L': 7.0, 'M': 1.0, 'N': -7.0, 'P': -7.0, 'Q': -5.0, 'R': -8.0, 'S': -8.0, 'T': -7.0, 'V': -2.0, 'W': -6.0, 'X': -6.0, 'Y': -7.0, 'Z': -7.0},
                        'M': {'*': -17.0, 'A': -5.0, 'B': -10.0, 'C': -13.0, 'D': -11.0, 'E': -7.0, 'F': -4.0, 'G': -8.0, 'H': -10.0, 'I': -1.0, 'K': -2.0, 'L': 1.0, 'M': 11.0, 'N': -9.0, 'P': -8.0, 'Q': -4.0, 'R': -4.0, 'S': -5.0, 'T': -4.0, 'V': -1.0, 'W': -13.0, 'X': -5.0, 'Y': -11.0, 'Z': -5.0},
                        'N': {'*': -17.0, 'A': -4.0, 'B': 6.0, 'C': -11.0, 'D': 2.0, 'E': -2.0, 'F': -9.0, 'G': -3.0, 'H': 0.0, 'I': -5.0, 'K': -1.0, 'L': -7.0, 'M': -9.0, 'N': 8.0, 'P': -6.0, 'Q': -3.0, 'R': -6.0, 'S': 0.0, 'T': -2.0, 'V': -8.0, 'W': -8.0, 'X': -3.0, 'Y': -4.0, 'Z': -3.0},
                        'P': {'*': -17.0, 'A': -2.0, 'B': -7.0, 'C': -8.0, 'D': -8.0, 'E': -5.0, 'F': -10.0, 'G': -6.0, 'H': -4.0, 'I': -8.0, 'K': -6.0, 'L': -7.0, 'M': -8.0, 'N': -6.0, 'P': 8.0, 'Q': -3.0, 'R': -4.0, 'S': -2.0, 'T': -4.0, 'V': -6.0, 'W': -14.0, 'X': -5.0, 'Y': -13.0, 'Z': -4.0},
                        'Q': {'*': -17.0, 'A': -4.0, 'B': -3.0, 'C': -14.0, 'D': -2.0, 'E': 1.0, 'F': -13.0, 'G': -7.0, 'H': 1.0, 'I': -8.0, 'K': -3.0, 'L': -5.0, 'M': -4.0, 'N': -3.0, 'P': -3.0, 'Q': 8.0, 'R': -2.0, 'S': -5.0, 'T': -5.0, 'V': -7.0, 'W': -13.0, 'X': -5.0, 'Y': -12.0, 'Z': 6.0},
                        'R': {'*': -17.0, 'A': -7.0, 'B': -7.0, 'C': -8.0, 'D': -10.0, 'E': -9.0, 'F': -9.0, 'G': -9.0, 'H': -2.0, 'I': -5.0, 'K': 0.0, 'L': -8.0, 'M': -4.0, 'N': -6.0, 'P': -4.0, 'Q': -2.0, 'R': 8.0, 'S': -3.0, 'T': -6.0, 'V': -8.0, 'W': -2.0, 'X': -6.0, 'Y': -10.0, 'Z': -4.0},
                        'S': {'*': -17.0, 'A': 0.0, 'B': -1.0, 'C': -3.0, 'D': -4.0, 'E': -4.0, 'F': -6.0, 'G': -2.0, 'H': -6.0, 'I': -7.0, 'K': -4.0, 'L': -8.0, 'M': -5.0, 'N': 0.0, 'P': -2.0, 'Q': -5.0, 'R': -3.0, 'S': 6.0, 'T': 0.0, 'V': -6.0, 'W': -5.0, 'X': -3.0, 'Y': -7.0, 'Z': -5.0},
                        'T': {'*': -17.0, 'A': -1.0, 'B': -3.0, 'C': -8.0, 'D': -5.0, 'E': -6.0, 'F': -9.0, 'G': -6.0, 'H': -7.0, 'I': -2.0, 'K': -3.0, 'L': -7.0, 'M': -4.0, 'N': -2.0, 'P': -4.0, 'Q': -5.0, 'R': -6.0, 'S': 0.0, 'T': 7.0, 'V': -3.0, 'W': -13.0, 'X': -4.0, 'Y': -6.0, 'Z': -6.0},
                        'V': {'*': -17.0, 'A': -2.0, 'B': -8.0, 'C': -6.0, 'D': -8.0, 'E': -6.0, 'F': -8.0, 'G': -5.0, 'H': -6.0, 'I': 2.0, 'K': -9.0, 'L': -2.0, 'M': -1.0, 'N': -8.0, 'P': -6.0, 'Q': -7.0, 'R': -8.0, 'S': -6.0, 'T': -3.0, 'V': 7.0, 'W': -15.0, 'X': -5.0, 'Y': -7.0, 'Z': -6.0},
                        'W': {'*': -17.0, 'A': -13.0, 'B': -10.0, 'C': -15.0, 'D': -15.0, 'E': -17.0, 'F': -4.0, 'G': -15.0, 'H': -7.0, 'I': -14.0, 'K': -12.0, 'L': -6.0, 'M': -13.0, 'N': -8.0, 'P': -14.0, 'Q': -13.0, 'R': -2.0, 'S': -5.0, 'T': -13.0, 'V': -15.0, 'W': 13.0, 'X': -11.0, 'Y': -5.0, 'Z': -14.0},
                        'X': {'*': -17.0, 'A': -3.0, 'B': -5.0, 'C': -9.0, 'D': -5.0, 'E': -5.0, 'F': -8.0, 'G': -5.0, 'H': -5.0, 'I': -5.0, 'K': -5.0, 'L': -6.0, 'M': -5.0, 'N': -3.0, 'P': -5.0, 'Q': -5.0, 'R': -6.0, 'S': -3.0, 'T': -4.0, 'V': -5.0, 'W': -11.0, 'X': -5.0, 'Y': -7.0, 'Z': -5.0},
                        'Y': {'*': -17.0, 'A': -8.0, 'B': -6.0, 'C': -4.0, 'D': -11.0, 'E': -8.0, 'F': 2.0, 'G': -14.0, 'H': -3.0, 'I': -6.0, 'K': -9.0, 'L': -7.0, 'M': -11.0, 'N': -4.0, 'P': -13.0, 'Q': -12.0, 'R': -10.0, 'S': -7.0, 'T': -6.0, 'V': -7.0, 'W': -5.0, 'X': -7.0, 'Y': 10.0, 'Z': -9.0},
                        'Z': {'*': -17.0, 'A': -3.0, 'B': 0.0, 'C': -14.0, 'D': 1.0, 'E': 6.0, 'F': -13.0, 'G': -5.0, 'H': -1.0, 'I': -6.0, 'K': -4.0, 'L': -7.0, 'M': -5.0, 'N': -3.0, 'P': -4.0, 'Q': 6.0, 'R': -4.0, 'S': -5.0, 'T': -6.0, 'V': -6.0, 'W': -14.0, 'X': -5.0, 'Y': -9.0, 'Z': 6.0}
                },
                'PAM70': {
                        '*': {'*': 1.0, 'A': -11.0, 'B': -11.0, 'C': -11.0, 'D': -11.0, 'E': -11.0, 'F': -11.0, 'G': -11.0, 'H': -11.0, 'I': -11.0, 'K': -11.0, 'L': -11.0, 'M': -11.0, 'N': -11.0, 'P': -11.0, 'Q': -11.0, 'R': -11.0, 'S': -11.0, 'T': -11.0, 'V': -11.0, 'W': -11.0, 'X': -11.0, 'Y': -11.0, 'Z': -11.0},
                        'A': {'*': -11.0, 'A': 5.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'E': -1.0, 'F': -6.0, 'G': 0.0, 'H': -4.0, 'I': -2.0, 'K': -4.0, 'L': -4.0, 'M': -3.0, 'N': -2.0, 'P': 0.0, 'Q': -2.0, 'R': -4.0, 'S': 1.0, 'T': 1.0, 'V': -1.0, 'W': -9.0, 'X': -2.0, 'Y': -5.0, 'Z': -1.0},
                        'B': {'*': -11.0, 'A': -1.0, 'B': 5.0, 'C': -8.0, 'D': 5.0, 'E': 2.0, 'F': -7.0, 'G': -1.0, 'H': 0.0, 'I': -4.0, 'K': -1.0, 'L': -6.0, 'M': -6.0, 'N': 5.0, 'P': -4.0, 'Q': -1.0, 'R': -4.0, 'S': 0.0, 'T': -1.0, 'V': -5.0, 'W': -7.0, 'X': -2.0, 'Y': -4.0, 'Z': 1.0},
                        'C': {'*': -11.0, 'A': -4.0, 'B': -8.0, 'C': 9.0, 'D': -9.0, 'E': -9.0, 'F': -8.0, 'G': -6.0, 'H': -5.0, 'I': -4.0, 'K': -9.0, 'L': -10.0, 'M': -9.0, 'N': -7.0, 'P': -5.0, 'Q': -9.0, 'R': -5.0, 'S': -1.0, 'T': -5.0, 'V': -4.0, 'W': -11.0, 'X': -6.0, 'Y': -2.0, 'Z': -9.0},
                        'D': {'*': -11.0, 'A': -1.0, 'B': 5.0, 'C': -9.0, 'D': 6.0, 'E': 3.0, 'F': -10.0, 'G': -1.0, 'H': -1.0, 'I': -5.0, 'K': -2.0, 'L': -8.0, 'M': -7.0, 'N': 3.0, 'P': -4.0, 'Q': 0.0, 'R': -6.0, 'S': -1.0, 'T': -2.0, 'V': -5.0, 'W': -10.0, 'X': -3.0, 'Y': -7.0, 'Z': 2.0},
                        'E': {'*': -11.0, 'A': -1.0, 'B': 2.0, 'C': -9.0, 'D': 3.0, 'E': 6.0, 'F': -9.0, 'G': -2.0, 'H': -2.0, 'I': -4.0, 'K': -2.0, 'L': -6.0, 'M': -4.0, 'N': 0.0, 'P': -3.0, 'Q': 2.0, 'R': -5.0, 'S': -2.0, 'T': -3.0, 'V': -4.0, 'W': -11.0, 'X': -3.0, 'Y': -6.0, 'Z': 5.0},
                        'F': {'*': -11.0, 'A': -6.0, 'B': -7.0, 'C': -8.0, 'D': -10.0, 'E': -9.0, 'F': 8.0, 'G': -7.0, 'H': -4.0, 'I': 0.0, 'K': -9.0, 'L': -1.0, 'M': -2.0, 'N': -6.0, 'P': -7.0, 'Q': -9.0, 'R': -7.0, 'S': -4.0, 'T': -6.0, 'V': -5.0, 'W': -2.0, 'X': -5.0, 'Y': 4.0, 'Z': -9.0},
                        'G': {'*': -11.0, 'A': 0.0, 'B': -1.0, 'C': -6.0, 'D': -1.0, 'E': -2.0, 'F': -7.0, 'G': 6.0, 'H': -6.0, 'I': -6.0, 'K': -5.0, 'L': -7.0, 'M': -6.0, 'N': -1.0, 'P': -3.0, 'Q': -4.0, 'R': -6.0, 'S': 0.0, 'T': -3.0, 'V': -3.0, 'W': -10.0, 'X': -3.0, 'Y': -9.0, 'Z': -3.0},
                        'H': {'*': -11.0, 'A': -4.0, 'B': 0.0, 'C': -5.0, 'D': -1.0, 'E': -2.0, 'F': -4.0, 'G': -6.0, 'H': 8.0, 'I': -6.0, 'K': -3.0, 'L': -4.0, 'M': -6.0, 'N': 1.0, 'P': -2.0, 'Q': 2.0, 'R': 0.0, 'S': -3.0, 'T': -4.0, 'V': -4.0, 'W': -5.0, 'X': -3.0, 'Y': -1.0, 'Z': 1.0},
                        'I': {'*': -11.0, 'A': -2.0, 'B': -4.0, 'C': -4.0, 'D': -5.0, 'E': -4.0, 'F': 0.0, 'G': -6.0, 'H': -6.0, 'I': 7.0, 'K': -4.0, 'L': 1.0, 'M': 1.0, 'N': -3.0, 'P': -5.0, 'Q': -5.0, 'R': -3.0, 'S': -4.0, 'T': -1.0, 'V': 3.0, 'W': -9.0, 'X': -3.0, 'Y': -4.0, 'Z': -4.0},
                        'K': {'*': -11.0, 'A': -4.0, 'B': -1.0, 'C': -9.0, 'D': -2.0, 'E': -2.0, 'F': -9.0, 'G': -5.0, 'H': -3.0, 'I': -4.0, 'K': 6.0, 'L': -5.0, 'M': 0.0, 'N': 0.0, 'P': -4.0, 'Q': -1.0, 'R': 2.0, 'S': -2.0, 'T': -1.0, 'V': -6.0, 'W': -7.0, 'X': -3.0, 'Y': -7.0, 'Z': -2.0},
                        'L': {'*': -11.0, 'A': -4.0, 'B': -6.0, 'C': -10.0, 'D': -8.0, 'E': -6.0, 'F': -1.0, 'G': -7.0, 'H': -4.0, 'I': 1.0, 'K': -5.0, 'L': 6.0, 'M': 2.0, 'N': -5.0, 'P': -5.0, 'Q': -3.0, 'R': -6.0, 'S': -6.0, 'T': -4.0, 'V': 0.0, 'W': -4.0, 'X': -4.0, 'Y': -4.0, 'Z': -4.0},
                        'M': {'*': -11.0, 'A': -3.0, 'B': -6.0, 'C': -9.0, 'D': -7.0, 'E': -4.0, 'F': -2.0, 'G': -6.0, 'H': -6.0, 'I': 1.0, 'K': 0.0, 'L': 2.0, 'M': 10.0, 'N': -5.0, 'P': -5.0, 'Q': -2.0, 'R': -2.0, 'S': -3.0, 'T': -2.0, 'V': 0.0, 'W': -8.0, 'X': -3.0, 'Y': -7.0, 'Z': -3.0},
                        'N': {'*': -11.0, 'A': -2.0, 'B': 5.0, 'C': -7.0, 'D': 3.0, 'E': 0.0, 'F': -6.0, 'G': -1.0, 'H': 1.0, 'I': -3.0, 'K': 0.0, 'L': -5.0, 'M': -5.0, 'N': 6.0, 'P': -3.0, 'Q': -1.0, 'R': -3.0, 'S': 1.0, 'T': 0.0, 'V': -5.0, 'W': -6.0, 'X': -2.0, 'Y': -3.0, 'Z': -1.0},
                        'P': {'*': -11.0, 'A': 0.0, 'B': -4.0, 'C': -5.0, 'D': -4.0, 'E': -3.0, 'F': -7.0, 'G': -3.0, 'H': -2.0, 'I': -5.0, 'K': -4.0, 'L': -5.0, 'M': -5.0, 'N': -3.0, 'P': 7.0, 'Q': -1.0, 'R': -2.0, 'S': 0.0, 'T': -2.0, 'V': -3.0, 'W': -9.0, 'X': -3.0, 'Y': -9.0, 'Z': -2.0},
                        'Q': {'*': -11.0, 'A': -2.0, 'B': -1.0, 'C': -9.0, 'D': 0.0, 'E': 2.0, 'F': -9.0, 'G': -4.0, 'H': 2.0, 'I': -5.0, 'K': -1.0, 'L': -3.0, 'M': -2.0, 'N': -1.0, 'P': -1.0, 'Q': 7.0, 'R': 0.0, 'S': -3.0, 'T': -3.0, 'V': -4.0, 'W': -8.0, 'X': -2.0, 'Y': -8.0, 'Z': 5.0},
                        'R': {'*': -11.0, 'A': -4.0, 'B': -4.0, 'C': -5.0, 'D': -6.0, 'E': -5.0, 'F': -7.0, 'G': -6.0, 'H': 0.0, 'I': -3.0, 'K': 2.0, 'L': -6.0, 'M': -2.0, 'N': -3.0, 'P': -2.0, 'Q': 0.0, 'R': 8.0, 'S': -1.0, 'T': -4.0, 'V': -5.0, 'W': 0.0, 'X': -3.0, 'Y': -7.0, 'Z': -2.0},
                        'S': {'*': -11.0, 'A': 1.0, 'B': 0.0, 'C': -1.0, 'D': -1.0, 'E': -2.0, 'F': -4.0, 'G': 0.0, 'H': -3.0, 'I': -4.0, 'K': -2.0, 'L': -6.0, 'M': -3.0, 'N': 1.0, 'P': 0.0, 'Q': -3.0, 'R': -1.0, 'S': 5.0, 'T': 2.0, 'V': -3.0, 'W': -3.0, 'X': -1.0, 'Y': -5.0, 'Z': -2.0},
                        'T': {'*': -11.0, 'A': 1.0, 'B': -1.0, 'C': -5.0, 'D': -2.0, 'E': -3.0, 'F': -6.0, 'G': -3.0, 'H': -4.0, 'I': -1.0, 'K': -1.0, 'L': -4.0, 'M': -2.0, 'N': 0.0, 'P': -2.0, 'Q': -3.0, 'R': -4.0, 'S': 2.0, 'T': 6.0, 'V': -1.0, 'W': -8.0, 'X': -2.0, 'Y': -4.0, 'Z': -3.0},
                        'V': {'*': -11.0, 'A': -1.0, 'B': -5.0, 'C': -4.0, 'D': -5.0, 'E': -4.0, 'F': -5.0, 'G': -3.0, 'H': -4.0, 'I': 3.0, 'K': -6.0, 'L': 0.0, 'M': 0.0, 'N': -5.0, 'P': -3.0, 'Q': -4.0, 'R': -5.0, 'S': -3.0, 'T': -1.0, 'V': 6.0, 'W': -10.0, 'X': -2.0, 'Y': -5.0, 'Z': -4.0},
                        'W': {'*': -11.0, 'A': -9.0, 'B': -7.0, 'C': -11.0, 'D': -10.0, 'E': -11.0, 'F': -2.0, 'G': -10.0, 'H': -5.0, 'I': -9.0, 'K': -7.0, 'L': -4.0, 'M': -8.0, 'N': -6.0, 'P': -9.0, 'Q': -8.0, 'R': 0.0, 'S': -3.0, 'T': -8.0, 'V': -10.0, 'W': 13.0, 'X': -7.0, 'Y': -3.0, 'Z': -10.0},
                        'X': {'*': -11.0, 'A': -2.0, 'B': -2.0, 'C': -6.0, 'D': -3.0, 'E': -3.0, 'F': -5.0, 'G': -3.0, 'H': -3.0, 'I': -3.0, 'K': -3.0, 'L': -4.0, 'M': -3.0, 'N': -2.0, 'P': -3.0, 'Q': -2.0, 'R': -3.0, 'S': -1.0, 'T': -2.0, 'V': -2.0, 'W': -7.0, 'X': -3.0, 'Y': -5.0, 'Z': -3.0},
                        'Y': {'*': -11.0, 'A': -5.0, 'B': -4.0, 'C': -2.0, 'D': -7.0, 'E': -6.0, 'F': 4.0, 'G': -9.0, 'H': -1.0, 'I': -4.0, 'K': -7.0, 'L': -4.0, 'M': -7.0, 'N': -3.0, 'P': -9.0, 'Q': -8.0, 'R': -7.0, 'S': -5.0, 'T': -4.0, 'V': -5.0, 'W': -3.0, 'X': -5.0, 'Y': 9.0, 'Z': -7.0},
                        'Z': {'*': -11.0, 'A': -1.0, 'B': 1.0, 'C': -9.0, 'D': 2.0, 'E': 5.0, 'F': -9.0, 'G': -3.0, 'H': 1.0, 'I': -4.0, 'K': -2.0, 'L': -4.0, 'M': -3.0, 'N': -1.0, 'P': -2.0, 'Q': 5.0, 'R': -2.0, 'S': -2.0, 'T': -3.0, 'V': -4.0, 'W': -10.0, 'X': -3.0, 'Y': -7.0, 'Z': 5.0}
                },
                'RAO': {
                        'A': {'A': 16.0, 'C': 11.0, 'D': 9.0, 'E': 10.0, 'F': 10.0, 'G': 8.0, 'H': 11.0, 'I': 9.0, 'K': 10.0, 'L': 11.0, 'M': 11.0, 'N': 9.0, 'P': 6.0, 'Q': 11.0, 'R': 8.0, 'S': 10.0, 'T': 10.0, 'V': 9.0, 'W': 11.0, 'Y': 9.0},
                        'C': {'A': 11.0, 'C': 16.0, 'D': 8.0, 'E': 9.0, 'F': 10.0, 'G': 8.0, 'H': 10.0, 'I': 8.0, 'K': 9.0, 'L': 11.0, 'M': 10.0, 'N': 9.0, 'P': 7.0, 'Q': 10.0, 'R': 8.0, 'S': 10.0, 'T': 10.0, 'V': 8.0, 'W': 11.0, 'Y': 10.0},
                        'D': {'A': 9.0, 'C': 8.0, 'D': 16.0, 'E': 11.0, 'F': 4.0, 'G': 9.0, 'H': 9.0, 'I': 3.0, 'K': 11.0, 'L': 6.0, 'M': 5.0, 'N': 11.0, 'P': 8.0, 'Q': 11.0, 'R': 10.0, 'S': 10.0, 'T': 9.0, 'V': 3.0, 'W': 6.0, 'Y': 7.0},
                        'E': {'A': 10.0, 'C': 9.0, 'D': 11.0, 'E': 16.0, 'F': 6.0, 'G': 6.0, 'H': 11.0, 'I': 4.0, 'K': 11.0, 'L': 7.0, 'M': 8.0, 'N': 10.0, 'P': 5.0, 'Q': 11.0, 'R': 9.0, 'S': 9.0, 'T': 8.0, 'V': 4.0, 'W': 7.0, 'Y': 6.0},
                        'F': {'A': 10.0, 'C': 10.0, 'D': 4.0, 'E': 6.0, 'F': 16.0, 'G': 7.0, 'H': 9.0, 'I': 12.0, 'K': 6.0, 'L': 11.0, 'M': 10.0, 'N': 6.0, 'P': 4.0, 'Q': 7.0, 'R': 5.0, 'S': 8.0, 'T': 10.0, 'V': 11.0, 'W': 11.0, 'Y': 10.0},
                        'G': {'A': 8.0, 'C': 8.0, 'D': 9.0, 'E': 6.0, 'F': 7.0, 'G': 16.0, 'H': 7.0, 'I': 6.0, 'K': 7.0, 'L': 6.0, 'M': 4.0, 'N': 10.0, 'P': 11.0, 'Q': 8.0, 'R': 7.0, 'S': 11.0, 'T': 10.0, 'V': 6.0, 'W': 8.0, 'Y': 10.0},
                        'H': {'A': 11.0, 'C': 10.0, 'D': 9.0, 'E': 11.0, 'F': 9.0, 'G': 7.0, 'H': 16.0, 'I': 8.0, 'K': 11.0, 'L': 10.0, 'M': 10.0, 'N': 10.0, 'P': 5.0, 'Q': 11.0, 'R': 10.0, 'S': 10.0, 'T': 10.0, 'V': 9.0, 'W': 10.0, 'Y': 9.0},
                        'I': {'A': 9.0, 'C': 8.0, 'D': 3.0, 'E': 4.0, 'F': 12.0, 'G': 6.0, 'H': 8.0, 'I': 16.0, 'K': 4.0, 'L': 10.0, 'M': 9.0, 'N': 5.0, 'P': 3.0, 'Q': 6.0, 'R': 4.0, 'S': 8.0, 'T': 10.0, 'V': 12.0, 'W': 11.0, 'Y': 10.0},
                        'K': {'A': 10.0, 'C': 9.0, 'D': 11.0, 'E': 11.0, 'F': 6.0, 'G': 7.0, 'H': 11.0, 'I': 4.0, 'K': 16.0, 'L': 7.0, 'M': 8.0, 'N': 11.0, 'P': 6.0, 'Q': 12.0, 'R': 11.0, 'S': 10.0, 'T': 9.0, 'V': 5.0, 'W': 7.0, 'Y': 7.0},
                        'L': {'A': 11.0, 'C': 11.0, 'D': 6.0, 'E': 7.0, 'F': 11.0, 'G': 6.0, 'H': 10.0, 'I': 10.0, 'K': 7.0, 'L': 16.0, 'M': 11.0, 'N': 7.0, 'P': 4.0, 'Q': 9.0, 'R': 6.0, 'S': 8.0, 'T': 9.0, 'V': 10.0, 'W': 11.0, 'Y': 9.0},
                        'M': {'A': 11.0, 'C': 10.0, 'D': 5.0, 'E': 8.0, 'F': 10.0, 'G': 4.0, 'H': 10.0, 'I': 9.0, 'K': 8.0, 'L': 11.0, 'M': 16.0, 'N': 6.0, 'P': 2.0, 'Q': 9.0, 'R': 6.0, 'S': 7.0, 'T': 8.0, 'V': 9.0, 'W': 10.0, 'Y': 8.0},
                        'N': {'A': 9.0, 'C': 9.0, 'D': 11.0, 'E': 10.0, 'F': 6.0, 'G': 10.0, 'H': 10.0, 'I': 5.0, 'K': 11.0, 'L': 7.0, 'M': 6.0, 'N': 16.0, 'P': 9.0, 'Q': 11.0, 'R': 10.0, 'S': 11.0, 'T': 10.0, 'V': 5.0, 'W': 8.0, 'Y': 8.0},
                        'P': {'A': 6.0, 'C': 7.0, 'D': 8.0, 'E': 5.0, 'F': 4.0, 'G': 11.0, 'H': 5.0, 'I': 3.0, 'K': 6.0, 'L': 4.0, 'M': 2.0, 'N': 9.0, 'P': 16.0, 'Q': 7.0, 'R': 6.0, 'S': 10.0, 'T': 8.0, 'V': 3.0, 'W': 6.0, 'Y': 8.0},
                        'Q': {'A': 11.0, 'C': 10.0, 'D': 11.0, 'E': 11.0, 'F': 7.0, 'G': 8.0, 'H': 11.0, 'I': 6.0, 'K': 12.0, 'L': 9.0, 'M': 9.0, 'N': 11.0, 'P': 7.0, 'Q': 16.0, 'R': 10.0, 'S': 10.0, 'T': 10.0, 'V': 6.0, 'W': 9.0, 'Y': 8.0},
                        'R': {'A': 8.0, 'C': 8.0, 'D': 10.0, 'E': 9.0, 'F': 5.0, 'G': 7.0, 'H': 10.0, 'I': 4.0, 'K': 11.0, 'L': 6.0, 'M': 6.0, 'N': 10.0, 'P': 6.0, 'Q': 10.0, 'R': 16.0, 'S': 9.0, 'T': 9.0, 'V': 5.0, 'W': 7.0, 'Y': 7.0},
                        'S': {'A': 10.0, 'C': 10.0, 'D': 10.0, 'E': 9.0, 'F': 8.0, 'G': 11.0, 'H': 10.0, 'I': 8.0, 'K': 10.0, 'L': 8.0, 'M': 7.0, 'N': 11.0, 'P': 10.0, 'Q': 10.0, 'R': 9.0, 'S': 16.0, 'T': 11.0, 'V': 8.0, 'W': 10.0, 'Y': 11.0},
                        'T': {'A': 10.0, 'C': 10.0, 'D': 9.0, 'E': 8.0, 'F': 10.0, 'G': 10.0, 'H': 10.0, 'I': 10.0, 'K': 9.0, 'L': 9.0, 'M': 8.0, 'N': 10.0, 'P': 8.0, 'Q': 10.0, 'R': 9.0, 'S': 11.0, 'T': 16.0, 'V': 10.0, 'W': 11.0, 'Y': 11.0},
                        'V': {'A': 9.0, 'C': 8.0, 'D': 3.0, 'E': 4.0, 'F': 11.0, 'G': 6.0, 'H': 9.0, 'I': 12.0, 'K': 5.0, 'L': 10.0, 'M': 9.0, 'N': 5.0, 'P': 3.0, 'Q': 6.0, 'R': 5.0, 'S': 8.0, 'T': 10.0, 'V': 16.0, 'W': 11.0, 'Y': 10.0},
                        'W': {'A': 11.0, 'C': 11.0, 'D': 6.0, 'E': 7.0, 'F': 11.0, 'G': 8.0, 'H': 10.0, 'I': 11.0, 'K': 7.0, 'L': 11.0, 'M': 10.0, 'N': 8.0, 'P': 6.0, 'Q': 9.0, 'R': 7.0, 'S': 10.0, 'T': 11.0, 'V': 11.0, 'W': 16.0, 'Y': 11.0},
                        'Y': {'A': 9.0, 'C': 10.0, 'D': 7.0, 'E': 6.0, 'F': 10.0, 'G': 10.0, 'H': 9.0, 'I': 10.0, 'K': 7.0, 'L': 9.0, 'M': 8.0, 'N': 8.0, 'P': 8.0, 'Q': 8.0, 'R': 7.0, 'S': 11.0, 'T': 11.0, 'V': 10.0, 'W': 11.0, 'Y': 16.0}
                },
                'RISLER': {
                        'A': {'A': 2.2, 'C': -1.5, 'D': 0.2, 'E': 1.7, 'F': 0.6, 'G': 0.6, 'H': -0.6, 'I': 1.7, 'K': 1.4, 'L': 1.3, 'M': 1.0, 'N': 1.3, 'P': -0.2, 'Q': 1.8, 'R': 1.5, 'S': 2.0, 'T': 1.9, 'V': 2.0, 'W': -0.9, 'Y': 0.2},
                        'C': {'A': -1.5, 'C': 2.2, 'D': -1.7, 'E': -1.5, 'F': -1.6, 'G': -1.7, 'H': -1.8, 'I': -1.6, 'K': -1.6, 'L': -1.5, 'M': -1.6, 'N': -1.6, 'P': -1.8, 'Q': -1.4, 'R': -1.5, 'S': -1.3, 'T': -1.4, 'V': -1.4, 'W': -1.8, 'Y': -1.1},
                        'D': {'A': 0.2, 'C': -1.7, 'D': 2.2, 'E': 1.0, 'F': -0.3, 'G': -0.4, 'H': -1.3, 'I': 0.0, 'K': 0.1, 'L': -0.2, 'M': -0.5, 'N': 0.8, 'P': -1.2, 'Q': 0.6, 'R': -0.1, 'S': 0.7, 'T': 0.0, 'V': 0.0, 'W': -1.4, 'Y': -0.4},
                        'E': {'A': 1.7, 'C': -1.5, 'D': 1.0, 'E': 2.2, 'F': 0.6, 'G': 0.3, 'H': -0.6, 'I': 1.5, 'K': 1.4, 'L': 0.9, 'M': 0.6, 'N': 1.4, 'P': -0.1, 'Q': 2.1, 'R': 1.9, 'S': 1.8, 'T': 1.6, 'V': 1.6, 'W': -1.0, 'Y': 0.2},
                        'F': {'A': 0.6, 'C': -1.6, 'D': -0.3, 'E': 0.6, 'F': 2.2, 'G': -0.4, 'H': -1.1, 'I': 1.0, 'K': 0.1, 'L': 1.0, 'M': -0.2, 'N': 0.4, 'P': -1.1, 'Q': 0.7, 'R': 0.4, 'S': 0.5, 'T': 0.3, 'V': 0.8, 'W': -0.9, 'Y': 2.0},
                        'G': {'A': 0.6, 'C': -1.7, 'D': -0.4, 'E': 0.3, 'F': -0.4, 'G': 2.2, 'H': -1.2, 'I': 0.0, 'K': -0.1, 'L': -0.2, 'M': -0.4, 'N': 0.2, 'P': -1.2, 'Q': 0.2, 'R': 0.1, 'S': 0.7, 'T': 0.2, 'V': 0.1, 'W': -1.3, 'Y': -0.2},
                        'H': {'A': -0.6, 'C': -1.8, 'D': -1.3, 'E': -0.6, 'F': -1.1, 'G': -1.2, 'H': 2.2, 'I': -0.8, 'K': -1.0, 'L': -0.9, 'M': -1.2, 'N': -0.3, 'P': -1.6, 'Q': -0.5, 'R': -0.4, 'S': -0.4, 'T': -0.9, 'V': -0.7, 'W': -1.7, 'Y': -0.8},
                        'I': {'A': 1.7, 'C': -1.6, 'D': 0.0, 'E': 1.5, 'F': 1.0, 'G': 0.0, 'H': -0.8, 'I': 2.2, 'K': 1.0, 'L': 2.1, 'M': 0.9, 'N': 0.9, 'P': -0.6, 'Q': 1.4, 'R': 1.4, 'S': 1.6, 'T': 1.6, 'V': 2.2, 'W': -0.7, 'Y': 0.4},
                        'K': {'A': 1.4, 'C': -1.6, 'D': 0.1, 'E': 1.4, 'F': 0.1, 'G': -0.1, 'H': -1.0, 'I': 1.0, 'K': 2.2, 'L': 0.7, 'M': 0.4, 'N': 1.0, 'P': -0.7, 'Q': 1.7, 'R': 2.1, 'S': 1.4, 'T': 1.2, 'V': 1.2, 'W': -1.1, 'Y': 0.5},
                        'L': {'A': 1.3, 'C': -1.5, 'D': -0.2, 'E': 0.9, 'F': 1.0, 'G': -0.2, 'H': -0.9, 'I': 2.1, 'K': 0.7, 'L': 2.2, 'M': 1.8, 'N': 0.8, 'P': -0.8, 'Q': 1.1, 'R': 1.2, 'S': 1.3, 'T': 1.2, 'V': 2.0, 'W': -0.8, 'Y': 0.5},
                        'M': {'A': 1.0, 'C': -1.6, 'D': -0.5, 'E': 0.6, 'F': -0.2, 'G': -0.4, 'H': -1.2, 'I': 0.9, 'K': 0.4, 'L': 1.8, 'M': 2.2, 'N': 0.0, 'P': -1.2, 'Q': 1.2, 'R': 1.1, 'S': 0.6, 'T': 0.8, 'V': 0.8, 'W': -1.3, 'Y': -0.2},
                        'N': {'A': 1.3, 'C': -1.6, 'D': 0.8, 'E': 1.4, 'F': 0.4, 'G': 0.2, 'H': -0.3, 'I': 0.9, 'K': 1.0, 'L': 0.8, 'M': 0.0, 'N': 2.2, 'P': -1.0, 'Q': 1.6, 'R': 1.2, 'S': 1.9, 'T': 1.1, 'V': 1.1, 'W': -1.1, 'Y': -0.1},
                        'P': {'A': -0.2, 'C': -1.8, 'D': -1.2, 'E': -0.1, 'F': -1.1, 'G': -1.2, 'H': -1.6, 'I': -0.6, 'K': -0.7, 'L': -0.8, 'M': -1.2, 'N': -1.0, 'P': 2.2, 'Q': -0.6, 'R': -0.3, 'S': -0.3, 'T': -0.5, 'V': -0.6, 'W': -1.6, 'Y': -1.2},
                        'Q': {'A': 1.8, 'C': -1.4, 'D': 0.6, 'E': 2.1, 'F': 0.7, 'G': 0.2, 'H': -0.5, 'I': 1.4, 'K': 1.7, 'L': 1.1, 'M': 1.2, 'N': 1.6, 'P': -0.6, 'Q': 2.2, 'R': 2.0, 'S': 1.8, 'T': 1.7, 'V': 1.5, 'W': -1.0, 'Y': 0.5},
                        'R': {'A': 1.5, 'C': -1.5, 'D': -0.1, 'E': 1.9, 'F': 0.4, 'G': 0.1, 'H': -0.4, 'I': 1.4, 'K': 2.1, 'L': 1.2, 'M': 1.1, 'N': 1.2, 'P': -0.3, 'Q': 2.0, 'R': 2.2, 'S': 2.0, 'T': 1.9, 'V': 1.5, 'W': -0.8, 'Y': 0.8},
                        'S': {'A': 2.0, 'C': -1.3, 'D': 0.7, 'E': 1.8, 'F': 0.5, 'G': 0.7, 'H': -0.4, 'I': 1.6, 'K': 1.4, 'L': 1.3, 'M': 0.6, 'N': 1.9, 'P': -0.3, 'Q': 1.8, 'R': 2.0, 'S': 2.2, 'T': 2.1, 'V': 1.8, 'W': -0.8, 'Y': 0.4},
                        'T': {'A': 1.9, 'C': -1.4, 'D': 0.0, 'E': 1.6, 'F': 0.3, 'G': 0.2, 'H': -0.9, 'I': 1.6, 'K': 1.2, 'L': 1.2, 'M': 0.8, 'N': 1.1, 'P': -0.5, 'Q': 1.7, 'R': 1.9, 'S': 2.1, 'T': 2.2, 'V': 1.6, 'W': -1.0, 'Y': 0.3},
                        'V': {'A': 2.0, 'C': -1.4, 'D': 0.0, 'E': 1.6, 'F': 0.8, 'G': 0.1, 'H': -0.7, 'I': 2.2, 'K': 1.2, 'L': 2.0, 'M': 0.8, 'N': 1.1, 'P': -0.6, 'Q': 1.5, 'R': 1.5, 'S': 1.8, 'T': 1.6, 'V': 2.2, 'W': -0.7, 'Y': 0.3},
                        'W': {'A': -0.9, 'C': -1.8, 'D': -1.4, 'E': -1.0, 'F': -0.9, 'G': -1.3, 'H': -1.7, 'I': -0.7, 'K': -1.1, 'L': -0.8, 'M': -1.3, 'N': -1.1, 'P': -1.6, 'Q': -1.0, 'R': -0.8, 'S': -0.8, 'T': -1.0, 'V': -0.7, 'W': 2.2, 'Y': -0.6},
                        'Y': {'A': 0.2, 'C': -1.1, 'D': -0.4, 'E': 0.2, 'F': 2.0, 'G': -0.2, 'H': -0.8, 'I': 0.4, 'K': 0.5, 'L': 0.5, 'M': -0.2, 'N': -0.1, 'P': -1.2, 'Q': 0.5, 'R': 0.8, 'S': 0.4, 'T': 0.3, 'V': 0.3, 'W': -0.6, 'Y': 2.2}
                },
                'STR': {
                        'A': {'A': 4.0, 'C': -2.0, 'D': -1.0, 'E': 0.0, 'F': -3.0, 'G': 0.0, 'H': -2.0, 'I': -2.0, 'K': -1.0, 'L': -2.0, 'M': 0.0, 'N': -1.0, 'P': -1.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': -1.0, 'V': 0.0, 'W': -3.0, 'Y': -3.0},
                        'C': {'A': -2.0, 'C': 11.0, 'D': -7.0, 'E': -3.0, 'F': -2.0, 'G': -6.0, 'H': -6.0, 'I': -4.0, 'K': -4.0, 'L': -6.0, 'M': -5.0, 'N': -6.0, 'P': -8.0, 'Q': -3.0, 'R': -2.0, 'S': -4.0, 'T': -5.0, 'V': -4.0, 'W': -6.0, 'Y': -6.0},
                        'D': {'A': -1.0, 'C': -7.0, 'D': 6.0, 'E': 2.0, 'F': -5.0, 'G': -1.0, 'H': 0.0, 'I': -3.0, 'K': -1.0, 'L': -6.0, 'M': -4.0, 'N': 2.0, 'P': -1.0, 'Q': 0.0, 'R': -2.0, 'S': 0.0, 'T': -1.0, 'V': -4.0, 'W': -6.0, 'Y': -3.0},
                        'E': {'A': 0.0, 'C': -3.0, 'D': 2.0, 'E': 5.0, 'F': -4.0, 'G': -2.0, 'H': -2.0, 'I': -3.0, 'K': 1.0, 'L': -4.0, 'M': -2.0, 'N': 0.0, 'P': -1.0, 'Q': 2.0, 'R': 0.0, 'S': -1.0, 'T': 0.0, 'V': -2.0, 'W': -6.0, 'Y': -2.0},
                        'F': {'A': -3.0, 'C': -2.0, 'D': -5.0, 'E': -4.0, 'F': 7.0, 'G': -6.0, 'H': -2.0, 'I': 1.0, 'K': -3.0, 'L': 2.0, 'M': 0.0, 'N': -3.0, 'P': -5.0, 'Q': -4.0, 'R': -4.0, 'S': -3.0, 'T': -3.0, 'V': -1.0, 'W': 2.0, 'Y': 3.0},
                        'G': {'A': 0.0, 'C': -6.0, 'D': -1.0, 'E': -2.0, 'F': -6.0, 'G': 5.0, 'H': -3.0, 'I': -5.0, 'K': -3.0, 'L': -5.0, 'M': -4.0, 'N': -1.0, 'P': -2.0, 'Q': -2.0, 'R': -2.0, 'S': -1.0, 'T': -3.0, 'V': -4.0, 'W': -4.0, 'Y': -3.0},
                        'H': {'A': -2.0, 'C': -6.0, 'D': 0.0, 'E': -2.0, 'F': -2.0, 'G': -3.0, 'H': 8.0, 'I': -5.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 2.0, 'P': -3.0, 'Q': 0.0, 'R': 0.0, 'S': -2.0, 'T': -2.0, 'V': -2.0, 'W': -3.0, 'Y': 0.0},
                        'I': {'A': -2.0, 'C': -4.0, 'D': -3.0, 'E': -3.0, 'F': 1.0, 'G': -5.0, 'H': -5.0, 'I': 6.0, 'K': -3.0, 'L': 2.0, 'M': 1.0, 'N': -3.0, 'P': -4.0, 'Q': -5.0, 'R': -3.0, 'S': -3.0, 'T': -2.0, 'V': 2.0, 'W': -2.0, 'Y': -1.0},
                        'K': {'A': -1.0, 'C': -4.0, 'D': -1.0, 'E': 1.0, 'F': -3.0, 'G': -3.0, 'H': 0.0, 'I': -3.0, 'K': 5.0, 'L': -2.0, 'M': -1.0, 'N': 0.0, 'P': -1.0, 'Q': 1.0, 'R': 2.0, 'S': -1.0, 'T': 0.0, 'V': -3.0, 'W': -3.0, 'Y': -2.0},
                        'L': {'A': -2.0, 'C': -6.0, 'D': -6.0, 'E': -4.0, 'F': 2.0, 'G': -5.0, 'H': -3.0, 'I': 2.0, 'K': -2.0, 'L': 5.0, 'M': 3.0, 'N': -3.0, 'P': -3.0, 'Q': -3.0, 'R': -3.0, 'S': -4.0, 'T': -3.0, 'V': 1.0, 'W': -1.0, 'Y': -2.0},
                        'M': {'A': 0.0, 'C': -5.0, 'D': -4.0, 'E': -2.0, 'F': 0.0, 'G': -4.0, 'H': -2.0, 'I': 1.0, 'K': -1.0, 'L': 3.0, 'M': 8.0, 'N': -2.0, 'P': -6.0, 'Q': 1.0, 'R': -4.0, 'S': -4.0, 'T': -2.0, 'V': 0.0, 'W': -2.0, 'Y': -1.0},
                        'N': {'A': -1.0, 'C': -6.0, 'D': 2.0, 'E': 0.0, 'F': -3.0, 'G': -1.0, 'H': 2.0, 'I': -3.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 5.0, 'P': -2.0, 'Q': 0.0, 'R': -1.0, 'S': 0.0, 'T': 0.0, 'V': -4.0, 'W': -5.0, 'Y': -1.0},
                        'P': {'A': -1.0, 'C': -8.0, 'D': -1.0, 'E': -1.0, 'F': -5.0, 'G': -2.0, 'H': -3.0, 'I': -4.0, 'K': -1.0, 'L': -3.0, 'M': -6.0, 'N': -2.0, 'P': 7.0, 'Q': -2.0, 'R': -2.0, 'S': -1.0, 'T': -1.0, 'V': -4.0, 'W': -4.0, 'Y': -6.0},
                        'Q': {'A': 0.0, 'C': -3.0, 'D': 0.0, 'E': 2.0, 'F': -4.0, 'G': -2.0, 'H': 0.0, 'I': -5.0, 'K': 1.0, 'L': -3.0, 'M': 1.0, 'N': 0.0, 'P': -2.0, 'Q': 6.0, 'R': 1.0, 'S': -1.0, 'T': 0.0, 'V': -2.0, 'W': -5.0, 'Y': -3.0},
                        'R': {'A': -1.0, 'C': -2.0, 'D': -2.0, 'E': 0.0, 'F': -4.0, 'G': -2.0, 'H': 0.0, 'I': -3.0, 'K': 2.0, 'L': -3.0, 'M': -4.0, 'N': -1.0, 'P': -2.0, 'Q': 1.0, 'R': 7.0, 'S': 0.0, 'T': -1.0, 'V': -3.0, 'W': -2.0, 'Y': -1.0},
                        'S': {'A': 0.0, 'C': -4.0, 'D': 0.0, 'E': -1.0, 'F': -3.0, 'G': -1.0, 'H': -2.0, 'I': -3.0, 'K': -1.0, 'L': -4.0, 'M': -4.0, 'N': 0.0, 'P': -1.0, 'Q': -1.0, 'R': 0.0, 'S': 4.0, 'T': 1.0, 'V': -3.0, 'W': -5.0, 'Y': -2.0},
                        'T': {'A': -1.0, 'C': -5.0, 'D': -1.0, 'E': 0.0, 'F': -3.0, 'G': -3.0, 'H': -2.0, 'I': -2.0, 'K': 0.0, 'L': -3.0, 'M': -2.0, 'N': 0.0, 'P': -1.0, 'Q': 0.0, 'R': -1.0, 'S': 1.0, 'T': 5.0, 'V': -1.0, 'W': -5.0, 'Y': -2.0},
                        'V': {'A': 0.0, 'C': -4.0, 'D': -4.0, 'E': -2.0, 'F': -1.0, 'G': -4.0, 'H': -2.0, 'I': 2.0, 'K': -3.0, 'L': 1.0, 'M': 0.0, 'N': -4.0, 'P': -4.0, 'Q': -2.0, 'R': -3.0, 'S': -3.0, 'T': -1.0, 'V': 5.0, 'W': -4.0, 'Y': -1.0},
                        'W': {'A': -3.0, 'C': -6.0, 'D': -6.0, 'E': -6.0, 'F': 2.0, 'G': -4.0, 'H': -3.0, 'I': -2.0, 'K': -3.0, 'L': -1.0, 'M': -2.0, 'N': -5.0, 'P': -4.0, 'Q': -5.0, 'R': -2.0, 'S': -5.0, 'T': -5.0, 'V': -4.0, 'W': 10.0, 'Y': 2.0},
                        'Y': {'A': -3.0, 'C': -6.0, 'D': -3.0, 'E': -2.0, 'F': 3.0, 'G': -3.0, 'H': 0.0, 'I': -1.0, 'K': -2.0, 'L': -2.0, 'M': -1.0, 'N': -1.0, 'P': -6.0, 'Q': -3.0, 'R': -1.0, 'S': -2.0, 'T': -2.0, 'V': -1.0, 'W': 2.0, 'Y': 7.0}}
        },
        'DNA': {
                'BLASTN': {
                        'A': {'A': 2.0, 'B': -3.0, 'C': -3.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -3.0, 'T': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -3.0},
                        'B': {'A': -3.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'C': {'A': -3.0, 'B': -1.0, 'C': 2.0, 'D': -3.0, 'G': -3.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -2.0, 'R': -3.0, 'S': -1.0, 'T': -3.0, 'V': -1.0, 'W': -3.0, 'Y': -1.0},
                        'D': {'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'G': {'A': -3.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': 2.0, 'H': -3.0, 'K': -1.0, 'M': -3.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -3.0, 'V': -1.0, 'W': -3.0, 'Y': -3.0},
                        'H': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'K': {'A': -3.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -3.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'M': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'N': {'A': -2.0, 'B': -2.0, 'C': -2.0, 'D': -2.0, 'G': -2.0, 'H': -2.0, 'K': -2.0, 'M': -2.0, 'N': -2.0, 'R': -2.0, 'S': -2.0, 'T': -2.0, 'V': -2.0, 'W': -2.0, 'Y': -2.0},
                        'R': {'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -3.0},
                        'S': {'A': -3.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -3.0, 'V': -1.0, 'W': -3.0, 'Y': -1.0},
                        'T': {'A': -3.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -3.0, 'N': -2.0, 'R': -3.0, 'S': -3.0, 'T': 2.0, 'V': -3.0, 'W': -1.0, 'Y': -1.0},
                        'V': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'T': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'W': {'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -3.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'Y': {'A': -3.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -3.0, 'S': -1.0, 'T': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0}
                },
                'NUC.4.4': {
                        'A': {'A': 5.0, 'B': -4.0, 'C': -4.0, 'D': -1.0, 'G': -4.0, 'H': -1.0, 'K': -4.0, 'M': 1.0, 'N': -2.0, 'R': 1.0, 'S': -4.0, 'T': -4.0, 'V': -1.0, 'W': 1.0, 'Y': -4.0},
                        'B': {'A': -4.0, 'B': -1.0, 'C': -1.0, 'D': -2.0, 'G': -1.0, 'H': -2.0, 'K': -1.0, 'M': -3.0, 'N': -1.0, 'R': -3.0, 'S': -1.0, 'T': -1.0, 'V': -2.0, 'W': -3.0, 'Y': -1.0},
                        'C': {'A': -4.0, 'B': -1.0, 'C': 5.0, 'D': -4.0, 'G': -4.0, 'H': -1.0, 'K': -4.0, 'M': 1.0, 'N': -2.0, 'R': -4.0, 'S': 1.0, 'T': -4.0, 'V': -1.0, 'W': -4.0, 'Y': 1.0},
                        'D': {'A': -1.0, 'B': -2.0, 'C': -4.0, 'D': -1.0, 'G': -1.0, 'H': -2.0, 'K': -1.0, 'M': -3.0, 'N': -1.0, 'R': -1.0, 'S': -3.0, 'T': -1.0, 'V': -2.0, 'W': -1.0, 'Y': -3.0},
                        'G': {'A': -4.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'G': 5.0, 'H': -4.0, 'K': 1.0, 'M': -4.0, 'N': -2.0, 'R': 1.0, 'S': 1.0, 'T': -4.0, 'V': -1.0, 'W': -4.0, 'Y': -4.0},
                        'H': {'A': -1.0, 'B': -2.0, 'C': -1.0, 'D': -2.0, 'G': -4.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -1.0, 'R': -3.0, 'S': -3.0, 'T': -1.0, 'V': -2.0, 'W': -1.0, 'Y': -1.0},
                        'K': {'A': -4.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'G': 1.0, 'H': -3.0, 'K': -1.0, 'M': -4.0, 'N': -1.0, 'R': -2.0, 'S': -2.0, 'T': 1.0, 'V': -3.0, 'W': -2.0, 'Y': -2.0},
                        'M': {'A': 1.0, 'B': -3.0, 'C': 1.0, 'D': -3.0, 'G': -4.0, 'H': -1.0, 'K': -4.0, 'M': -1.0, 'N': -1.0, 'R': -2.0, 'S': -2.0, 'T': -4.0, 'V': -1.0, 'W': -2.0, 'Y': -2.0},
                        'N': {'A': -2.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -2.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'R': {'A': 1.0, 'B': -3.0, 'C': -4.0, 'D': -1.0, 'G': 1.0, 'H': -3.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -1.0, 'S': -2.0, 'T': -4.0, 'V': -1.0, 'W': -2.0, 'Y': -4.0},
                        'S': {'A': -4.0, 'B': -1.0, 'C': 1.0, 'D': -3.0, 'G': 1.0, 'H': -3.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -2.0, 'S': -1.0, 'T': -4.0, 'V': -1.0, 'W': -4.0, 'Y': -2.0},
                        'T': {'A': -4.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'G': -4.0, 'H': -1.0, 'K': 1.0, 'M': -4.0, 'N': -2.0, 'R': -4.0, 'S': -4.0, 'T': 5.0, 'V': -4.0, 'W': 1.0, 'Y': 1.0},
                        'V': {'A': -1.0, 'B': -2.0, 'C': -1.0, 'D': -2.0, 'G': -1.0, 'H': -2.0, 'K': -3.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'T': -4.0, 'V': -1.0, 'W': -3.0, 'Y': -3.0},
                        'W': {'A': 1.0, 'B': -3.0, 'C': -4.0, 'D': -1.0, 'G': -4.0, 'H': -1.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -2.0, 'S': -4.0, 'T': 1.0, 'V': -3.0, 'W': -1.0, 'Y': -2.0},
                        'Y': {'A': -4.0, 'B': -1.0, 'C': 1.0, 'D': -3.0, 'G': -4.0, 'H': -1.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -4.0, 'S': -2.0, 'T': 1.0, 'V': -3.0, 'W': -2.0, 'Y': -1.0}},
                'TRANS': {
                        'A': {'A': 5.0, 'C': 0.0, 'G': 4.0, 'T': 0.0},
                        'C': {'A': 0.0, 'C': 5.0, 'G': 0.0, 'T': 4.0},
                        'G': {'A': 4.0, 'C': 0.0, 'G': 5.0, 'T': 0.0},
                        'T': {'A': 0.0, 'C': 4.0, 'G': 0.0, 'T': 5.0}}
        },
        'RNA': {
                'BLASTN': {
                        'A': {'A': 2.0, 'B': -3.0, 'C': -3.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -3.0, 'U': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -3.0},
                        'B': {'A': -3.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'C': {'A': -3.0, 'B': -1.0, 'C': 2.0, 'D': -3.0, 'G': -3.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -2.0, 'R': -3.0, 'S': -1.0, 'U': -3.0, 'V': -1.0, 'W': -3.0, 'Y': -1.0},
                        'D': {'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'G': {'A': -3.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': 2.0, 'H': -3.0, 'K': -1.0, 'M': -3.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -3.0, 'V': -1.0, 'W': -3.0, 'Y': -3.0},
                        'H': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'K': {'A': -3.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -3.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'M': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'N': {'A': -2.0, 'B': -2.0, 'C': -2.0, 'D': -2.0, 'G': -2.0, 'H': -2.0, 'K': -2.0, 'M': -2.0, 'N': -2.0, 'R': -2.0, 'S': -2.0, 'U': -2.0, 'V': -2.0, 'W': -2.0, 'Y': -2.0},
                        'R': {'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -3.0},
                        'S': {'A': -3.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -3.0, 'V': -1.0, 'W': -3.0, 'Y': -1.0},
                        'U': {'A': -3.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -3.0, 'N': -2.0, 'R': -3.0, 'S': -3.0, 'U': 2.0, 'V': -3.0, 'W': -1.0, 'Y': -1.0},
                        'V': {'A': -1.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -1.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -1.0, 'U': -3.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'W': {'A': -1.0, 'B': -1.0, 'C': -3.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -1.0, 'S': -3.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'Y': {'A': -3.0, 'B': -1.0, 'C': -1.0, 'D': -1.0, 'G': -3.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -2.0, 'R': -3.0, 'S': -1.0, 'U': -1.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0}
                },
                'NUC.4.4': {
                        'A': {'A': 5.0, 'B': -4.0, 'C': -4.0, 'D': -1.0, 'G': -4.0, 'H': -1.0, 'K': -4.0, 'M': 1.0, 'N': -2.0, 'R': 1.0, 'S': -4.0, 'U': -4.0, 'V': -1.0, 'W': 1.0, 'Y': -4.0},
                        'B': {'A': -4.0, 'B': -1.0, 'C': -1.0, 'D': -2.0, 'G': -1.0, 'H': -2.0, 'K': -1.0, 'M': -3.0, 'N': -1.0, 'R': -3.0, 'S': -1.0, 'U': -1.0, 'V': -2.0, 'W': -3.0, 'Y': -1.0},
                        'C': {'A': -4.0, 'B': -1.0, 'C': 5.0, 'D': -4.0, 'G': -4.0, 'H': -1.0, 'K': -4.0, 'M': 1.0, 'N': -2.0, 'R': -4.0, 'S': 1.0, 'U': -4.0, 'V': -1.0, 'W': -4.0, 'Y': 1.0},
                        'D': {'A': -1.0, 'B': -2.0, 'C': -4.0, 'D': -1.0, 'G': -1.0, 'H': -2.0, 'K': -1.0, 'M': -3.0, 'N': -1.0, 'R': -1.0, 'S': -3.0, 'U': -1.0, 'V': -2.0, 'W': -1.0, 'Y': -3.0},
                        'G': {'A': -4.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'G': 5.0, 'H': -4.0, 'K': 1.0, 'M': -4.0, 'N': -2.0, 'R': 1.0, 'S': 1.0, 'U': -4.0, 'V': -1.0, 'W': -4.0, 'Y': -4.0},
                        'H': {'A': -1.0, 'B': -2.0, 'C': -1.0, 'D': -2.0, 'G': -4.0, 'H': -1.0, 'K': -3.0, 'M': -1.0, 'N': -1.0, 'R': -3.0, 'S': -3.0, 'U': -1.0, 'V': -2.0, 'W': -1.0, 'Y': -1.0},
                        'K': {'A': -4.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'G': 1.0, 'H': -3.0, 'K': -1.0, 'M': -4.0, 'N': -1.0, 'R': -2.0, 'S': -2.0, 'U': 1.0, 'V': -3.0, 'W': -2.0, 'Y': -2.0},
                        'M': {'A': 1.0, 'B': -3.0, 'C': 1.0, 'D': -3.0, 'G': -4.0, 'H': -1.0, 'K': -4.0, 'M': -1.0, 'N': -1.0, 'R': -2.0, 'S': -2.0, 'U': -4.0, 'V': -1.0, 'W': -2.0, 'Y': -2.0},
                        'N': {'A': -2.0, 'B': -1.0, 'C': -2.0, 'D': -1.0, 'G': -2.0, 'H': -1.0, 'K': -1.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'U': -2.0, 'V': -1.0, 'W': -1.0, 'Y': -1.0},
                        'R': {'A': 1.0, 'B': -3.0, 'C': -4.0, 'D': -1.0, 'G': 1.0, 'H': -3.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -1.0, 'S': -2.0, 'U': -4.0, 'V': -1.0, 'W': -2.0, 'Y': -4.0},
                        'S': {'A': -4.0, 'B': -1.0, 'C': 1.0, 'D': -3.0, 'G': 1.0, 'H': -3.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -2.0, 'S': -1.0, 'U': -4.0, 'V': -1.0, 'W': -4.0, 'Y': -2.0},
                        'U': {'A': -4.0, 'B': -1.0, 'C': -4.0, 'D': -1.0, 'G': -4.0, 'H': -1.0, 'K': 1.0, 'M': -4.0, 'N': -2.0, 'R': -4.0, 'S': -4.0, 'U': 5.0, 'V': -4.0, 'W': 1.0, 'Y': 1.0},
                        'V': {'A': -1.0, 'B': -2.0, 'C': -1.0, 'D': -2.0, 'G': -1.0, 'H': -2.0, 'K': -3.0, 'M': -1.0, 'N': -1.0, 'R': -1.0, 'S': -1.0, 'U': -4.0, 'V': -1.0, 'W': -3.0, 'Y': -3.0},
                        'W': {'A': 1.0, 'B': -3.0, 'C': -4.0, 'D': -1.0, 'G': -4.0, 'H': -1.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -2.0, 'S': -4.0, 'U': 1.0, 'V': -3.0, 'W': -1.0, 'Y': -2.0},
                        'Y': {'A': -4.0, 'B': -1.0, 'C': 1.0, 'D': -3.0, 'G': -4.0, 'H': -1.0, 'K': -2.0, 'M': -2.0, 'N': -1.0, 'R': -4.0, 'S': -2.0, 'U': 1.0, 'V': -3.0, 'W': -2.0, 'Y': -1.0}
                },
                'TRANS': {
                        'A': {'A': 5.0, 'C': 0.0, 'G': 4.0, 'U': 0.0},
                        'C': {'A': 0.0, 'C': 5.0, 'G': 0.0, 'U': 4.0},
                        'G': {'A': 4.0, 'C': 0.0, 'G': 5.0, 'U': 0.0},
                        'U': {'A': 0.0, 'C': 4.0, 'G': 0.0, 'U': 5.0}
                }
        }
}


# basic color mapping
IDENTITY_COLORS = {
        0: {'type': 'identical',
            'color': 'lightgrey'},
        -1: {'type': 'different',
            'color': 'peru'},
        -2: {'type': 'mask',
            'color': 'dimgrey'},
        -3: {'type': 'ambiguity',
            'color': 'black'}
}

# extended color mappings
CHAR_COLORS = {
    'DNA': {
        'standard': {
            'A': '#F2A45E',
            'T': '#F26D6D',
            'G': '#64A87B',
            'C': '#7095D1',
            '-': 'silver'
        },
        'purine_pyrimidine': {
            'A': '#A18C6C',
            'G': '#A18C6C',
            'C': '#7994B1',
            'T': '#7994B1',
            '-': 'silver'
        },
        'strong_weak': {
            'G': '#6B9E88',
            'C': '#6B9E88',
            'A': '#CC9C6B',
            'T': '#CC9C6B',
            '-': 'silver'
        }
    },
    'RNA': {
        'standard': {
            'A': '#F2A45E',
            'U': '#F26D6D',
            'G': '#64A87B',
            'C': '#7095D1',
            '-': 'silver'
        },
        'purine_pyrimidine': {
            'A': '#A18C6C',
            'G': '#A18C6C',
            'C': '#7994B1',
            'U': '#7994B1',
            '-': 'silver'
        },
        'strong_weak': {
            'G': '#6B9E88',
            'C': '#6B9E88',
            'A': '#CC9C6B',
            'U': '#CC9C6B',
            '-': 'silver'
        }
    },
    'AA': {
        'standard': {
                'A': '#D99058', 'V': '#E2A876', 'I': '#E0A45F', 'L': '#DA8A54', 'M': '#D99058',
                'F': '#B05B61', 'W': '#805F9C', 'S': '#D98B81', 'T': '#D9A86B', 'N': '#AB6D9F',
                'Q': '#986DB2', 'Y': '#C76F61', 'C': '#D1B758', 'D': '#D96666', 'E': '#D96F5F',
                'K': '#5D93C4', 'R': '#5DA5A4', 'H': '#916FA7', 'G': '#6FAF99', 'P': '#77A86F',
                '-': 'silver'
        },
        'zappo': {
                'I': '#b89494', 'L': '#b89494', 'V': '#b89494', 'A': '#b89494', 'M': '#b89494',
                'F': '#c1a14c', 'W': '#c1a14c', 'Y': '#c1a14c',
                'K': '#5c6ec2', 'R': '#5c6ec2', 'H': '#5c6ec2',
                'D': '#b45050', 'E': '#b45050',
                'S': '#5cbf67', 'T': '#5cbf67', 'N': '#5cbf67', 'Q': '#5cbf67',
                'P': '#cc85cc', 'G': '#cc85cc',
                'C': '#e0c64a',
                '-': 'silver'
        },
        'hydrophobicity': {
                'I': '#cc3333',
                'V': '#c73c3c',
                'L': '#b84747',
                'F': '#993f4f',
                'C': '#8d4053',
                'M': '#7c4a5f',
                'A': '#794d60',
                'G': '#6e5b8e',
                'X': '#6a5890',
                'T': '#655593',
                'S': '#615297',
                'W': '#5e509a',
                'Y': '#554c9f',
                'P': '#4f48a3',
                'H': '#3f4db0',
                'E': '#374fb5',
                'Z': '#374fb5',
                'Q': '#374fb5',
                'D': '#374fb5',
                'B': '#374fb5',
                'N': '#374fb5',
                'K': '#2d52bb',
                'R': '#2d52bb',
                '-': 'silver'
        },
        'clustal': {
                'A': '#6e8cc2', 'I': '#6e8cc2', 'L': '#6e8cc2',
                'M': '#6e8cc2', 'F': '#6e8cc2', 'W': '#6e8cc2', 'V': '#6e8cc2',
                'K': '#c54a3c', 'R': '#c54a3c',
                'D': '#aa64aa', 'E': '#aa64aa',
                'N': '#58b858', 'Q': '#58b858', 'S': '#58b858', 'T': '#58b858',
                'G': '#db9c66',
                'P': '#c2c24d',
                'C': '#c9aaaa',
                'H': '#5db3b3', 'Y': '#5db3b3',
                '-': '#f0f0f0'
        }
    }
}

# used for group allocation in legend
CHAR_GROUPS = {
    'DNA': {
        'purine_pyrimidine': {
                'Purines (A, G)': ['A', 'G'],
                'Pyrimidines (C, T)': ['C', 'T'],
        },
        'strong_weak': {
                'Strong (G≡C)': ['G', 'C'],
                'Weak (A=T)': ['A', 'T'],
        }
    },
    'RNA': {
        'purine_pyrimidine': {
                'Purines (A, G)': ['A', 'G'],
                'Pyrimidines (C, U)': ['C', 'U'],
        },
        'strong_weak': {
                'Strong (G≡C)': ['G', 'C'],
                'Weak (A=U)': ['A', 'U'],
        }
    },
    'AA': {
        'zappo': {
                'aliphatic/hydrophobic': ['I', 'L', 'V', 'A', 'M'],
                'aromatic': ['F', 'W', 'Y'],
                'positive': ['K', 'R', 'H'],
                'negative': ['D', 'E'],
                'hydrophilic': ['S', 'T', 'N', 'Q'],
                'conformationally special': ['P', 'G'],
                'cysteine': ['C']
        },
        'hydrophobicity': {
                'hydrophobic': ['I', 'V', 'L', 'F', 'C', 'M', 'A'],
                'moderate': ['T', 'S', 'W', 'Y', 'P', 'G'],
                'hydrophilic': ['H', 'E', 'Q', 'D', 'N', 'K', 'R']
        },
        'clustal': {
                'hydrophobic': ['A', 'I', 'L', 'M', 'F', 'W', 'V'],
                'positive': ['K', 'R'],
                'negative': ['D', 'E'],
                'polar': ['N', 'Q', 'S', 'T'],
                'cysteine': ['C'],
                'glycine': ['G'],
                'proline': ['P'],
                'aromatic': ['H', 'Y']
        }
    }
}

# shiny app specific settings
STANDARD_HEIGHT_RATIOS = {
        5: (1, 5, 2),
        10: (1, 10, 2),
        20: (1, 14, 2),
        30: (1, 18, 2),
        40: (1, 22, 2),
        50: (1, 26, 2),
        60: (1, 30, 2),
        70: (1, 34, 2),
        80: (1, 38, 2),
        90: (1, 42, 2),
        100: (1, 46, 2),
        200: (1, 50, 2),
        500: (1, 54, 2)
}